################################################################
use strict;

=head1 LICENSE

   Bweb - A Bacula web interface
   Bacula® - The Network Backup Solution

   Copyright (C) 2010-2014 Bacula Systems(R) SA
   All rights reserved.

   The main author of Bweb is Eric Bollengier.

   Bacula® is a registered trademark of Kern Sibbald.
   Bacula Systems(R) is a trademark of Bacula Systems SA.
   Bacula Enterprise(TM) is a trademark of Bacula Systems SA.

   The licensor of Bacula Enterprise(TM) is Bacula Systems(R) SA,
   Rue Galilee 5, 1400 Yverdon-les-Bains, Switzerland.

   Licensees holding a valid Bacula Systems SA license may use this file
   and others of this release in accordance with the proprietary license
   agreement provided in the LICENSE file. Redistribution of any part of
   this release is not permitted.

=cut

# Treat strings and regex modifiers as Unicode strings
eval('use feature "unicode_strings"');
use Encode;

# Print stderr messages with timestamp
sub STDERR {
    my @out = @_;
    my $ret;
    if (scalar(@out) > 0) { # called as a function
        # add timestamp and print output
        my $timestamp = localtime();
        my @outlog = map { Encode::encode_utf8($_) } @out;
        print STDERR "$timestamp: @outlog";
        $ret = '';
    } else { # called as file handler
        # Return real STDERR for external modules that refer to STDERR
        # as I/O typeglob, not as a function.
        $ret = \*STDERR;
    }
    return $ret;
}

# Print Perl warnings with timestamp
$SIG{__WARN__} = sub {
    STDERR(@_);
};

# When used Lighttpd on Ubuntu the HOME environment variable can be unavailable
# due to not passing it here by the Lighttpd. It is Ubuntu bug #1610553.
# Here the HOME value is retrieved to original state if it doesn't exist.
if (!defined($ENV{HOME})) {
    my @pwfields = getpwuid($<);
    $ENV{HOME} = $pwfields[7];
}

package Bweb::Gui;

=head1 PACKAGE

    Bweb::Gui - Base package for all Bweb object

=head2 DESCRIPTION

    This package define base fonction like new, display, etc..

=cut

use CGI '-utf8';
use CGI::Session;
use HTML::Template;
our $template_dir='/opt/bweb/tpl';

our $default_lang = 'en';
our $cgi_session;

# Global parameters that existence don't cause sending common HTTP headers
my %hold_headers_params = ();

=head1 FUNCTION

    new - creation a of new Bweb object

=head2 DESCRIPTION

    This function take an hash of argument and place them
    on bless ref

    IE : $obj = new Obj(name => 'test', age => '10');

         $obj->{name} eq 'test' and $obj->{age} eq 10

=cut

sub new
{
    my ($class, %arg) = @_;
    my $self = bless {
        name => undef,
        info => undef,
        translate => undef
    }, $class;

    map { $self->{lc($_)} = $arg{$_} } keys %arg ;
    $self->set_username();

    return $self;
}

sub debug
{
    my ($self, $what) = @_;
    if ($self->{debug}) {
        print STDERR Data::Dumper::Dumper($what);
    }
}

sub edebug
{
    my ($self, $what) = @_;
    my ($pkg, $filename, $line) = caller;
    if (ref $what) {
        print STDERR "$filename:$line ", Data::Dumper::Dumper($what);
    } else {
        print STDERR "$filename:$line ", $what;
    }
}

sub fdebug
{
    my ($self, $what) = @_;

    my $old = $self->{debug};
    $self->{debug}=1;
    $self->debug($what);
    $self->{debug} = $old;
}

=head1 FUNCTION

    error - display an error to the user

=head2 DESCRIPTION

    this function set $self->{error} with arg, display a message with
    error.tpl and return 0

=head2 EXAMPLE

    unless (...) {
        return $self->error("Can't use this file");
    }

=cut

sub error
{
    my ($self, $what, $args) = @_;
    if (defined($args)) {
        $args->{error} = $what;
    } else {
        $args = {error => $what};
    }
    $self->display($args, 'error.tpl');
    return 0;
}

# send content type the first time, see man CGI to overwrite
# values
#
# NOTE! Temporary disable sending content type. The goal is to remove it because
# content type is already sent in CGI session for each request.
my $send_content_type_done=1;
sub send_content_type
{
    my ($self, %arg) = @_;
    my $info = $self->{info} || $self;

    if (!$send_content_type_done) { # display it once
        $send_content_type_done = 1;

        %arg = (-type => 'text/html', -charset => 'UTF-8', %arg);
#        if ($info->{current_conf}) {
#            $arg{'-cookies'} = [CGI::cookie(
#                                    -name=>'dir',
#                                    -value=>$info->{current_conf})];
#        } elsif ($info->{subconf}) {
#            $arg{'-cookies'} = [CGI::cookie( # try to remove it
#                                    -name=>'dir',
#                                    -value=>'')];
#        }
        print CGI::header(%arg);
    }
}

=head1 FUNCTION

    display - display an html page with HTML::Template

=head2 DESCRIPTION

    this function is use to render all html codes. it takes an
    ref hash as arg in which all param are usable in template.

    it will use user template_dir then global template_dir 
    to search the template file.

    hash keys are not sensitive. See HTML::Template for more
    explanations about the hash ref. (it's can be quiet hard to understand) 

    It uses the following variables: template_dir lang director

=head2 EXAMPLE

    $ref = { name => 'me', age => 26 };
    $self->display($ref, "people.tpl");

=cut

sub display
{
    my ($self, $hash, $tpl) = @_ ;
    my $info = $self->{info} || $self;

    my $template_paths = $self->get_template_paths();
    my $template = HTML::Template->new(filename => $tpl,
                                       path => $template_paths,
                                       die_on_bad_params => 0,
                                       case_sensitive => 0);

    foreach my $var (qw/limit offset/) {

        unless ($hash->{$var}) {
            my $value = CGI::param($var) || '';

            if ($value =~ /^(\d+)$/) {
                $template->param($var, $1) ;
            }
        }
    }

    $template->param('bweb_js_version', $Bweb::BWEB_JS_VERSION);
    $template->param('bconfig_js_version', $Bweb::BCONFIG_JS_VERSION);
    if ($self->can('get_ver')) {
        $template->param('bweb_version', $self->get_ver());
    }

    $template->param('thisurl', CGI::url(-relative => 1, -query=>1));
    $template->param('loginname', $self->{loginname});

    if ($info->{current_conf}) {
        my $cur = $info->{current_conf};
        $template->param('cur_conf', $cur);
        $template->param('cur_name', $info->{subconf}->{$cur}->{name});
    }
    if ($info->{subconf}) {
        my $l = $info->{subconf}; #{ key => { name => 'key name' },... }
        $template->param('DirectorDefined',
         [ map { {id => $_, name => $l->{$_}->{name}} } keys %$l ]);
    }

    # Prepare plugins configuration for templates
    $self->setup_plugins_menu($template);

    # TODO: Add/Remove bconfig buttons if available
    #if (!$self->cant_do('r_configure')) {
        $template->param('DisplayBconfig', 1);
    #}

    # Check if the documentation is installed to display
    # links in bconfig
    if ($self->{info}->{html_dir} && -d "$self->{info}->{html_dir}/docs") {
        $template->param('UseDocs', 1);
    }

    if ($self->{ComponentName}) {
        $template->param('ComponentName', $self->{ComponentName});
    }
    if ($self->{Component}) {
        $template->param('Component', $self->{Component});
    }
    if ($self->{security}) {
        $template->param('security', $self->get_missing_roles());
    }

    $template->param($hash);
    $self->send_content_type();
    print $template->output();
}

# Get binaries location, will use bconsole path first
sub get_location
{
    my ($self, $bin) = @_;

    my $bindir = '/opt/bacula/bin';
    my $script_dir = "/opt/bweb/script";
    my $bweb_bin_dir = "/opt/bweb/bin";

    my $info = $self->{info} || $self;

    if (!defined($info->{bconsole})) {
        $self->edebug('Bconsole path is not set in Bweb config file.');
    } elsif ($info->{bconsole} =~ m:^(.+?)/bconsole:) {
        $bindir = $1;
    } else {
        $self->edebug('Invalid bconsole path. Path bindir will not be set.');
    }

    if (!defined($info->{html_dir})) {
        $self->edebug('HTML directory path is not set in Bweb config file.');
    } elsif ($info->{html_dir} =~ m:^(.+?)/html:) {
        $script_dir = "$1/script";
        $bweb_bin_dir = "$1/bin";
    } else {
        $self->edebug('Invalid html_dir path. Paths script_dir and bweb_bin_dir will not be set.');
    }

    my @default = ($bindir, $script_dir, $bweb_bin_dir, 
                   '/opt/bacula/bin', '/opt/bacula/scripts', 
                   '/usr/local/bin', '/usr/bin', '/sbin', '/bin',
                   '/usr/sbin', '/opt/bacula/etc', '/opt/bweb/etc');

    foreach my $d (@default) {
        if (defined $d && -r "$d/$bin") {
            return $d;
        }
    }
    if ($self->{debug}) {
        $self->edebug("Unable to find $bin\n");
    }
    return;
}

# Get actually set language (short name, for example 'en' or 'fr')
# If no lang set then is returned default language name.
sub get_current_lang
{
    my $self = shift;
    my $info = $self->{info} || $self;
    my $lang = $info->{current_lang} || $info->{lang} || $default_lang;
    return $lang;
}

# Get all possible templates locations.
# Returns:
#  arrayref list of template paths
sub get_template_paths
{
    my $self = shift;
    my $dir = $self->{template_dir} || $template_dir;
    my $lang = $self->get_current_lang();

    # Order is important - first subdirectories, next directories.
    # It is due to recognize available templates dir
    my @tpl_paths = ("$dir/$lang/tpl", "$dir/$lang", $dir);

    return \@tpl_paths;
}

# Get current (first available) templates path.
# Returns:
#  string current templates path
sub get_current_templates_path
{
    my $self = shift;
    my $current_tpl_path="./";

    my $tpl_paths = $self->get_template_paths();
    for (my $i = 0; $i < scalar(@{$tpl_paths}); $i++) {
        if ( -d $tpl_paths->[$i]) {
            $current_tpl_path = $tpl_paths->[$i];
            last;
        }
    }
    return $current_tpl_path;
}

# Prepare plugins template variables.
sub setup_plugins_menu
{
    my ($self, $template) = @_;

    # Every class that has defined 'get_support' method is able to provide
    # to template information about supported plugins.
    if ($self->can('get_support')) {
        # Check BWeb plugins support
        if (ref($self) eq "Bweb") {
            # Exchange restore plugin check
            my $sir_support = $self->get_support(qw/sir/);
            if ($sir_support->{sir}->{exchange}) {
                $template->param('ExchangeDefined', 1);
            }
            # VMware restore plugin check
            if ($sir_support->{sir}->{vmware}) {
                $template->param('VMwareDefined', 1);
            }
        }
    }
}

# Set session.
# Currently it is object CGI::Session
sub set_session
{
    my $self = shift;
    if (!defined($cgi_session) || $cgi_session->is_expired || $cgi_session->is_empty) {
        my $cgi = CGI->new();
        $cgi->header(-type  => 'text/html', -charset => 'utf-8');
        my $session = CGI::Session->load(undef, $cgi);
        if (!defined($session) || $session->is_empty) {
            $session = CGI::Session->new(undef, $cgi);
        }
        if (defined($session)) {
            $session->expire('+1h');
        }
        $cgi_session = $session;
    }
}


# Return session object.
# If session exists it returns object, otherwise it returns undef
sub get_session
{
    my $self = shift;

    if (!defined($cgi_session) || $cgi_session->is_expired || $cgi_session->is_empty) {
        # if no session exists, initialize it
        $self->set_session();
    }
    return $cgi_session;
}

# Print all headers required to show content by CGI
sub print_headers
{
    my $self = shift;

    $self->prepare_headers();

    if ($self->{hold_headers}) {
        # Hold printing headers. Usually used when headers are provided
        # in an action method after headers printing.
        return;
    }

    my $session = $self->get_session();
    my $query_type = $self->get_http_query_type();
    if (defined($query_type)) {
        if ($query_type eq 'XMLHttpRequest') {
            # AJAX request, set text content header
            print $session->header(-type => "text/plain");
        } else {
            # sync request with requested-with header
            print $session->header();
        }
    } else {
        # no AJAX, print general session header
        print $session->header();
    }
}

# Various actions to prepare headers to show
# Used before showing headers.
sub prepare_headers
{
    my $self = shift;
    my $hold_headers = 0;

    my $value;
    my @result;
    my %hold_headers_params = $self->get_hold_headers_params();
    foreach my $key (keys %hold_headers_params) {
        $value = CGI::param($key);
        if (defined($value)) {
            if (ref($hold_headers_params{$key}) eq 'ARRAY') {
                @result = grep {$_ eq $value} @{$hold_headers_params{$key}};
                if (scalar(@result) == 1) {
                    $hold_headers = 1;
                    last;
                }
            } else {
                $hold_headers = 1;
                last;
            }
        }
    }

    $self->{hold_headers} = $hold_headers;
}

# Get HTTP parameters that cause not sending default headers.
# There parameters mean that headers are sent somewhere inside
# a specific action implementation.
sub get_hold_headers_params {
    my $self = shift;
    my %params = %Bweb::Gui::hold_headers_params;

    if (defined($self->{hold_headers_params})) {
        %params = (%params, %{$self->{hold_headers_params}});
    }
    return %params;
}

# Get HTTP request type
sub get_http_query_type {
    my $self = shift;

    my $session = $self->get_session();
    my $requested_by = $session->query()->http('X-Requested-With');
    if (!defined($requested_by)) {
        $requested_by = '';
    }
    return $requested_by;
}

sub set_username {
    my $self = shift;
    my $user;
    if ($self->is_system_auth() == 1) {
        $user = $self->get_session()->param('user');
    } else {
        $user = CGI::remote_user();
    }
    $self->{loginname} = $user;
}

# Check if system user auth is enabled
# Returns: 1 - enabled, 0 - disabled
sub is_system_auth() {
    my $self = shift;
    my $is_auth = 0;

    if (defined($self->{info}->{enable_system_auth}) && $self->{info}->{enable_system_auth} eq 'on') {
        $is_auth = 1;
    }
    return $is_auth;
}

# Check if Self Restore User service is enabled
# Returns: 1 - enabled, 0 - disabled
sub is_self_user_restore() {
    my $self = shift;
    my $is_service = 0;

    if (defined($self->{info}->{enable_self_user_restore}) && $self->{info}->{enable_self_user_restore} eq 'on') {
        $is_service = 1;
    }
    return $is_service;
}

# Set values related to self user restore service.
# Currently it is only client list
sub set_self_user_restore() {
    my $self = shift;

    my @clients;
    my $session = $self->get_session();
    if ($self->is_self_user_restore() == 1 && !defined($session->param('user_res'))) {
        @clients = $self->get_clients_per_remote_addr();
    }

    if (!defined($session->param('user_res'))) {
        $session->param('user_res', {clients => \@clients});
    }
}

1;

################################################################

package Bweb::Config;

use base q/Bweb::Gui/;

=head1 PACKAGE
    
    Bweb::Config - read, write, display, modify configuration

=head2 DESCRIPTION

    this package is used for manage configuration

=head2 USAGE

    $conf = new Bweb::Config(config_file => '/path/to/conf');
    $conf->load();

    $conf->edit();

    $conf->save();

=head2 NOTES

    main_conf:
     Contains the original configuration object.

    main_conf->subconf:
     This sub hash contains all defined directors

    current_conf:
     It points to the current subconf item used, it also means that
     we are using a subconfiguration item.

    Unnamed is a special director name that avoid to print error
    when it is not found in the configuration.

=cut

use CGI '-utf8';
$CGI::POST_MAX=102400; # Limit post to 100kB
$CGI::DISABLE_UPLOADS=1; # Disable file uploads

# Make CGI param get in array context silent
# Usually in grep { } is used CGI::param() that results warning in web server log
$CGI::LIST_CONTEXT_WARN = 0;

=head1 PACKAGE VARIABLE

    %k_re - hash of all acceptable option.

=head2 DESCRIPTION

    this variable permit to check all option with a regexp.

=cut

our %k_re = ( dbi      => qr/^(dbi:(Pg|mysql|SQLite):(?:\w+=[\w\d\.\/\-]+;?)+)$/i,
              user     => qr/^([\w\d\.-]+)$/i,
              password => qr/^(.*)$/,
              fv_write_path => qr!^([/\w\d\.-]*)$!,
              template_dir => qr!^([/\w\d\.-]+)$!,
              config_dir => qr!^([/\w\d\.-]+)?$!,
              html_dir => qr!^([/\w\d\.-]+)?$!,
              workset_dir => qr!^([/\w\d\.-]+)?$!,
              ssl_dir => qr!^([/\w\d\.-]+)?$!,
              debug    => qr/^(on)?$/,
              lang     => qr/^(\w\w)?$/,
              email_media => qr/^([\w\d\.-]+@[\d\w\.-]+)$/,
              graph_font  => qr!^([/\w\d\.-]+.ttf)?$!,
              bconsole    => qr!^([a-z0-9_\-/]*/bconsole( [\-a-z0-9\. /]+)*)$!i,
              wiki_url    => qr!(.*)$!,
              stat_job_table => qr!^(\w*)$!,
              display_log_time => qr!^(on|off)?$!,
              enable_security => qr/^(on)?$/,
              enable_security_acl => qr/^(on)?$/,
              enable_system_auth => qr/^(on)?$/,
              enable_self_user_restore => qr/^(on)?$/,
              enable_web_bconsole => qr/^(on)?$/,
              default_age => qr/^((?:\d+(?:[ywdhms]\s*?)?)+)\s*$/,
              default_limit => qr/^(\d+)\s*$/,
              rows_per_page => qr/^(\d+)$/,
              name => qr/^([\w\s\d\.\-]+)$/,
              dir_ver => qr/^(\d+(\.\d+)?)$/,
              size_unit => qr/^(Bi?)$/,
              # can be subitem
              subconf => qr/^$/,
              achs => qr/^$/,
              ach_list => qr/^$/,
              url => qr!^(https?://[\w\.\d/@?;]+)$!,
              );

=head1 FUNCTION

    apply_subconf - overload a conf object  

=head2 DESCRIPTION

    This function permits to handle multiple directors in the same configuration
    file.

    It returns 1 if the supplied argument is a valid subconfiguration item

=cut

sub apply_subconf
{
    my ($self, $dir) = @_ ;
    
    if ($dir && $self->{main_conf}) {
        my $main = $self->{main_conf};
        if (exists $main->{subconf} and defined $main->{subconf}->{$dir}) {
            my $conf = $main->{subconf}->{$dir};
            $self->{current_conf} = $dir;
            # TODO apply only a small set of the configuration
            foreach my $k (keys %$conf) { 
                $self->{$k} = $conf->{$k} if (defined $k_re{$k} and defined $conf->{$k});
            }
        } elsif ($dir ne "Unnamed") { # TODO: find a better way to handle new director
            print STDERR "W: Unable to find sub configuration for $dir in configuration\n";
            return 0;
        }
    }
    return 1;
}

=head1 FUNCTION

    restore_mainconf - restore main conf

=head2 DESCRIPTION

    This function permits to edit the main configuration

=cut

sub restore_mainconf
{
    my ($self) = @_ ;
    if (!defined $self->{main_conf}) {
        return;
    }
    my $c = $self->{main_conf};

    foreach my $k (keys %k_re) {
        delete $self->{$k};
        if (defined $c->{$k}) {
            $self->{$k} = $c->{$k}
        }
    }
    $self->{default_age} ||= '7d';
    delete $self->{current_conf};

    my $current_templates_path = $self->get_current_templates_path();
    $self->{lang_file} = "$current_templates_path/$Bweb::Lang::lang_filename";
}

sub get_current_conf
{
    my ($self) = @_ ;
    my $conf;
    if (defined $self->{current_conf}) {
        $conf = $self->{main_conf}->{subconf}->{$self->{current_conf}};
        $conf->{current_conf} = $self->{current_conf};

    } elsif (defined $self->{main_conf}) {
        $conf = $self->{main_conf}

    } else {
        $conf = $self;
    }
    return $conf;
}

=head1 FUNCTION

    load - load config_file

=head2 DESCRIPTION

    this function load the specified config_file.

=cut

sub load
{
    my ($self) = @_ ;

    unless (open(FP, $self->{config_file}))
    {
        return $self->error("can't load config_file $self->{config_file} : $!");
    }
    my $f=''; my $tmpbuffer;
    while(read FP,$tmpbuffer,4096)
    {
        $f .= $tmpbuffer;
    }
    close(FP);

    my $VAR1;

    no strict; # I have no idea of the contents of the file
    eval "$f" ;
    use strict;

    if ($f and $@) {
        return $self->error("Something is wrong with your configuration file...") ;
    }

    $self->{main_conf} = {subconf => {}};
    # keep a backup of the original config
    foreach my $k (keys %$VAR1) {
        if (exists $k_re{$k} and defined $VAR1->{$k}) {
            $self->{main_conf}->{$k} = $VAR1->{$k};
        }
    }
    $self->restore_mainconf();
    return 1;
}

=head1 FUNCTION

    save - save the current configuration to config_file

=cut

sub save
{
    my ($self) = @_ ;
    my ($main, $conf);

    if (defined $self->{current_conf}) {
        $conf = $self->{main_conf}->{subconf}->{$self->{current_conf}};
        $main = $self->{main_conf};

    } elsif (defined $self->{main_conf}) {
        $main = $conf = $self->{main_conf}

    } else {
        $main = $conf = $self;
    }

    if ($conf->{ach_list}) {
        # shortcut for display_begin
        $conf->{achs} = [ map {{ name => $_ }} 
                          keys %{$conf->{ach_list}}
                        ];
    }

    unless (open(FP, ">$self->{config_file}"))
    {
        return $self->error("$self->{config_file} : $!\n" .
                            "You must add this to your config file\n" 
                            . Data::Dumper::Dumper($main));
    }

    print FP Data::Dumper::Dumper($main);
    
    close(FP);       
    return 1;
}

=head1 FUNCTIONS
    
    edit, view, modify - html form ouput

=cut

sub edit
{
    my ($self) = @_ ;
    my $conf = $self->get_current_conf();
    $self->debug($conf);
    $self->display($conf, "config_edit.tpl");
}

sub view
{
    my ($self) = @_ ;
    my $conf = $self->get_current_conf();
    
    # if we are a subconf, we display default values from main config
    if ($self->{current_conf}) {
        foreach my $k (keys %k_re) {
            if (defined $self->{main_conf}->{$k}) {
                $conf->{"main_$k"} = $self->{main_conf}->{$k};
            }
        }
    }

    # If we have the error parameter, display it
    my $error = CGI::param('error') || '';
    if ($error =~ /^([\w\s\.]+)$/) {
        $conf->{error} = $1;
    }

    $self->display($conf, "config_view.tpl");

    # we cleanup default values from subconf
    if ($self->{current_conf}) {
        foreach my $k (keys %k_re) {
            delete $conf->{"main_$k"};
        }
    }
}

sub del
{
    my ($self) = @_ ;
    my $cur = $self->{current_conf};
    if (!$cur || !$self->{main_conf}->{subconf}) {
        return $self->error("Error while deleting Director");
    }
    delete $self->{main_conf}->{subconf}->{$cur};

    $self->restore_mainconf();
    $self->save();
    $self->view();
}

# apply, create or rename subconf
sub handle_new_conf
{
    my ($self, $new_dir) = @_;
    my $subconf = $self->{main_conf}->{subconf};
    my $cur = $self->{current_conf};

    if ($new_dir && $cur && $new_dir eq $cur) {
        return 1;               # nothing to do

    # Create or edit a subconf from main menu
    } elsif ($new_dir && !$cur) {
        if (!$subconf->{$new_dir}) {
            $subconf->{$new_dir} = {};
        }

        $self->apply_subconf($new_dir);
        return 1;

    # Rename the current director
    } elsif ($new_dir ne $cur) {
        $subconf->{$new_dir} = $subconf->{$cur};
        delete $subconf->{$cur};
        $self->apply_subconf($new_dir);
        return 1;
        
    } elsif (!$new_dir) {
        return $self->error("Can't find new director name");
    }
}

sub modify
{
    my ($self) = @_;
    my $conf = $self->get_current_conf();

    $self->{error} = '';
    # we need to reset checkbox first
    $conf->{debug} = 0;
    $conf->{display_log_time} = 1;
    $conf->{enable_security} = 0;
    $conf->{enable_security_acl} = 0;
    $conf->{enable_system_auth} = 0;
    $conf->{enable_self_user_restore} = 0;
    $conf->{enable_web_bconsole} = 0;

    foreach my $k (CGI::param())
    {
        next unless (exists $k_re{$k}) ;
        my $val = CGI::param($k);

        # waiting for a new form
        if ($k eq 'dbi') {
            $val =~ s/dbi:pg/DBI:Pg/i;
            $val =~ s/dbi:mysql/DBI:mysql/i;
            $val =~ s/dbi:sqlite/DBI:SQLite/i;
        }

        # use parameter only if the format is ok
        if ($val =~ $k_re{$k}) {
            $conf->{$k} = $1;

        # else, try with the default one
        } elsif ($self->{main_conf}->{$k}) { # we take default one
            delete $conf->{$k};

        } else {
            $self->{error} .= "bad parameter : $k = [$val]";
        }
    }
    if ($self->{current_conf}) {
        $self->apply_subconf($self->{current_conf});
    } else {
        $self->restore_mainconf();
    }
    #$self->debug($conf);
    $self->view();

    if ($self->{error}) {       # an error as occured
        $self->display($self, 'error.tpl');
    } else {
        $self->save();
    }
}

1;

################################################################

package Bweb::Client;

use base q/Bweb::Gui/;

=head1 PACKAGE
    
    Bweb::Client - Bacula FD

=head2 DESCRIPTION

    this package is use to do all Client operations like, parse status etc...

=head2 USAGE

    $client = new Bweb::Client(name => 'zog-fd');
    $client->status();            # do a 'status client=zog-fd'

=cut

=head1 FUNCTION

    display_running_job - Html display of a running job

=head2 DESCRIPTION

    this function is used to display information about a current job

=cut

sub display_running_job
{
    my ($self, $bweb, $jobid, $infos) = @_ ;
    my $status = $self->status($bweb->{info});

    if ($jobid) {
        if ($status->{$jobid}) {
            $status = $status->{$jobid};
            $status->{last_jobbytes} = $infos->{jobbytes};
            $status->{last_jobfiles} = $infos->{jobfiles};
            $status->{corr_jobbytes} = $infos->{corr_jobbytes};
            $status->{corr_jobfiles} = $infos->{corr_jobfiles};
            $status->{jobbytes}=$status->{Bytes}; 

            $status->{jobbytes} =~ s![^\d]!!g
                if ($status->{jobbytes});

            $status->{Bwlimit} =~ s/,//g
                if ($status->{Bwlimit});

            # Quick fix for new status output
            if ($status->{"AveBytes/sec"}) {
                $status->{"Bytes/sec"} = $status->{"AveBytes/sec"};
            }

            if ($status->{jobtype} eq 'R') {
                $status->{restore} = 1;
                if ($status->{Completed}) {
                    $status->{corr_jobfiles}=1;
                    $status->{last_jobfiles}=$status->{Expected};
                    $status->{jobfiles}=$status->{Restored}; 
                }
            } else {
                $status->{jobfiles}=$status->{'Files Examined'} || $status->{'Examined'}; 
                $status->{files}=$status->{'Files'} || $status->{'Backedup'}; 
                $status->{jobfiles} =~ s/,//g;
            }

            if (!defined($status->{tls})) {
                $status->{tls} = 0;
            }

            $status->{tls_info_available} = Bweb::tls_support_check($status->{FdVersion});

            $bweb->debug($status);
            $bweb->display($status, "client_job_status.tpl");

        } else {   # Nothing found on client output, job not yet started or already finished
            print "<script type='text/javascript'>", 
                   "refresh_time=10000; bweb_add_refresh();", 
                  "</script>\n";
        }
    } else {
        for my $id (keys %$status) {
            $bweb->display($status->{$id}, "client_job_status.tpl");
        }
    }
}

=head1 FUNCTION

    $client = new Bweb::Client(name => 'plume-fd');
                               
    $client->status($bweb);

=head2 DESCRIPTION

    dirty hack to parse "status client=xxx-fd"

=head2 INPUT

JobId 5 Job alldrives-test.2012-03-26_08.45.56_21 is running.
    VSS Full Backup Job started: 25-Mar-12 23:45
    Files=693 Bytes=493,385 Bytes/sec=685 Errors=0
    Bwlimit=102,400
    Files: Examined=693 Backed up=693
    Processing file: C:/Program Files/Microsoft/Exchange Server/V14/Logging/
    SDReadSeqNo=5 fd=1272

 Old
   JobId 105 Job Full_plume.2006-06-06_17.22.23 is running.
       Backup Job started: 06-jun-06 17:22
       Files=8,971 Bytes=194,484,132 Bytes/sec=7,480,158
       Files: Examined=10,697 Backed up=245
       Bwlimit=102,400
       Processing file: /home/eric/.openoffice.org2/user/config/standard.sod
       SDReadSeqNo=5 fd=5
   

JobId 5 Job RestoreFiles.2012-03-09_09.54.38_09 is running.
      Restore Job started: 09-Mar-12 00:54
    Files=7 Bytes=0 Bytes/sec=0 Errors=7
    Bwlimit=0
    Files: Restored=8 Expected=32 Completed=25%
    Processing file: c:/test2/

JobId=6
 Job=RestoreFiles.2012-03-26_11.28.31_12
 VSS=0
 Level= 
 JobType=R
 JobStarted=2012-03-26 11:28:33
 Files=1287
 Bytes=111318146
 Bytes/sec=691417
 Errors=0
 Bwlimit=512000
    Files Examined=1,287 Expected Files=2,676 Percent Complete=0
 Processing file=/tmp/regress/tmp/bacula-restores/tmp/regress/build/src/win32/win64_installer/release64/bacula.dll

=head2 OUTPUT

    $VAR1 = { 105 => {
                JobName => Full_plume.2006-06-06_17.22.23,
                JobId => 105,
                Files => 8,971,
                Bytes => 194,484,132,
                ...
              },
              ...
    };

=cut

sub status
{
    my ($self, $conf) = @_ ;

    if (defined $self->{cur_jobs}) {
        return $self->{cur_jobs} ;
    }

    my $arg = {};
    my $b = $self->{bweb}->get_bconsole();
    my $ret = $b->send_cmd("st client=\"$self->{name}\"");
    my @param;
    my $jobid;
    my $fdversion;

    for my $r (split(/\n/, $ret)) {
        chomp($r);
        $r =~ s/(^\s+|\s+$)//g;
        $r =~ s/^\s*Files: //g;
        $r =~ s/Backed up=/Backedup=/g;

        if (!defined($fdversion) && $r =~ /\s+Version\:\s+([\d\.a-z]+)\s+/) {
            $fdversion = $1;
        }

        if ($r =~ /JobId (\d+) Job (\S+)/) {
            if ($jobid) {
                $arg->{$jobid} = { @param, JobId => $jobid } ;
            }

            $jobid = $1;
            @param = ( JobName => $2 );
        } elsif ($r =~ /^Director connected using TLS.+/) {
            push @param, "tls", 1;

        } elsif ($r =~ /(\w)\w+ Job started/) {
            push @param, "jobtype", $1;

        } elsif ($r =~ /=.+=/) {
            push @param, split(/\s+|\s*=\s*/, $r) ;

        } elsif ($r =~ /=/) {   # one per line
            push @param, split(/\s*=\s*/, $r) ;

        } elsif ($r =~ /:/) {   # one per line
            push @param, split(/\s*:\s*/, $r, 2) ;
        }
    }

    if ($jobid and @param) {
        $arg->{$jobid} = { @param,
                           JobId => $jobid, 
                           Client => $self->{name},
                           FdVersion => $fdversion
                       } ;
    }

    $self->{cur_jobs} = $arg ;

    return $arg;
}
1;

################################################################

package Bweb::Autochanger;

use base q/Bweb::Gui/;

=head1 PACKAGE
    
    Bweb::Autochanger - Object to manage Autochanger

=head2 DESCRIPTION

    this package will parse the mtx output and manage drives.

=head2 USAGE

    $auto = new Bweb::Autochanger(precmd => 'sudo');
    or
    $auto = new Bweb::Autochanger(precmd => 'ssh root@robot');
                                  
    $auto->status();

    $auto->slot_is_full(10);
    $auto->transfer(10, 11);

=cut

sub new
{
    my ($class, %arg) = @_;
    my $self = $class->SUPER::new();

    my %prop = (
        name  => '',    # autochanger name
        label => {},    # where are volume { label1 => 40, label2 => drive0 }
        drive => [],    # drive use [ 'media1', 'empty', ..]
        slot  => [],    # slot use [ undef, 'empty', 'empty', ..] no slot 0
        io    => [],    # io slot number list [ 41, 42, 43...]
        info  => {slot => 0,    # informations (slot, drive, io)
                  io   => 0,
                  drive=> 0,
                 },
        mtxcmd => '/usr/sbin/mtx',
        debug => 0,
        device => '/dev/changer',
        precmd => '',   # ssh command
        bweb => undef,  # link to bacula web object (use for display) 
        %arg
    ) ;

    bless($self, $class);

    map { $self->{lc($_)} = $prop{$_} } keys %prop ;

    return $self;
}

=head1 FUNCTION

    status - parse the output of mtx status

=head2 DESCRIPTION

    this function will launch mtx status and parse the output. it will
    give a perlish view of the autochanger content.

    it uses ssh if the autochanger is on a other host.

=cut

sub status
{
    my ($self) = @_;
    my @out = `$self->{precmd} $self->{mtxcmd} -f $self->{device} status` ;

    # TODO : reset all infos
    $self->{info}->{drive} = 0;
    $self->{info}->{slot}  = 0;
    $self->{info}->{io}    = 0;

    #my @out = `cat /home/eric/travail/brestore/plume/mtx` ;

#
#  Storage Changer /dev/changer:2 Drives, 45 Slots ( 5 Import/Export )
#Data Transfer Element 0:Full (Storage Element 1 Loaded):VolumeTag = 000000
#Data Transfer Element 1:Empty
#      Storage Element 1:Empty
#      Storage Element 2:Full :VolumeTag=000002
#      Storage Element 3:Empty
#      Storage Element 4:Full :VolumeTag=000004
#      Storage Element 5:Full :VolumeTag=000001
#      Storage Element 6:Full :VolumeTag=000003
#      Storage Element 7:Empty
#      Storage Element 41 IMPORT/EXPORT:Empty
#      Storage Element 41 IMPORT/EXPORT:Full :VolumeTag=000002
#

    for my $l (@out) {

        #          Storage Element 7:Empty
        #          Storage Element 2:Full :VolumeTag=000002
        if ($l =~ /Storage Element (\d+):(Empty|Full)(\s+:VolumeTag=([\w\d.-]+))?/){

            if ($2 eq 'Empty') {
                $self->set_empty_slot($1);
            } else {
                $self->set_slot($1, $4);
            }

        } elsif ($l =~ /Data Transfer.+?(\d+):(Full|Empty)(\s+.Storage Element (\d+) Loaded.(:VolumeTag = ([\w\d.-]+))?)?/) {

            if ($2 eq 'Empty') {
                $self->set_empty_drive($1);
            } else {
                $self->set_drive($1, $4, $6);
            }

        } elsif ($l =~ /Storage Element (\d+).+IMPORT\/EXPORT:(Empty|Full)( :VolumeTag=([\d\w.-]+))?/) 
        {
            if ($2 eq 'Empty') {
                $self->set_empty_io($1);
            } else {
                $self->set_io($1, $4);
            }

#       Storage Changer /dev/changer:2 Drives, 30 Slots ( 1 Import/Export )

        } elsif ($l =~ /Storage Changer .+:(\d+) Drives, (\d+) Slots/) {
            $self->{info}->{drive} = $1;
            $self->{info}->{slot} = $2;
            if ($l =~ /(\d+)\s+Import/) {
                $self->{info}->{io} = $1 ;
            } else {
                $self->{info}->{io} = 0;
            }
        } 
    }

    $self->debug($self) ;
}

sub is_slot_loaded
{
    my ($self, $slot) = @_;

    # no barcodes
    if ($self->{slot}->[$slot] eq 'loaded') {
        return 1;
    } 

    my $label = $self->{slot}->[$slot] ;

    return $self->is_media_loaded($label);
}

sub unload
{
    my ($self, $drive, $slot) = @_;

    return 0 if (not defined $drive or $self->{drive}->[$drive] eq 'empty') ;
    return 0 if     ($self->slot_is_full($slot)) ;

    my $out = `$self->{precmd} $self->{mtxcmd} -f $self->{device} unload $slot $drive 2>&1`;
    
    if ($? == 0) {
        my $content = $self->get_slot($slot);
        print "content = $content<br/> $drive => $slot<br/>";
        $self->set_empty_drive($drive);
        $self->set_slot($slot, $content);
        return 1;
    } else {
        $self->{error} = $out;
        return 0;
    }
}

# TODO: load/unload have to use mtx script from bacula
sub load
{
    my ($self, $drive, $slot) = @_;

    return 0 if (not defined $drive or $self->{drive}->[$drive] ne 'empty') ;
    return 0 unless ($self->slot_is_full($slot)) ;

    print "Loading drive $drive with slot $slot<br/>\n";
    my $out = `$self->{precmd} $self->{mtxcmd} -f $self->{device} load $slot $drive 2>&1`;
    
    if ($? == 0) {
        my $content = $self->get_slot($slot);
        print "content = $content<br/> $slot => $drive<br/>";
        $self->set_drive($drive, $slot, $content);
        return 1;
    } else {
        $self->{error} = $out;
        print $out;
        return 0;
    }
}

sub is_media_loaded
{
    my ($self, $media) = @_;

    unless ($self->{label}->{$media}) {
        return 0;
    }

    if ($self->{label}->{$media} =~ /drive\d+/) {
        return 1;
    }

    return 0;
}

sub have_io
{
    my ($self) = @_;
    return (defined $self->{info}->{io} and $self->{info}->{io} > 0);
}

sub set_io
{
    my ($self, $slot, $tag) = @_;
    $self->{slot}->[$slot] = $tag || 'full';
    push @{ $self->{io} }, $slot;

    if ($tag) {
        $self->{label}->{$tag} = $slot;
    } 
}

# Check if port is IMPORT/EXPORT type
# Returns: 0 - not I/O port, 1 - I/O port
sub is_io
{
    my ($self, $slot) = @_;

    my @io = grep { $_ == $slot } @{$self->{io}};
    my $io_port = scalar(@io);
    return $io_port;
}

sub set_empty_io
{
    my ($self, $slot) = @_;

    push @{ $self->{io} }, $slot;

    unless ($self->{slot}->[$slot]) {       # can be loaded (parse before) 
        $self->{slot}->[$slot] = 'empty';
    }
}

sub get_slot
{
    my ($self, $slot) = @_;
    return $self->{slot}->[$slot];
}

sub set_slot
{
    my ($self, $slot, $tag) = @_;
    $self->{slot}->[$slot] = $tag || 'full';

    if ($tag) {
        $self->{label}->{$tag} = $slot;
    }
}

sub set_empty_slot
{
    my ($self, $slot) = @_;

    unless ($self->{slot}->[$slot]) {       # can be loaded (parse before) 
        $self->{slot}->[$slot] = 'empty';
    }
}

sub set_empty_drive
{
    my ($self, $drive) = @_;
    $self->{drive}->[$drive] = 'empty';
}

sub set_drive
{
    my ($self, $drive, $slot, $tag) = @_;
    $self->{drive}->[$drive] = $tag || $slot;
    $self->{drive_slot}->[$drive] = $slot;

    $self->{slot}->[$slot] = $tag || 'loaded';

    if ($tag) {
        $self->{label}->{$tag} = "drive$drive";
    }
}

sub slot_is_full
{
    my ($self, $slot) = @_;
    
    # slot don't exists => full
    if (not defined $self->{slot}->[$slot]) {
        return 0 ;
    }

    if ($self->{slot}->[$slot] eq 'empty') {
        return 0;
    }
    return 1;                   # vol, full, loaded
}

sub slot_get_first_free
{
    my ($self) = @_;
    for (my $slot=1; $slot < $self->{info}->{slot}; $slot++) {
        return $slot unless ($self->slot_is_full($slot));
    }
}

sub io_get_first_free
{
    my ($self) = @_;
    
    foreach my $slot (@{ $self->{io} }) {
        return $slot unless ($self->slot_is_full($slot));       
    }
    return 0;
}

sub io_get_first_full
{
    my ($self) = @_;
    my $slot;

    foreach my $io (@{ $self->{io} }) {
        if ($self->slot_is_full($io)) {
            $slot = $io;
            last;
        }
    }
    return $slot;
}

sub get_media_slot
{
    my ($self, $media) = @_;

    return $self->{label}->{$media} ;    
}

sub have_media
{
    my ($self, $media) = @_;

    return defined $self->{label}->{$media} ;    
}

sub send_to_io
{
    my ($self, $slot) = @_;

    unless ($self->slot_is_full($slot)) {
        print "Autochanger $self->{name} slot $slot is empty<br>\n";
        return 1;               # ok
    }

    # first, eject it
    if ($self->is_slot_loaded($slot)) {
        # bconsole->umount
        # self->eject
        print "Autochanger $self->{name} $slot is currently in use<br>\n";
        return 0;
    }

    # autochanger must have I/O
    unless ($self->have_io()) {
        print "Autochanger $self->{name} don't have I/O, you can take media yourself<br>\n";
        return 0;
    }

    my $dst = $self->io_get_first_free();

    unless ($dst) {
        print "Autochanger $self->{name} mailbox is full, you must empty I/O first<br>\n";
        return 0;
    }

    $self->transfer($slot, $dst);
}

sub transfer
{
    my ($self, $src, $dst) = @_ ;
    if ($self->{debug}) {
        print "<pre>$self->{precmd} $self->{mtxcmd} -f $self->{device} transfer $src $dst</pre>\n";
    }
    my $out = `$self->{precmd} $self->{mtxcmd} -f $self->{device} transfer $src $dst 2>&1`;
    
    if ($? == 0) {
        my $content = $self->get_slot($src);
        $self->{slot}->[$src] = 'empty';
        $self->set_slot($dst, $content);
        return 1;
    } else {
        $self->{error} = $out;
        return 0;
    }
}

sub get_drive_name
{
    my ($self, $index) = @_;
    return $self->{drive_name}->[$index];
}

# TODO : do a tapeinfo request to get informations
sub tapeinfo
{
    my ($self) = @_;
}

sub clear_io
{
    my ($self) = @_;

    print "<table><tr>\n";
    for my $slot (@{$self->{io}})
    {
        if ($self->is_slot_loaded($slot)) {
            print "<td></td><td>Slot $slot is currently loaded</td></tr>\n";
            next;
        }

        if ($self->slot_is_full($slot))
        {
            my $free = $self->slot_get_first_free() ;
            print "</tr><tr><td>move slot $slot to $free :</td>";

            if ($free) {
                if ($self->transfer($slot, $free)) {
                    print "<td><img src='/bweb/T.png' alt='ok'></td>\n";
                } else {
                    print "<td><img src='/bweb/E.png' alt='ok' title='$self->{error}'></td>\n";
                }
                
            } else {
                $self->{error} = "<td><img src='/bweb/E.png' alt='ok' title='E : Can t find free slot'></td>\n";
            }
        }
    }
    print "</tr></table>\n";
}

# TODO : this is with mtx status output,
# we can do an other function from bacula view (with StorageId)
sub display_content
{
    my ($self) = @_;
    my $bweb = $self->{bweb};

    # $self->{label} => ('vol1', 'vol2', 'vol3', ..);
    my $media_list = $bweb->dbh_join( keys %{ $self->{label} });

    my $query="
SELECT Media.VolumeName  AS volumename,
       Media.VolStatus   AS volstatus,
       Media.LastWritten AS lastwritten,
       Media.VolBytes    AS volbytes,
       Media.MediaType   AS mediatype,
       Media.Slot        AS slot,
       Media.InChanger   AS inchanger,
       Pool.Name         AS name,
       $bweb->{sql}->{MEDIA_EXPIRE} AS expire
FROM Media 
 INNER JOIN Pool USING (PoolId) 

WHERE Media.VolumeName IN ($media_list)
";

    my $all = $bweb->dbh_selectall_hashref($query, 'volumename') ;

    # TODO : verify slot and bacula slot
    my $param = [];
    my @to_update;

    for (my $slot=1; $slot <= $self->{info}->{slot} ; $slot++) {

        if ($self->slot_is_full($slot)) {

            my $vol = $self->{slot}->[$slot];
            if (defined $all->{$vol}) {    # TODO : autochanger without barcodes 

                my $bslot = $all->{$vol}->{slot} ;
                my $inchanger = $all->{$vol}->{inchanger};

                # if bacula slot or inchanger flag is bad, we display a message
                if ((($bslot != $slot && !$self->is_io($slot)) || ($bslot == 0 && $self->is_io($slot))) or !$inchanger) {
                    push @to_update, $slot;
                }
                
                $all->{$vol}->{realslot} = $slot;

                push @{ $param }, $all->{$vol};

            } else {            # empty or no label
                push @{ $param }, {realslot => $slot,
                                   volstatus => 'Unknown',
                                   volumename => $self->{slot}->[$slot]} ;
            }
        } else {                # empty
            push @{ $param }, {realslot => $slot, volumename => 'empty'} ;
        }
    }

    my $i=0; my $drives = [] ;
    foreach my $d (@{ $self->{drive} }) {
        $drives->[$i] = { index => $i,
                          load  => $self->{drive}->[$i],
                          name  => $self->{drive_name}->[$i],
                      };
        $i++;
    }

    $bweb->display({ Name   => $self->{name},
                     nb_drive => $self->{info}->{drive},
                     nb_io => $self->{info}->{io},
                     Drives => $drives,
                     Slots  => $param,
                     IOFull => $self->io_get_first_full(),
                     Update => scalar(@to_update) },
                   'ach_content.tpl');

}

1;

################################################################

package Bweb::Sched;
use base q/Bweb::Gui/;

=head1 PACKAGE

    Bweb::Sched() - Bweb package that parse show schedule ouput

    new Bweb::Sched(format => '%Y-%m-%d', year => 2008);

=head2 USAGE

 my $b = $bweb->get_bconsole();
 my $s = $b->send_cmd("show schedule");
 my $sched = new Bweb::Sched(begin => '2007-01-01', end => '2007-01-02 12:00');
 $sched->parse_scheds(split(/\r?\n/, $s));


$VAR1 = {
          'event' => [
                       '2007-01-04 04:05',
                       '2007-12-05 04:05',
                       '2007-12-12 04:05',
                       '2007-12-19 04:05',
                       '2007-12-26 04:05'
                     ],
          'level' => 'Differential',
          'pool' => 'Semaine'
        };
=cut

sub new
{
    my ($class, @arg) = @_;
    my $self = $class->SUPER::new(@arg);

    # we compare the current schedule date with begin and end
    # in a float form ex: 20071212.1243 > 20070101
    if ($self->{begin} and $self->{end}) {
        ($self->{fbegin}, $self->{fend}) = ($self->{begin}, $self->{end});
        $self->{fbegin} =~ s/(-|:)//g;  $self->{fbegin} =~ s/ /./;
        $self->{fend}   =~ s/(-|:)//g;  $self->{fend}   =~ s/ /./;
    } 

    bless($self,$class);

    if ($self->{bconsole}) {
        my $sel = $self->{name}?"=\"$self->{name}\"":'';
        my $b = $self->{bconsole};
        my $out = $b->send_cmd("show schedule$sel");
        $self->{show_output}=$out;
        $self->parse_scheds(split(/\r?\n/, $out));
        undef $self->{bconsole}; # useless now
    }

    return $self;
}

# cleanup and add a schedule
sub add_sched
{
    my ($self, $name, $info) = @_;
    # bacula uses dates that start from 0, we start from 1
    foreach (@{$info->{mday}},@{$info->{month}}) { $_++ };

    # get events
    $info->{event} = [ $self->get_events($info, $self->{format}) ];

    foreach my $i (qw/hour mday month wday wom woy mins/) {
        delete $info->{$i};
    }

    push @{$self->{schedules}->{$name}}, $info;
}

# return the name of all schedules
sub list_scheds
{
    my ($self, $name) = @_;
    my %ret;
    return keys %{ $self->{schedules} };
}

# return an array of all schedule
sub get_scheds
{
    my ($self, $sched) = @_;
    return $self->{schedules}->{$sched};
}

# return an ref array of all events
# [ '2007-12-19 04:05', '2007-12-19 04:05' ]
sub get_event
{
    my ($self, $sched) = @_;
    return $sched->{event};
}

# return the pool name
sub get_pool
{
    my ($self, $sched) = @_;
    return $sched->{pool} || '';
}

# return the level name (Incremental, Differential, Full)
sub get_level
{
    my ($self, $sched) = @_;
    return $sched->{level};
}

# parse bacula sched bitmap
sub parse_scheds
{
    my ($self, @output) = @_;

    my $run;
    my $name;
    my $schedule;
    my $elt;

    foreach my $ligne (@output) {
        if ($ligne =~ /Schedule: Name=(.+(?= Enabled=\d)|.+)/i) {
            if ($name and $elt) {
                $elt->{level} = $run;
                $self->add_sched($name, $elt);
                $elt=undef;
            }
            $name = $1;
        }
        elsif ($ligne =~ /Run Level=(.+)/i) {
            if ($name and $elt) {
                $elt->{level} = $run;
                $self->add_sched($name, $elt);
                $elt=undef;
            }
            $run = $1;
        }
        elsif ($ligne =~ /(hour|mday|month|mins)=(.+)/i) {
            # All theses lines have the same format
            
            my ($k,$v) = ($1,$2);
            # we get all values (0 1 4 9)
            $elt->{$k}=[split (/\s/,$v)];
        }
        # we make a bitmap for this kind of data (0 0 1 0 0 0 1) for a quick access
        elsif ($ligne =~ /(wday|wom|woy)=(.+)/i) {
            my ($k,$v) = ($1,$2);
            foreach my $e (split (/\s/,$v)) {
                $elt->{$k}->[$e]=1;
            }
        }
        elsif ($ligne =~ /((?:Incremental|Differential|Full)?Pool): name=(.+?) PoolType/i) {
            $elt->{lc($1)} = $2;
        }
    }

    if ($name and $elt) {
        $elt->{level} = $run;
        $self->add_sched($name, $elt);
    }
}

use Date::Calc qw(:all);

# read bacula schedule bitmap and get $format date string
sub get_events
{
    use integer;
    my ($self, $s,$format) = @_;
    my $year = $self->{year} || ((localtime($Bweb::btime))[5] + 1900);
    $format = $format || '%u-%02u-%02u %02u:%02u';
    my @ret;
    foreach my $m (@{$s->{month}})              # mois de l'annee
    {
        foreach my $md (@{$s->{mday}})  # jour du mois
        {
#           print "  month=$m mday=$md\n";
            # we check if this day exists (31 fev)
            next if (!check_date($year,$m,$md));
#           print "    check_date ok\n";

            my $w = ($md-1)/7; # we use the same thing than bacula
            next if (! $s->{wom}->[$w]);
#           print "      wom ok\n";

            # on recupere le jour de la semaine 
            my $wd = Day_of_Week($year,$m,$md);
            $wd = 0 if ($wd == 7) ; # sunday is 0 on bacula
            next if (! $s->{wday}->[$wd]);
#           print "          wday=$wd ok\n";

            my ($w1,undef) = Week_of_Year($year,$m,$md);
            next if (! $s->{woy}->[$w1-1]); # bacula 0-51
#           print "        woy=$w1 ok\n";

            foreach my $h (@{$s->{hour}}) # hour of the day
            {
                foreach my $min (@{$s->{mins}}) # minute
                {
                    if ($self->{fbegin}) {
                        no integer;
                        my $d = sprintf('%d%0.2d%0.2d.%0.2d%0.2d',
                                        $year,$m,$md,$h,$min);
                        next if ($d < $self->{fbegin} or $d > $self->{fend});
                    }
                    push @ret, sprintf($format, $year,$m,$md,$h,$min);
                }
            }
        }
    }
    return @ret;
}
1;

################################################################

package Bweb;

use base q/Bweb::Gui/;

=head1 PACKAGE

    Bweb - main Bweb package

=head2

    this package is use to compute and display informations

=cut

use DBI;
use POSIX qw/strftime mktime/;
use Expect;
use File::Basename;
use File::Copy qw(copy);
use File::Spec;
use File::Temp qw/tempfile/;
use JSON;

# Which tool for which component
our %JSONTools = (
    Director => 'bdirjson',
    Storage => 'bsdjson',
    Console => 'bbconsjson',
    FileDaemon => 'bfdjson'
);

our $BCONFIG_JS_VERSION=1;
our $BWEB_JS_VERSION=1;
our $VERSION="8.8.3";
our $RELEASE_TYPE="";
our $config_file= '/opt/bweb/etc/bweb.conf';

# Version number from which is introduced info about components encrypted
# connection. The info is shown for each component status command (or .status
# for .api 2 commands)
our $tls_support_version = {release => 8, major => 1, minor => 0};

if ($ENV{BWEBCONF} && -f $ENV{BWEBCONF}) {
    $config_file = $ENV{BWEBCONF};
}

our $cur_id=0;

our %Services = (
    'Director' => 'bacula-dir',
    'Storage' => 'bacula-sd',
    'FileDaemon' => 'bacula-fd'
);

our $restore_single_file_subdir = 'vmware_restore';

# Prefix added to each result file taken by Restore Single File interface
our $restore_filename_prefix = 'restore-';

=head1 VARIABLE

    %sql_func - hash to make query mysql/postgresql compliant

=cut

our %sql_func = ( 
          Pg => { 
              UNIX_TIMESTAMP => '',
              FROM_UNIXTIME => '',
              TO_SEC => " interval '1 second' * ",
              SEC_TO_INT => "SEC_TO_INT",
              SEC_TO_TIME => '',
              MATCH => " ~* ",
              MEDIA_EXPIRE => "date_part('epoch', Media.LastWritten) + Media.VolRetention",
              ENDTIME_SEC => " date_part('epoch', EndTime) ",
              JOB_DURATION => " date_part('epoch', EndTime) -  date_part('epoch', StartTime) ",
              STARTTIME_SEC  => " date_part('epoch', Job.StartTime) ",
              STARTTIME_DAY  => " date_trunc('day', Job.StartTime) ",
              STARTTIME_HOUR => " date_trunc('hour', Job.StartTime) ",
              STARTTIME_MONTH  => " date_trunc('month', Job.StartTime) ",
              STARTTIME_WEEK => " date_trunc('week', Job.StartTime) ",
              STARTTIME_PHOUR=> " date_part('hour', Job.StartTime) ",
              STARTTIME_PDAY => " date_part('day', Job.StartTime) ",
              STARTTIME_PMONTH => " date_part('month', Job.StartTime) ",
              STARTTIME_PWEEK => " date_part('week', Job.StartTime) ",
              DB_SIZE => " SELECT pg_database_size(current_database()) ",
              CAT_POOL_TYPE => " MediaType || '_' || Pool.Name ",
              CONCAT_SEP => "",
              NOW => "NOW()",
              #NOW  => "TIMESTAMP '2010-07-15 00:00:00' "
          },
          mysql => {
              UNIX_TIMESTAMP => 'UNIX_TIMESTAMP',
              FROM_UNIXTIME => 'FROM_UNIXTIME',
              SEC_TO_INT => '',
              TO_SEC => '',
              SEC_TO_TIME => 'SEC_TO_TIME',
              MATCH => " REGEXP ",
              MEDIA_EXPIRE => 'UNIX_TIMESTAMP(Media.LastWritten)+Media.VolRetention',
              ENDTIME_SEC => " UNIX_TIMESTAMP(EndTime) ",
              JOB_DURATION => " UNIX_TIMESTAMP(EndTime) - UNIX_TIMESTAMP(StartTime) ",
              STARTTIME_SEC => " UNIX_TIMESTAMP(Job.StartTime) ",
              STARTTIME_DAY  => " DATE_FORMAT(Job.StartTime, '%Y-%m-%d') ",
              STARTTIME_HOUR => " DATE_FORMAT(Job.StartTime, '%Y-%m-%d %H') ",
              STARTTIME_MONTH => " DATE_FORMAT(Job.StartTime, '%Y-%m') ",
              STARTTIME_WEEK => " DATE_FORMAT(Job.StartTime, '%Y-%v') ",
              STARTTIME_PHOUR=> " DATE_FORMAT(Job.StartTime, '%H') ",
              STARTTIME_PDAY => " DATE_FORMAT(Job.StartTime, '%d') ",
              STARTTIME_PMONTH => " DATE_FORMAT(Job.StartTime, '%m') ",
              STARTTIME_PWEEK => " DATE_FORMAT(Job.StartTime, '%v') ",
              # with mysql < 5, you have to play with the ugly SHOW command
              #DB_SIZE => " SELECT 0 ",
              # works only with mysql 5
              DB_SIZE => " SELECT sum(DATA_LENGTH) FROM INFORMATION_SCHEMA.TABLES ",
              CAT_POOL_TYPE => " CONCAT(MediaType,'_',Pool.Name) ",
              CONCAT_SEP => " SEPARATOR '' ",
              NOW => "NOW()",
          },
          SQLite => {
              UNIX_TIMESTAMP => '',
              FROM_UNIXTIME => '',
              SEC_TO_INT => '',
              TO_SEC => '',
              SEC_TO_TIME => '',
              MATCH => " REGEXP ",
              MEDIA_EXPIRE => "strftime('%s', Media.LastWritten) + Media.VolRetention",
              ENDTIME_SEC => " strftime('%s', EndTime) ",
              STARTTIME_SEC =>  " strftime('%s', Job.StartTime) ",
              JOB_DURATION => " strftime('%s', EndTime) -  strftime('%s', StartTime)",

              STARTTIME_DAY  => " strftime('%Y-%m-%d', Job.StartTime) ",
              STARTTIME_HOUR => " strftime('%Y-%m-%d %H', Job.StartTime) ",
              STARTTIME_MONTH => " strftime('%Y-%m', Job.StartTime) ",
              STARTTIME_WEEK => " strftime('%Y-%W', Job.StartTime) ",
              STARTTIME_PHOUR=> " strftime('%H', Job.StartTime) ",
              STARTTIME_PDAY => " strftime('%d', Job.StartTime) ",
              STARTTIME_PMONTH => " strftime('%m', Job.StartTime) ",
              STARTTIME_PWEEK => " strftime('%W', Job.StartTime) ",
              DB_SIZE => " SELECT 0 ",
              CAT_POOL_TYPE => " MediaType || Pool.Name ",
              CONCAT_SEP => "",
              NOW => "strftime('%Y-%m-%d %H:%M:%S', 'now')",
          },
         );

use Exporter 'import';
our @EXPORT_OK = qw($btime);

#our $btime = 1279144800;
our $btime = time;
sub guess_conf
{
    my ($self, $cfg) = @_;
    my $bin_path = $self->get_location("dbcheck");
    return $self->edebug("Unable to find bacula-dir.conf")
        unless (defined $bin_path && -x "$bin_path/dbcheck");
    
    my $dir_conf = $self->get_location("bacula-dir.conf");
    return $self->edebug("Unable to find bacula-dir.conf")
        unless (defined $dir_conf && -r "$dir_conf/bacula-dir.conf");
        
    my @conf = `$bin_path/dbcheck -B -c $dir_conf/bacula-dir.conf`;
    if ($? != 0) {
        $self->edebug("ERROR: dbcheck returned an error\n@conf");
        return;
    }
        
    my $found=0;
    my %dbi;
    
    foreach my $l (@conf) {
        chomp($l);
        if ($l !~ /=/) {
            $self->edebug("Strange line on dbcheck output $l\n");
            next;
        }
        if ($l =~ /catalog=/) {
            if ($found++ > 0) {
                $self->edebug("Will choose the first catalog, discarding $l\n");
                $self->edebug("To migrate $l, use command line parameters\n");
                last;
            }
        }
        my ($k, $v) = split(/=/, $l);
        $dbi{$k} = $v;
    }
    
    return $self->edebug(["Unable to get dbcheck info", %dbi])
        unless ($dbi{db_type} and $dbi{db_name});

    my $dbi_string;
    if (lc($dbi{db_type}) eq 'postgresql') {
        $dbi_string = "DBI:Pg";

    } elsif (lc($dbi{db_type}) eq 'mysql') {
        $dbi_string = "DBI:mysql";

    } else {
        return $self->edebug("Unsupported dbi string $dbi{db_type}\n");
    }

    # Is present!
    if ($dbi{db_name}) {
        $dbi_string .= ":database=$dbi{db_name}";
    }

    if ($dbi{db_address}) {
        $dbi_string .= ";host=$dbi{db_address}";
    }

    if ($dbi{db_port}) {
        $dbi_string .= ";port=$dbi{db_port}";
    }

    if ($dbi{db_socket}) {
        $dbi_string .= ";host=$dbi{db_socket}";
    }
    
    $cfg->{dbi}      = $dbi_string;
    $cfg->{user}     = $dbi{db_user};
    $cfg->{password} = $dbi{db_password};

    return;
}

sub dbh_is_mysql
{
    my ($self) = @_;
    return $self->{info}->{dbi} =~ /dbi:mysql/i;
}

sub dbh_is_sqlite
{
    my ($self) = @_;
    return $self->{info}->{dbi} =~ /dbi:sqlite/i;
}

sub dbh_is_pg
{
    my ($self) = @_;
    return $self->{info}->{dbi} =~ /dbi:pg/i;
}

sub dbh_disconnect
{
    my ($self) = @_;
    if ($self->{dbh}) {
       $self->{dbh}->disconnect();
       undef $self->{dbh};
    }
}

sub dbh_selectall_arrayref
{
    my ($self, $query) = @_;
    $self->connect_db();
    $self->debug($query);
    return $self->{dbh}->selectall_arrayref($query);
}

sub dbh_join
{
    my ($self, @what) = @_;
    return join(',', $self->dbh_quote(@what)) ;
}

sub dbh_quote
{
    my ($self, @what) = @_;

    $self->connect_db();
    if (wantarray) {
        return map { $self->{dbh}->quote($_) } @what;
    } else {
        return $self->{dbh}->quote($what[0]) ;
    }
}

sub dbh_do
{
    my ($self, $query) = @_ ; 
    $self->connect_db();
    $self->debug($query);
    return $self->{dbh}->do($query);
}

# For sqlite, convert UNIX_TIMESTAMP(a) to strftime('%s', a)
sub dbh_convert
{
    my ($self, $query) = @_ ; 
    if ($self->dbh_is_sqlite()) {
        $query =~ s/UNIX_TIMESTAMP\(([^)]+)\)/strftime('%s', $1)/gs;
    }
    return $query;
}

sub dbh_selectall_hashref
{
    my ($self, $query, $join) = @_;
    
    $self->connect_db();
    $self->debug($query);
    return $self->{dbh}->selectall_hashref($query, $join) ;
}

sub dbh_selectrow_hashref
{
    my ($self, $query) = @_;
    
    $self->connect_db();
    $self->debug($query);
    return $self->{dbh}->selectrow_hashref($query) ;
}

sub dbh_strcat
{
    my ($self, @what) = @_;
    if ($self->dbh_is_mysql()) {
        return 'CONCAT(' . join(',', @what) . ')' ;
    } else {
        return join(' || ', @what);
    }
}

sub dbh_prepare
{
    my ($self, $query) = @_;
    $self->debug($query, up => 1);
    return $self->{dbh}->prepare($query);    
}

# display Mb/Gb/Kb
sub human_size
{
    my @unit = qw(B KB MB GB TB);
    my $val = shift || 0;
    my $i=0;
    my $format = '%i %s';
    while ($val / 1024 > 1) {
        $i++;
        $val /= 1024;
    }
    $format = ($i>0)?'%0.1f %s':'%i %s';
    return sprintf($format, $val, $unit[$i]);
}

sub human_sec_unit
{
    my $val = shift;
    my $orig = $val;
    if ($val =~ /^(?:\d+(?:[ywdhms]\s*)?)+\s*$/) {
        for ($val) {
            s/y/*52w/g;
            s/w/*7d/g;
            s/d/*24h/g;
            s/h/*60m/g;
            s/m/*60s/g;
            s/s(\d)/+$1/g;
            s/s//;
        }
        $val = eval $val;
    }
    return int($val);
}

# display Day, Hour, Year
sub human_sec
{
    use integer;

    my $val = shift;
    $val /= 60;                 # sec -> min

    if ($val / 60 <= 1) {
        return "$val mins";
    } 

    $val /= 60;                 # min -> hour
    if ($val / 24 <= 1) {
        return "$val hours";
    } 

    $val /= 24;                 # hour -> day
    if ($val / 365 < 2) {
        return "$val days";
    } 

    $val /= 365 ;               # day -> year

    return "$val years";   
}

sub human_msec
{
    my $val = shift;
    my $seconds = $val / 1000;
    return human_sec($seconds);
}

# display Enabled
sub human_enabled
{
    my $val = shift || 0;

    if ($val eq '1' or $val eq "yes") {
        return "yes";
    } elsif ($val eq '2' or $val eq "archived") {
        return "archived";
    } else {
        return  "no";
    }
}

# display Enabled
sub from_human_enabled
{
    my $val = shift || 0;

    if ($val eq '1' or $val eq "yes") {
        return 1;
    } elsif ($val eq '2' or $val eq "archived") {
        return 2;
    } else {
        return  0;
    }
}

# get Day, Hour, Year
sub from_human_sec
{
    use integer;

    my $val = shift;
    unless ($val =~ /^\s*(\d+)\s*(\w)\w*\s*$/) {
        return 0;
    }

    my %times = ( m   => 60,
                  h   => 60*60,
                  d   => 60*60*24,
                  m   => 60*60*24*31,
                  y   => 60*60*24*365,
                  );
    my $mult = $times{$2} || 0;

    return $1 * $mult;   
}

# get long term statistic table
sub get_stat_table
{
    my ($self) = @_;
    my $ret = $self->{info}->{stat_job_table} || 'JobHisto';
    if ($ret !~ m/^job$/i) {
        $ret = "(SELECT * FROM Job UNION SELECT * FROM $ret)";
    }
    return $ret;
}

sub connect_db
{
    my ($self) = @_;

    unless ($self->{dbh}) {
        
        if ($self->{info}->{dbi} !~ /Pg|mysql|SQLite/) {
            return $self->error("It looks you have a case problem with your DBI string.\n" . 
                                "$self->{info}->{dbi} doesn't contain Pg, mysql or SQLite");
        }

        $self->{dbh} = DBI->connect($self->{info}->{dbi}, 
                                    $self->{info}->{user},
                                    $self->{info}->{password});

        return $self->error("Can't connect to your database:\n$DBI::errstr\n")
            unless ($self->{dbh});

        $self->{dbh}->{FetchHashKeyName} = 'NAME_lc';

        if ($self->dbh_is_mysql()) {
            $self->{dbh}->do("SET group_concat_max_len=1000000");
        } elsif ($self->dbh_is_pg()) {
            $self->{dbh}->do("SET datestyle TO 'ISO, YMD'");
        } elsif ($self->dbh_is_sqlite()) {
            eval {
                $self->{dbh}->func('REGEXP', 2, sub {
                    my ($regex, $string) = @_;
                    return $string =~ /$regex/;
                                   }, 'create_function');
            }
        }
    }
}

sub new
{
    my ($class, %arg) = @_;
    my $self = $class->SUPER::new();

    my %prop = (
        dbh => undef,           # connect_db();
        info => {
            dbi   => '', # DBI:Pg:database=bacula;host=127.0.0.1
            user  => 'bacula',
            password => 'test', 
        },
        %arg
    );

    bless($self, $class);

    map { $self->{lc($_)} = $prop{$_} } keys %prop ;

    $self->set_username();
    $self->{debug} = $self->{info}->{debug};
    $self->{template_dir} = $self->{info}->{template_dir};

    my $args = $self->get_form('dir', 'lang');
    $self->set_lang($args->{lang});
    $self->set_director($args->{dir});

    return $self;
}

# TODO: remove with Overview panel
sub list_conf
{
    my ($self) = @_;
    foreach my $c (keys %{$self->{info}->{subconf}}) {
        $self->{info}->restore_mainconf();
        $self->set_director($c);
        print "<br>$c<br>";
        $self->display_general();
    }
    $self->clear_director();
}

sub display_begin
{
    my ($self) = @_;

    if ((defined($self->{info}->{enable_security}) && $self->{info}->{enable_security}) || $self->is_system_auth() == 1) {
        $self->get_roles();     # get lang
    }

    my $session = $self->get_session();
    $self->{info}->{loginname} = $self->{loginname};
    $self->{info}->{manage_users} = ((defined($self->{info}->{enable_security}) && $self->{info}->{enable_security}) || $self->is_system_auth()) ? 1 : 0;
    $self->{info}->{ver} = $self->get_ver();

    $self->debug($self->{info});
    if ($self->is_self_user_restore() == 1 && $self->{loginname} ne 'admin') {
        my $user_res = $session->param('user_res');
        $self->{info}->{user_remote_addr} =  $session->query()->remote_addr();
        $self->{info}->{empty_client_list} = scalar(@{$user_res->{clients}}) == 0 ? 1 : 0;
        $self->display($self->{info}, "begin_self_user_restore.tpl");
    } else {
        $self->display($self->{info}, "begin.tpl");
    }
}

sub display_end
{
    my ($self) = @_;
    $self->display($self->{info}, "end.tpl");
}

# Get value unique for version
sub get_ver
{
    my $self = shift;
    return $VERSION;
}

sub display_user_actions
{
    my $self = shift;
    my $res = {};
    my $session = $self->get_session();
    my $user_res = $session->param('user_res');
    if (exists($user_res->{clients}) && scalar(@{$user_res->{clients}}) == 1) {
        $res->{client} = $user_res->{clients}[0];
    }
    $self->display($res, "user_actions.tpl");
}

sub test_conf_dbi
{
    my ($self) = @_;
    my ($status, $msg, $error) = ('', '', '');
    my $dbi = CGI::param('dbi');
    my $user = CGI::param('user');
    my $passwd = CGI::param('password');

    if ($dbi !~ $k_re{dbi}) {
        $status = 'f';
        $msg = 'Bad DBI string format';
    }

    if ($user !~ $k_re{user}) {
        $status = 'f';
        $msg = 'Bad user string format';
    }

    my $dbh;
    eval {
        $dbh = DBI->connect($dbi,
                            $user,
                            $passwd, {RaiseError => 0});
    };
    if (!$dbh) {
        $status = 'f';
        $msg = $DBI::errstr || $msg;
        $msg =~ s/"/'/g;
        $msg =~ s/\n/ /g;
    } else {
        $status = 'T';
        my $row = $dbh->selectrow_hashref("SELECT VersionId FROM Version");
        if (!$row) {
            $status = 'f';
            $msg = "Unable to get VersionId from Bacula schema";

        } else {
            my $row =$dbh->selectrow_hashref("SELECT 1 FROM Client LEFT JOIN client_group_member USING (ClientId) LIMIT 1;");
            if (!$row) {
                $status = 'W';
                my $install_script = $self->get_location('install_bweb.sh');
                if (defined($install_script)) {
                    $install_script .= '/';
                }
                # get translated message
                $msg = $self->{translate}->get('The database access is OK, but you need to load the specific BWeb SQL functions. Execute as root %s to correct the setup.');
                $msg = sprintf($msg, "'${install_script}install_bweb.sh'");
            }
        }
    }

    $self->send_content_type(-type => "text/plain");
    print "{\"status\":\"$status\", \"error\":\"$error\", \"msg\":\"$msg\"}\n";
}

sub test_conf_bconsole
{
    my ($self) = @_;
    my ($status, $msg, $error) = ('T', '', '');
    my $bconsole = CGI::param('bconsole');

    if ($bconsole !~ $k_re{bconsole}) {
        $status = 'f';
        $msg = 'Bad bconsole string format';
    } else {
        my $version = `echo version | $bconsole -u 10 2>&1`;
        if ($? != 0) {
            $status = 'f';
            $msg = $version;
            $msg =~ s/"/'/g;
            $msg =~ s/\n/ /g;
        }
    }

    $self->send_content_type(-type => "text/plain");
    print "{\"status\":\"$status\", \"error\":\"$error\", \"msg\":\"$msg\"}\n";
}

sub client_edit
{
    my ($self) = @_;
    my $arg = $self->get_form("qclient");
    my $f1 = $self->get_client_group_filter();
    my $f2 = $self->get_client_filter();
 
# client_group_name | here 
#-------------------+-----
#      GROUP1       | 1
#      GRP3         | 0

    my $query = "
SELECT client_group_name, max(here) AS here FROM (
  SELECT client_group_name, 1 AS here
    FROM client_group
    JOIN client_group_member USING (client_group_id)
    JOIN Client USING (ClientId) $f2
   WHERE Name = $arg->{qclient}
  UNION ALL
  SELECT client_group_name, 0 
    FROM client_group $f1
) AS temp
GROUP by client_group_name";

    my $all = $self->dbh_selectall_hashref($query, 'client_group_name');
    
    $self->display({ client_group => [ values %$all ], %$arg }, "client_edit.tpl");
}

# Returns hash reference contains clients information like:
# Name, Uname, AutoPrune, FileRetention, JobRetention.
#
# Optional options parameter is hash reference with one key from:
# { qre_client => client_regexp } - takes client name by regular expression
# { client => client_name } - returns one specific here client information
# { jclient_groups => client_groups } - comma separated clients groups
# { qnotingroup => 0/1 } - if set to 1 then only clients without group are listed
#
sub clients_list
{
    my ($self, $options) = @_;

    my $where = "";

    if (exists($options->{qre_client})) {
        $where = "WHERE Name $self->{sql}->{MATCH} $options->{qre_client} ";
    } elsif (exists($options->{client})) {
        $where = "WHERE Name = '$options->{client}' ";
    } elsif (exists($options->{jclient_groups})) {
        # $filter could already contains client_group_member 
        $where = "
 JOIN client_group_member USING (ClientId) 
 JOIN client_group USING (client_group_id)
 WHERE client_group_name IN ($options->{jclient_groups}) ";
    } elsif (exists($options->{qnotingroup})) {
        $where =   "
  WHERE NOT EXISTS
   (SELECT 1 FROM client_group_member
     WHERE Client.ClientId = client_group_member.ClientId
   )
";
    }

    my $query = "
SELECT Name   AS name,
       Uname  AS uname,
       AutoPrune AS autoprune,
       FileRetention AS fileretention,
       JobRetention  AS jobretention
FROM Client " . $self->get_client_filter() .
$where ;

    my $clients = $self->dbh_selectall_hashref($query, 'name');
    return $clients;
}

sub display_clients
{
    my ($self) = @_;

    my $arg = $self->get_form("client", "qre_client", 
                              "jclient_groups", "qnotingroup");
    my $all = $self->clients_list($arg);

    my $dsp = { ID => $cur_id++,
                clients => [ values %$all] };

    $self->display($dsp, "client_list.tpl") ;
}

sub get_limit
{
    my ($self, %arg) = @_;

    my $limit = '';
    my $label = '';
    my $sql = $self->{sql};

    if ($arg{since} and $arg{age}) {
        my $since = iso_to_unix($arg{since});
        my $d = strftime('%Y-%m-%d %H:%M:%S', localtime($since + $arg{age}));
        $limit .= "
 AND StartTime > '$arg{since}'  
 AND EndTime < '$d' ";

        $label .= $self->{translate}->get("since") . " $arg{since} " .
        $self->{translate}->get("and during") . " " . human_sec($arg{age});
    } elsif ($arg{age}) {
        my $when = $btime - $arg{age};
        $limit .= "AND JobTDate > $when";

        $label = $self->{translate}->get("last") . " " . human_sec($arg{age});
    }

    if ($arg{groupby}) {
        $limit .= " GROUP BY $arg{groupby} ";
    }

    if ($arg{order}) {
        $limit .= " ORDER BY $arg{order} ";
    }

    if ($arg{limit}) {
        $limit .= " LIMIT $arg{limit} ";
        $label .= " " . $self->{translate}->get("limited to") . " $arg{limit}";
    }

    if ($arg{offset}) {
        $limit .= " OFFSET $arg{offset} ";
        $label .= " " . $self->{translate}->get("with") . " $arg{offset} ".
        $self->{translate}->get("offset");
    }

    unless ($label) {
        $label = $self->{translate}->get("no filter");
    }

    return ($limit, $label);
}

sub get_item
{
    my ($what, $default) = @_;
    my %opt_cookies = ( dir => 1 );

    my $ret = CGI::param($what);

    if ($opt_cookies{$what} && !$ret) {
        $ret = CGI::cookie($what);
    }

    if (!defined($ret) || $ret eq "") {
        $ret = $default;
    }

    return $ret;
    
}

=head1 FUNCTION

    $bweb->get_form(...) - Get useful stuff

=head2 DESCRIPTION

    This function get and check parameters against regexp.
    
    If word begin with 'q', the return will be quoted or join quoted
    if it's end with 's'.
    

=head2 EXAMPLE

    $bweb->get_form('jobid', 'qclient', 'qpools') ;

    { jobid    => 12,
      qclient  => 'plume-fd',
      qpools   => "'plume-fd', 'test-fd', '...'",
    }

=cut

sub get_form
{
    my ($self, @what) = @_;
    my %what = map { $_ => 1 } @what;
    my %ret;

    # Integer list
    my %opt_is = ( jobids => 0,
                   slots  => 0,
        );

    my %opt_i = (
                 files  => 0,
                 limit  => $self->{info}->{default_limit} || 100,
                 cost   =>  10,
                 offset =>   0,
                 start  =>   0,
                 width  => 640,
                 height => 480,
                 jobid  =>   0,
                 slot   =>   0,
                 drive  =>   0,
                 priority => 10,
                 age    => $self->{info}->{default_age},
                 days   => 1,
                 maxvoljobs  => 0,
                 maxvolbytes => 0,
                 maxvolfiles => 0,
                 filenameid => 0,
                 pathid => 0,
                 nb => 0,
                 bwlimit => 0,
                 debuglevel => 0,
                 time => 0,
                 remove_file => 0,
                 detect_system_layout => 0,
                 bytes => 0,
                 );

    my %opt_ss =(               # string with space
                 name    => 1,
                 dir     => 1,
                 device_name => 1,
                 new_dir => 1,
                 job     => 1,
                 verifyjob => 1,
                 storage => 1,
                 pool   => 1,
                 poolrecycle => 1,
                 restorejob  => 1,
                 client => 1,
                 exchange_server => 1,
                 exchange_database => 1,
                 );
    my %opt_s = (               # default to ''
                 ach    => 1,
                 component => 1,
                 status => 1,
                 volstatus => 1,
                 inchanger => 1,
                 level  => 1,
                 media  => 1,
                 ach    => 1,
                 jobtype=> 1,
                 file_access_method => 1,
                 graph  => 1,
                 gtype  => 1,
                 type   => 1,
                 replace => 1,
                 expired => 1,
                 enabled => 1,
                 rolename => 1,
                 storage_cmd => 1,
                 tag => 1,
                 vm_disk => 1,
                 content_id => 1,
                 moref => 1,
                 );

    my %opt_u = ( username => 1 ); # option with username

    my %opt_pwd = ( password => 1 );

    my %opt_p = (               # option with path
                 fileset=> 1,
                 mtxcmd => 1,
                 precmd => 1,
                 device => 1,
                 path => 1,
                 file => 1,
                 );
    my %opt_pl = (paths => 1); # option with paths list
    my %opt_ssl = (jobs => 1); # string with space list option
    my %opt_w = (where => 1);
    my %opt_r = (regexwhere => 1);
    my %opt_d = (               # option with date
                 cacheretention => 1,
                 voluseduration=> 1,
                 volretention => 1,
                );
    my %opt_t = (when => 2,     # option with time 
                 begin => 1,    # 1 hh:min are optionnal
                 end => 1,      # 2 hh:min are required
                 );
    my %opt_ascii = (               # option with ASCII characters set
                 token => 1,
                 bconsole_cmd => 1,
                 );

    foreach my $i (@what) {
        if (exists $opt_i{$i}) {# integer param
            my $value = get_item($i, $opt_i{$i}) ;
            if ($value =~ /^(\d+)$/) {
                $ret{$i} = $1;
            } elsif ($i eq 'age' &&  # can have unit
                     $value =~ /^(?:\d+(?:[ywdhms]\s*)?)+\s*$/) # 2y1h2m34s
            {
                $ret{$i} = human_sec_unit($value);
            }
        } elsif ($opt_s{$i}) {  # simple string param
            my $value = get_item($i, '');
            if ($value =~ /^([\w\d\.-]+)$/) {
                $ret{$i} = $1;
            }
        } elsif ($opt_u{$i}) {  # simple string param
            my $value = CGI::param($i) || '';
            if ($value =~ /^([\w\d\.\@\-]+)$/) {
                $ret{$i} = $1;
            }
        } elsif ($opt_ss{$i}) { # simple string param (with space)
            my $value = get_item($i, '');
            if ($value =~ /^([\w\d\.\-\s]+)$/) {
                $ret{$i} = $1;
            }
        } elsif (exists $opt_is{$i}) {
            my $label = $i;
            $label =~ s/s$//;   # jobids => search for jobid, store in jobids
            $ret{$i} = join(',', grep { /^(\d+)$/ } CGI::param($label));

        } elsif (exists $opt_ssl{$i}) { # string with space list
            my $label = $i;
            $label =~ s/s$//;   # ex. jobs => search for job, store in jobs
            my @value = CGI::param($label);
            $ret{$i} = join(',', grep { /^([\w\d\.\-\s]+)$/ } @value);
        } elsif ($i =~ /^j(\w+)s$/) { # quote join args "'arg1', 'arg2'"
            my @value = grep { ! /^\s*$/ } CGI::param($1) ;
            if (@value) {
                $ret{$i} = $self->dbh_join(@value) ;
            }

        } elsif ($i =~ /^q(\w+[^s])$/) { # 'arg1'
            my $value = CGI::param($1) ;
            if ($value) {
                $ret{$i} = $self->dbh_quote($value);
            }

        } elsif ($i =~ /^q(\w+)s$/) { #[ 'arg1', 'arg2']
            $ret{$i} = [ map { { name => $self->dbh_quote($_) } } 
                                           grep { ! /^\s*$/ } CGI::param($1) ];
        } elsif (exists $opt_p{$i} || exists $opt_pwd{$i}) { # path or password
            my $value = get_item($i, '');
            if ($value =~ /^([\w\d\.\/\s:\@\-]+)$/) {
                $ret{$i} = Encode::encode_utf8($1);
            }
        } elsif (exists $opt_pl{$i}) {
            my @values = CGI::param($i . '[]');
            my @rets;
            for (my $i = 0; $i < scalar(@values); $i++) {
                if ($values[$i] =~ /^([\w\d\.\/\s:\@\-]+)$/) {
                    push @rets, Encode::encode_utf8($values[$i]);
                }
            }
            $ret{$i} = \@rets;
        } elsif (exists $opt_r{$i}) {
            my $value = get_item($i, '');
            if ($value =~ /^([^'"']+)$/) {
                $ret{$i} = Encode::encode_utf8($1);
            }
        } elsif (exists $opt_w{$i}) {
            my $value = get_item($i, '');
            if ($value =~ /^(>?[\w\d\.\/\s:\@\-]+)$/) {
                $ret{$i} = Encode::encode_utf8($1);
            }
        } elsif (exists $opt_d{$i}) {
            my $value = get_item($i, '');
            if ($value =~ /^\s*(\d+\s+\w+)$/) {
                $ret{$i} = $1;
            }
        } elsif (exists $opt_t{$i}) { # 1: hh:min optionnal, 2: hh:min required
            my $when = get_item($i, '');
            if ($when =~ /(\d{4}-\d{2}-\d{2}( \d{2}:\d{2}(:\d{2})?)?)/) {
                if ($opt_t{$i} == 1 or defined $2) {
                    $ret{$i} = $1;
                }
            }
        } elsif ($opt_ascii{$i}) {
            my $value = get_item($i, '');
            if ($value =~ /^([\w\s\d\.\-,;:!?@#+=%\$\*\/]+)$/) {
                $ret{$i} = $1;
            }
        }
    }

    if ($what{comment}) {
        my $s = CGI::param('comment');
        if ($s) {
            $s =~ s/["\\'<>]/ /g; # strip some characters
            $ret{comment}=$s;
        }
    }

    if ($what{storage_cmd} && $ret{storage_cmd}) {
        if (!grep {/^\Q$ret{storage_cmd}\E$/} ('mount', 'umount', 'release','status', 'setdebug')) {
            delete $ret{storage_cmd};
        }
    }

    if ($what{since}) {
        my $age = $ret{age} || human_sec_unit($opt_i{age});
        my $since = CGI::param('since') || strftime('%F %T', localtime($btime - $age));
        if ($since =~ /^(\d{4}-\d{2}-\d{2}( \d{2}:\d{2}:\d{2})?)$/) {
            $ret{since} = $1;
        }
    }

    if ($what{lang}) {
        my $lang = get_item('lang', '');
        if ($lang =~ /^(\w+)$/) {
            $ret{lang} = $1;
        }
    }

    if ($what{db_clients}) {
        my $filter='';
        if ($what{filter}) {
            # get security filter only if asked
            $filter = $self->get_client_filter();
        }

        my $query = "
SELECT Client.Name as clientname
  FROM Client $filter
";

        my $clients = $self->dbh_selectall_hashref($query, 'clientname');
        $ret{db_clients} = [sort {$a->{clientname} cmp $b->{clientname} } 
                              values %$clients] ;
    }

    if ($what{db_client_groups}) {
        my $filter='';
        if ($what{filter}) {
            # get security filter only if asked
            $filter = $self->get_client_group_filter();
        }

        my $query = "
SELECT client_group_name AS name, comment AS comment
  FROM client_group $filter
";
        my $grps = $self->dbh_selectall_hashref($query, 'name');
        $ret{db_client_groups} = [sort {$a->{name} cmp $b->{name} } 
                                  values %$grps] ;
    }

    if ($what{db_usernames}) {
        my $query = "
SELECT username, comment
  FROM bweb_user
";
        my $users = $self->dbh_selectall_hashref($query, 'username');
        $ret{db_usernames} = [sort {$a->{username} cmp $b->{username} } 
                                  values %$users] ;
    }

    if ($what{db_roles}) {
        my $query = "
SELECT rolename, comment
  FROM bweb_role
";
        my $r = $self->dbh_selectall_hashref($query, 'rolename');
        $ret{db_roles} = [sort {$a->{rolename} cmp $b->{rolename} } 
                                  values %$r] ;
    }

    if ($what{db_mediatypes}) {
        my $query = "
SELECT MediaType as mediatype
  FROM MediaType
";
        my $media = $self->dbh_selectall_hashref($query, 'mediatype');
        $ret{db_mediatypes} = [sort {$a->{mediatype} cmp $b->{mediatype} } 
                                  values %$media] ;
    }

    if ($what{db_locations}) {
        my $query = "
SELECT Location as location, Cost as cost 
  FROM Location
";
        my $loc = $self->dbh_selectall_hashref($query, 'location');
        $ret{db_locations} = [ sort { $a->{location} 
                                      cmp 
                                      $b->{location} 
                                  } values %$loc ];
    }

    if ($what{db_pools}) {
        my $query = "SELECT Name as name FROM Pool";

        my $all = $self->dbh_selectall_hashref($query, 'name') ;
        $ret{db_pools} = [ sort { $a->{name} cmp $b->{name} } values %$all ];
    }

    if ($what{db_filesets}) {
        my $query = "
SELECT FileSet.FileSet AS fileset 
  FROM FileSet
";
        my $filesets = $self->dbh_selectall_hashref($query, 'fileset');

        $ret{db_filesets} = [sort {lc($a->{fileset}) cmp lc($b->{fileset}) } 
                               values %$filesets] ;
    }

    if ($what{db_jobnames}) {
        my $filter='';
        if ($what{filter}) {
            $filter = " JOIN Client USING (ClientId) " . $self->get_client_filter();
        }
        my $query = "
SELECT DISTINCT Job.Name AS jobname 
  FROM Job $filter
";
        my $jobnames = $self->dbh_selectall_hashref($query, 'jobname');

        $ret{db_jobnames} = [sort {lc($a->{jobname}) cmp lc($b->{jobname}) } 
                               values %$jobnames] ;
    }

    if ($what{db_devices}) {
        my $query = "
SELECT Device.Name AS name
  FROM Device
";
        my $devices = $self->dbh_selectall_hashref($query, 'name');

        $ret{db_devices} = [sort {lc($a->{name}) cmp lc($b->{name}) } 
                               values %$devices] ;
    }

    return \%ret;
}

sub display_graph
{
    my ($self) = @_;
    $self->can_do('r_view_stat');
    my $fields = $self->get_form(qw/age level status clients filesets 
                                    graph gtype type filter db_clients
                                    limit db_filesets width height
                                    qclients qfilesets qjobnames db_jobnames/);
    
    my $url = CGI::url(-full => 0,
                      -base => 0,
                      -query => 1);
    $url =~ s/^.+?\?//; # http://path/to/bweb.pl?arg => arg

# this organisation is to keep user choice between 2 click
# TODO : fileset and client selection doesn't work

    $self->display({
        url => $url,
        ver => $self->get_ver(),
        %$fields,
    }, "graph.tpl");

    if ($fields->{gtype} and $fields->{gtype} eq 'balloon') {
        system("./bgraph.pl");
    }
}

sub get_selected_media_location
{
    my ($self) = @_ ;

    my $media = $self->get_form('jmedias');

    unless ($media->{jmedias}) {
        return undef;
    }

    my $query = "
SELECT Media.VolumeName AS volumename, Location.Location AS location
FROM Media LEFT JOIN Location ON (Media.LocationId = Location.LocationId)
WHERE Media.VolumeName IN ($media->{jmedias})
";

    my $all = $self->dbh_selectall_hashref($query, 'volumename') ;
  
    # { 'vol1' => { [volumename => 'vol1', location => 'ici'],
    #               ..
    #             }
    # }
    return $all;
}

sub move_media
{
    my ($self, $in) = @_ ;
    $self->can_do('r_media_mgnt');
    my $media = $self->get_selected_media_location();

    unless ($media) {
        return ;
    }

    my $elt = $self->get_form('db_locations');

    $self->display({ ID => $cur_id++,
                     enabled => human_enabled($in),
                     %$elt,     # db_locations
                     media => [ 
            sort { $a->{volumename} cmp $b->{volumename} } values %$media
                               ],
                     },
                   "move_media.tpl");
}

sub help_extern
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');

    my $elt = $self->get_form(qw/db_pools db_mediatypes db_locations/) ;
    $self->debug($elt);
    $self->display($elt, "help_extern.tpl");
}

sub help_extern_compute
{
    my ($self) = @_;
    $self->can_do('r_media_mgnt');

    my $number = CGI::param('limit') || '' ;
    unless ($number =~ /^(\d+)$/) {
        return $self->error("Bad arg number : $number ");
    }

    my ($sql, undef) = $self->get_param('pools', 
                                        'locations', 'mediatypes');

    my $query = "
SELECT Media.VolumeName  AS volumename,
       Media.VolStatus   AS volstatus,
       Media.LastWritten AS lastwritten,
       Media.MediaType   AS mediatype,
       Media.VolMounts   AS volmounts,
       Pool.Name         AS name,
       Media.Recycle     AS recycle,
       $self->{sql}->{MEDIA_EXPIRE} AS expire
FROM Media 
 INNER JOIN Pool     ON (Pool.PoolId = Media.PoolId)
 LEFT  JOIN Location ON (Media.LocationId = Location.LocationId)

WHERE Media.InChanger = 1
  AND Media.VolStatus IN ('Disabled', 'Error', 'Full', 'Used')
  $sql
ORDER BY expire DESC, recycle, Media.VolMounts DESC
LIMIT $number
" ;
    
    my $all = $self->dbh_selectall_hashref($query, 'volumename') ;

    $self->display({ Media => [ values %$all ] },
                   "help_extern_compute.tpl");
}

sub help_intern
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');

    my $param = $self->get_form(qw/db_locations db_pools db_mediatypes/) ;
    $self->display($param, "help_intern.tpl");
}

sub help_intern_compute
{
    my ($self) = @_;
    $self->can_do('r_media_mgnt');

    my $number = CGI::param('limit') || '' ;
    unless ($number =~ /^(\d+)$/) {
        return $self->error("Bad arg number : $number ");
    }

    my ($sql, undef) = $self->get_param('pools', 'locations', 'mediatypes');

    if (CGI::param('expired')) {
        # we take only expired volumes or purged/recycle ones
        $sql = "
AND (
 ( ($self->{sql}->{MEDIA_EXPIRE}) < $btime
 ) OR ( 
  Media.VolStatus IN ('Purged', 'Recycle')
 )
)
 " . $sql ;
    }

    my $query = "
SELECT Media.VolumeName  AS volumename,
       Media.VolStatus   AS volstatus,
       Media.LastWritten AS lastwritten,
       Media.MediaType   AS mediatype,
       Media.VolMounts   AS volmounts,
       Pool.Name         AS name,
       $self->{sql}->{MEDIA_EXPIRE} AS expire
FROM Media 
 INNER JOIN Pool ON (Pool.PoolId = Media.PoolId) 
 LEFT  JOIN Location ON (Location.LocationId = Media.LocationId)

WHERE Media.InChanger <> 1
  AND Media.VolStatus IN ('Purged', 'Full', 'Append', 'Recycle')
  AND Media.Recycle = 1
  $sql
ORDER BY Media.VolUseDuration DESC, Media.VolMounts ASC, expire ASC 
LIMIT $number
" ;
    
    my $all = $self->dbh_selectall_hashref($query, 'volumename') ;

    $self->display({ Media => [ values %$all ] },
                   "help_intern_compute.tpl");

}

sub display_general
{
    my ($self, %arg) = @_ ;

    my ($limit, $label) = $self->get_limit(%arg);
    my $filter = $self->get_client_filter();
    $filter = $filter? " JOIN Client USING (ClientId) $filter " : '';
    my $query = "
SELECT
    (SELECT count(Pool.PoolId)   FROM Pool)   AS nb_pool,
    (SELECT count(Media.MediaId) FROM Media)  AS nb_media,
    (SELECT count(Job.JobId)     FROM Job)    AS nb_job,
    (SELECT sum(VolBytes)        FROM Media)  AS nb_bytes,
    ($self->{sql}->{DB_SIZE})                 AS db_size,
    (SELECT count(Job.JobId)
      FROM Job $filter
      WHERE Job.JobStatus IN ('E','e','f','A')
      $limit
    )                                         AS nb_err,
    (SELECT count(Client.ClientId) FROM Client) AS nb_client
";

    my $row = $self->dbh_selectrow_hashref($query) ;

    $row->{nb_bytes} = human_size($row->{nb_bytes});

    $row->{db_size} = human_size($row->{db_size});
    $row->{label} = $label;
    $row->{age} = $arg{age};
    $row->{current_url} = $self->get_session()->query()->url(-absolute=>1);

    $self->display($row, "general.tpl");
}

sub get_param
{
    my ($self, @what) = @_ ;
    my %elt = map { $_ => 1 } @what;
    my %ret;

    my $limit = '';

    if ($elt{clients}) {
        my @clients = grep { ! /^\s*$/ } CGI::param('client');
        if (@clients) {
            $ret{clients} = \@clients;
            my $str = $self->dbh_join(@clients);
            $limit .= "AND Client.Name IN ($str) ";
        }
    }

    if ($elt{client_groups}) {
        my @clients = grep { ! /^\s*$/ } CGI::param('client_group');
        if (@clients) {
            $ret{client_groups} = \@clients;
            my $str = $self->dbh_join(@clients);
            $limit .= "AND client_group_name IN ($str) ";
        }
    }

    if ($elt{filesets}) {
        my @filesets = grep { ! /^\s*$/ } CGI::param('fileset');
        if (@filesets) {
            $ret{filesets} = \@filesets;
            my $str = $self->dbh_join(@filesets);
            $limit .= "AND FileSet.FileSet IN ($str) ";
        }
    }

    if ($elt{mediatypes}) {
        my @media = grep { ! /^\s*$/ } CGI::param('mediatype');
        if (@media) {
            $ret{mediatypes} = \@media;
            my $str = $self->dbh_join(@media);
            $limit .= "AND Media.MediaType IN ($str) ";
        }
    }

    if ($elt{client}) {
        my $client = CGI::param('client');
        if ($client) {
            $ret{client} = $client;
            $client = $self->dbh_quote($client);
            $limit .= "AND Client.Name = $client ";
        }
    }

    if ($elt{level}) {
        my $level = CGI::param('level') || '';
        if ($level =~ /^(\w)$/) {
            $ret{level} = $1;
            $limit .= "AND Job.Level = '$1' ";
        }
    }

    if ($elt{jobid}) {
        my $jobid = CGI::param('jobid') || '';

        if ($jobid =~ /^(\d+)$/) {
            $ret{jobid} = $1;
            $limit .= "AND Job.JobId = '$1' ";
        }
    }

    if ($elt{status}) {
        my $status = CGI::param('status') || '';
        if ($status =~ /^(\w)$/) {
            $ret{status} = $1;
            if ($1 eq 'f') {
                $limit .= "AND Job.JobStatus IN ('E','e','f','A') ";            
            } elsif ($1 eq 'W') {
                $limit .= "AND ((Job.JobStatus = 'T' AND Job.JobErrors > 0) OR (Job.JobStatus = 'W')) ";
            } else {
                $limit .= "AND Job.JobStatus = '$1' ";          
            }
        }
    }

    if ($elt{volstatus}) {
        my $status = CGI::param('volstatus') || '';
        if ($status =~ /^(\w+)$/) {
            $ret{status} = $1;
            $limit .= "AND Media.VolStatus = '$1' ";            
        }
    }

    if ($elt{locations}) {
        my @location = grep { ! /^\s*$/ } CGI::param('location') ;
        if (@location) {
            $ret{locations} = \@location;           
            my $str = $self->dbh_join(@location);
            $limit .= "AND Location.Location IN ($str) ";
        }
    }

    if ($elt{pools}) {
        my @pool = grep { ! /^\s*$/ } CGI::param('pool') ;
        if (@pool) {
            $ret{pools} = \@pool; 
            my $str = $self->dbh_join(@pool);
            $limit .= "AND Pool.Name IN ($str) ";
        }
    }

    if ($elt{location}) {
        my $location = CGI::param('location') || '';
        if ($location) {
            $ret{location} = $location;
            $location = $self->dbh_quote($location);
            $limit .= "AND Location.Location = $location ";
        }
    }

    if ($elt{pool}) {
        my $pool = CGI::param('pool') || '';
        if ($pool) {
            $ret{pool} = $pool;
            $pool = $self->dbh_quote($pool);
            $limit .= "AND Pool.Name = $pool ";
        }
    }

    if ($elt{jobtype}) {
        my $jobtype = CGI::param('jobtype') || '';
        if ($jobtype =~ /^(\w)$/) {
            $ret{jobtype} = $1;
            $limit .= "AND Job.Type = '$1' ";
        }
    }

    return ($limit, %ret);
}

=head1

    get last backup

=cut 

sub display_job
{
    my ($self, %arg) = @_ ;
    return if $self->cant_do('r_view_job');

    my $args = $self->get_form('job');

    $arg{order} = ' Job.JobId DESC ';

    my ($limit, $label) = $self->get_limit(%arg);
    my ($where, undef) = $self->get_param('clients',
                                          'client_groups',
                                          'level',
                                          'filesets',
                                          'jobtype',
                                          'pools',
                                          'jobid',
                                          'status');
    my $cgq='';
    if (CGI::param('client_group')) {
        $cgq .= "
JOIN client_group_member USING (ClientId)
JOIN client_group USING (client_group_id)
";
    }

    if ($args->{job}) {
       $where .= " AND Job.Name = '$args->{job}' ";
    }

    my $filter = $self->get_client_filter();
    my $comment = $self->get_db_field('Job.Comment');
    my $rb = $self->get_db_field('ReadBytes');
    my $query="
SELECT  Job.JobId       AS jobid,
        Client.Name     AS client,
        FileSet.FileSet AS fileset,
        Job.Name        AS jobname,
        Level           AS level,
        StartTime       AS starttime,
        EndTime         AS endtime,
	SchedTime       AS schedtime,
        Pool.Name       AS poolname,
        JobFiles        AS jobfiles, 
        JobBytes        AS jobbytes,
        JobStatus       AS jobstatus,
        PriorJobId      AS priorjobid,
        HasBase         AS hasbase,
        HasCache        AS hascache,
        Type            AS jobtype,
        Reviewed        AS reviewed,
        PurgedFiles     AS purgedfiles,
        $rb             AS readbytes,
        $comment        AS comment,
        $self->{sql}->{JOB_DURATION} AS duration,
        JobErrors       AS joberrors

 FROM Client $filter $cgq, 
      Job LEFT JOIN Pool     ON (Job.PoolId    = Pool.PoolId)
          LEFT JOIN FileSet  ON (Job.FileSetId = FileSet.FileSetId)
 WHERE Client.ClientId=Job.ClientId
   AND Job.JobStatus NOT IN ('R', 'C')
 $where
 $limit
";

    my $all = $self->dbh_selectall_hashref($query, 'jobid') ;

    $self->display({ Filter => $label,
                     ID => $cur_id++,
                     Jobs => 
                           [ 
                             sort { $a->{jobid} <=>  $b->{jobid} } 
                                        values %$all 
                             ],
                   },
                   "display_job.tpl");
}

# Adapt the code to the Schema version
# TODO: can use the Version field
sub get_db_field
{
    my ($self, $what) = @_ ;

    my %feature = ('Job.Comment' => 4, 'ReadBytes' => 4);
    my %replacement = ('Job.Comment' => "''", 'ReadBytes' => 'JobBytes');

    if (!$self->{info}->{dir_ver} or 
        $self->{info}->{dir_ver} >= $feature{$what})
    {
        return $what;
    } else {
        return $replacement{$what};
    }
}

# display job informations
sub display_job_zoom
{
    my ($self, $jobid) = @_ ;
    $self->can_do('r_view_job');

    $jobid = $self->dbh_quote($jobid);

    # get security filter
    my $filter = $self->get_client_filter();
    my $comment = $self->get_db_field('Job.Comment');
    my $rb = $self->get_db_field('ReadBytes');
    my $query="
SELECT DISTINCT Job.JobId       AS jobid,
                Client.Name     AS client,
                Job.Name        AS jobname,
                FileSet.FileSet AS fileset,
                Level           AS level,
                Pool.Name       AS poolname,
                StartTime       AS starttime,
                JobFiles        AS jobfiles, 
                JobBytes        AS jobbytes,
                JobStatus       AS jobstatus,
                JobErrors       AS joberrors,
                Type            AS jobtype,
                $rb             AS readbytes,
                $comment        AS comment,
                $self->{sql}->{JOB_DURATION} AS duration
 FROM Client $filter,
      Job LEFT JOIN FileSet ON (Job.FileSetId = FileSet.FileSetId)
          LEFT JOIN Pool    ON (Job.PoolId    = Pool.PoolId)
 WHERE Client.ClientId=Job.ClientId
 AND Job.JobId = $jobid
";

    my $row = $self->dbh_selectrow_hashref($query) ;

    # display all volumes associate with this job
    $query="
SELECT Media.VolumeName as volumename
FROM Job,Media,JobMedia
WHERE Job.JobId = $jobid
 AND JobMedia.JobId=Job.JobId 
 AND JobMedia.MediaId=Media.MediaId
";

    my $all = $self->dbh_selectall_hashref($query, 'volumename');

    $row->{hide_bfileview} = $self->dbh_is_sqlite();
    $row->{volumes} = [ values %$all ] ;
    $row->{wiki_url} = $self->{info}->{wiki_url};

    $self->display($row, "display_job_zoom.tpl");
}

sub display_job_group
{
    my ($self, %arg) = @_;
    $self->can_do('r_view_job');

    my ($limit, $label) = $self->get_limit(groupby => 'client_group_name',  %arg);

    my ($where, undef) = $self->get_param('client_groups',
                                          'level',
                                          'pools');
    my $filter = $self->get_client_group_filter();
    my $query = 
"
SELECT client_group.client_group_name AS client_group_name,
       COALESCE(jobok.jobfiles,0)  + COALESCE(joberr.jobfiles,0)  AS jobfiles,
       COALESCE(jobok.jobbytes,0)  + COALESCE(joberr.jobbytes,0)  AS jobbytes,
       COALESCE(jobok.joberrors,0) + COALESCE(joberr.joberrors,0) AS joberrors,
       COALESCE(jobok.nbjobs,0)  AS nbjobok,
       COALESCE(joberr.nbjobs,0) AS nbjoberr,
       COALESCE(jobok.duration, '0') AS duration

FROM client_group $filter LEFT JOIN (
    SELECT client_group_name AS client_group_name, COUNT(1) AS nbjobs, 
           SUM(JobFiles) AS jobfiles, SUM(JobBytes) AS jobbytes, 
           SUM(JobErrors) AS joberrors,
           SUM($self->{sql}->{JOB_DURATION}) AS duration
    FROM Job JOIN client_group_member ON (Job.ClientId = client_group_member.ClientId)
             JOIN client_group USING (client_group_id)
    
    WHERE Type IN ('B', 'R') AND JobStatus IN ('T', 'W')
    $where
    $limit
) AS jobok USING (client_group_name) LEFT JOIN

(
    SELECT client_group_name AS client_group_name, COUNT(1) AS nbjobs, 
           SUM(JobFiles) AS jobfiles, SUM(JobBytes) AS jobbytes, 
           SUM(JobErrors) AS joberrors
    FROM Job JOIN client_group_member ON (Job.ClientId = client_group_member.ClientId)
             JOIN client_group USING (client_group_id)
    
    WHERE Type IN ('B', 'R') AND JobStatus IN ('f','E', 'A')
    $where
    $limit
) AS joberr USING (client_group_name)

    ";

    my $all = $self->dbh_selectall_hashref($query, 'client_group_name');

    my $rep = { groups => [ values %$all ], age => $arg{age}, filter => $label };
                
    $self->debug($rep);
    $self->display($rep, "display_job_group.tpl");
}

sub display_media
{
    my ($self, %arg) = @_ ;
    $self->can_do('r_view_media');

    my ($limit, $label) = $self->get_limit(%arg);    
    my ($where, %elt) = $self->get_param('pools',
                                         'mediatypes',
                                         'volstatus',
                                         'locations');

    my $arg = $self->get_form('jmedias', 'qre_media', 'expired');

    if ($arg->{jmedias}) {
        $where = "AND Media.VolumeName IN ($arg->{jmedias}) $where"; 
    }
    if ($arg->{qre_media}) {
        $where = "AND Media.VolumeName $self->{sql}->{MATCH} $arg->{qre_media} $where"; 
    }
    if ($arg->{expired}) {
        $where = " 
        AND VolStatus IN ('Full', 'Used')
        AND ( $self->{sql}->{MEDIA_EXPIRE} ) < $btime  " . $where ;
    }

    my $query="
SELECT Media.VolumeName  AS volumename, 
       Media.VolBytes    AS volbytes,
       Media.VolStatus   AS volstatus,
       Media.MediaType   AS mediatype,
       Media.InChanger   AS online,
       Media.LastWritten AS lastwritten,
       Location.Location AS location,
       (volbytes*100/COALESCE(media_avg_size.size,-1))  AS volusage,
       Pool.Name         AS poolname,
       $self->{sql}->{MEDIA_EXPIRE} AS expire
FROM      Pool, Media 
LEFT JOIN Location ON (Media.LocationId = Location.LocationId)
LEFT JOIN (SELECT avg(Media.VolBytes) AS size,
                  Media.MediaType     AS MediaType
           FROM Media 
          WHERE Media.VolStatus = 'Full' 
          GROUP BY Media.MediaType
           ) AS media_avg_size ON (Media.MediaType = media_avg_size.MediaType)

WHERE Media.PoolId=Pool.PoolId
$where
$limit
";
    my $all = $self->dbh_selectall_hashref($query, 'volumename') ;

    $self->display({ ID => $cur_id++,
                     Pool => $elt{pool},
                     Location => $elt{location},
                     Media => [ values %$all ],
                   },
                   "display_media.tpl");
}

sub display_allmedia
{
    my ($self) = @_ ;

    my $pool = $self->get_form('db_pools');
    
    foreach my $name (@{ $pool->{db_pools} }) {
        CGI::param('pool', $name->{name});
        $self->display_media();
    }
}

sub display_media_zoom
{
    my ($self) = @_ ;
    $self->can_do('r_view_media');

    my $media = $self->get_form('jmedias');
    
    unless ($media->{jmedias}) {
        return $self->error("Can't get media selection");
    }
    
    my $query="
SELECT InChanger     AS online,
       Media.Enabled AS enabled,
       VolBytes      AS nb_bytes,
       VolABytes     AS nb_abytes,
       VolFiles      AS nb_files,
       VolJobs       AS nb_jobs,
       VolBlocks     AS nb_blocks,
       VolAPadding   AS volapadding,
       VolHoleBytes  AS volholebytes,
       VolHoles      AS volholes,
       (Media.VolBytes+Media.VolAbytes) AS volsize,
       VolumeName    AS volumename,
       VolStatus     AS volstatus,
       VolMounts     AS nb_mounts,
       VolWrites     AS volwrites,
       Media.ActionOnPurge AS actiononpurge,
       Media.VolUseDuration   AS voluseduration,
       Media.MaxVolJobs AS maxvoljobs,
       Media.MaxVolFiles AS maxvolfiles,
       Media.MaxVolBytes AS maxvolbytes,
       VolErrors     AS nb_errors,
       Pool.Name     AS poolname,
       Location.Location AS location,
       Media.Recycle AS recycle,
       Media.VolRetention AS volretention,
       Media.FirstWritten  AS firstwritten,
       Media.LastWritten  AS lastwritten,
       Media.VolReadTime/1000000  AS volreadtime,
       Media.VolWriteTime/1000000 AS volwritetime,
       Media.RecycleCount AS recyclecount,
       Media.Comment      AS comment,
       Media.CacheRetention AS cacheretention,
       Media.VolType AS voltype,
       $self->{sql}->{MEDIA_EXPIRE} AS expire
 FROM Pool,
      Media LEFT JOIN Location ON (Media.LocationId = Location.LocationId)
 WHERE Pool.PoolId = Media.PoolId
 AND VolumeName IN ($media->{jmedias})
";

    my $all = $self->dbh_selectall_hashref($query, 'volumename') ;
    foreach my $media (values %$all) {
        if (($media->{voltype} == 0 && $media->{nb_bytes} == 0) || $media->{voltype} == 14) {
            $media->{cloud} = 1;
        }

        my $mq = $self->dbh_quote($media->{volumename});

        $query = "
SELECT DISTINCT Job.JobId AS jobid,
                Job.Name  AS name,
                Job.StartTime AS starttime,
                Job.Type  AS type,
                Job.Level AS level,
                Job.JobFiles AS files,
                Job.JobBytes AS bytes,
                Job.JobErrors AS joberrors,
                Job.jobstatus AS status
 FROM Media,JobMedia,Job
 WHERE Media.VolumeName=$mq
 AND Media.MediaId=JobMedia.MediaId              
 AND JobMedia.JobId=Job.JobId
";

        my $jobs = $self->dbh_selectall_hashref($query, 'jobid') ;

        $query = "
SELECT LocationLog.Date    AS date,
       Location.Location   AS location,
       LocationLog.Comment AS comment
 FROM Media,LocationLog INNER JOIN Location ON (LocationLog.LocationId = Location.LocationId)
 WHERE Media.MediaId = LocationLog.MediaId
   AND Media.VolumeName = $mq
";

        my $logtxt = '';
        my $log = $self->dbh_selectall_arrayref($query) ;
        if ($log) {
            $logtxt = join("\n", map { ($_->[0] . ' ' . $_->[1] . ' ' . $_->[2])} @$log ) ;
        }

        $self->display({ jobs => [ values %$jobs ],
                         LocationLog => $logtxt,
                         %$media },
                       "display_media_zoom.tpl");
    }
}

# Get single media params by volume name
sub get_media_by_name
{
    my ($self, $volumename) = @_;

    my $media = $self->dbh_quote($volumename);
    my $query="
SELECT Media.VolumeName  AS volumename,
       Media.VolStatus   AS volstatus,
       Media.LastWritten AS lastwritten,
       Media.VolBytes    AS volbytes,
       Media.MediaType   AS mediatype,
       Media.Slot        AS slot,
       Media.InChanger   AS inchanger,
       Pool.Name         AS name
FROM Media
 INNER JOIN Pool USING (PoolId)
 WHERE Media.VolumeName = $media;
";

    my $row = $self->dbh_selectrow_hashref($query);
    return $row;
}

sub location_edit
{
    my ($self) = @_ ;
    $self->can_do('r_location_mgnt');

    my $loc = $self->get_form('qlocation');
    unless ($loc->{qlocation}) {
        return $self->error("Can't get location");
    }

    my $query = "
SELECT Location.Location AS location, 
       Location.Cost   AS cost,
       Location.Enabled AS enabled
FROM Location
WHERE Location.Location = $loc->{qlocation}
";

    my $row = $self->dbh_selectrow_hashref($query);
    $row->{enabled} = human_enabled($row->{enabled});
    $self->display({ ID => $cur_id++,
                     %$row }, "location_edit.tpl") ;
}

sub location_save
{
    my ($self) = @_ ;
    $self->can_do('r_location_mgnt');

    my $arg = $self->get_form(qw/qlocation qnewlocation cost enabled/) ;
    unless ($arg->{qlocation}) {
        return $self->error("Can't get location");
    }    
    unless ($arg->{qnewlocation}) {
        return $self->error("Can't get new location name");
    }
    unless ($arg->{cost}) {
        return $self->error("Can't get new cost");
    }

    my $enabled = from_human_enabled($arg->{enabled});

    my $query = "
UPDATE Location SET Cost     = $arg->{cost}, 
                    Location = $arg->{qnewlocation},
                    Enabled   = $enabled
WHERE Location.Location = $arg->{qlocation}
";

    $self->dbh_do($query);

    $self->location_display();
}

sub location_del
{
    my ($self) = @_ ;
    $self->can_do('r_location_mgnt');

    my $arg = $self->get_form(qw/qlocation/) ;

    unless ($arg->{qlocation}) {
        return $self->error("Can't get location");
    }

    my $query = "
SELECT count(Media.MediaId) AS nb 
  FROM Media INNER JOIN Location USING (LocationID)
WHERE Location = $arg->{qlocation}
";

    my $res = $self->dbh_selectrow_hashref($query);

    if ($res->{nb}) {
        return $self->error("Sorry, the location must be empty");
    }

    $query = "
DELETE FROM Location WHERE Location = $arg->{qlocation}
";

    $self->dbh_do($query);

    $self->location_display();
}

sub location_add
{
    my ($self) = @_ ;
    $self->can_do('r_location_mgnt');

    my $arg = $self->get_form(qw/qlocation cost/) ;

    unless ($arg->{qlocation}) {
        $self->display({}, "location_add.tpl");
        return 1;
    }
    unless ($arg->{cost}) {
        return $self->error("Can't get new cost");
    }

    my $enabled = CGI::param('enabled') || '';
    $enabled = from_human_enabled($enabled);

    my $query = "
INSERT INTO Location (Location, Cost, Enabled) 
       VALUES ($arg->{qlocation}, $arg->{cost}, $enabled)
";

    $self->dbh_do($query);

    $self->location_display();
}

sub location_display
{
    my ($self) = @_ ;

    my $query = "
SELECT Location.Location AS location, 
       Location.Cost     AS cost,
       Location.Enabled  AS enabled,
       (SELECT count(Media.MediaId) 
         FROM Media 
        WHERE Media.LocationId = Location.LocationId
       ) AS volnum
FROM Location
";

    my $location = $self->dbh_selectall_hashref($query, 'location');
    $self->display({ ID => $cur_id++,
                     Locations => [ values %$location ] },
                   "display_location.tpl");
}

sub update_location
{
    my ($self) = @_ ;

    my $media = $self->get_selected_media_location();
    unless ($media) {
        return ;
    }

    my $arg = $self->get_form('db_locations', 'qnewlocation');

    $self->display({ email  => $self->{info}->{email_media},
                     %$arg,
                     media => [ values %$media ],
                   },
                   "update_location.tpl");
}

###########################################################

sub client_save
{
    my ($self) = @_;
    my $arg = $self->get_form(qw/jclient_groups qclient/);

    unless ($arg->{qclient}) {
        return $self->error("Can't get client name");
    }

    $self->can_do('r_group_mgnt');

    my $f1 = $self->get_client_filter();
    my $f2 = $self->get_client_group_filter();

    $self->{dbh}->begin_work();

    my $query = "
DELETE FROM client_group_member 
      WHERE ClientId IN 
           (SELECT ClientId 
              FROM Client $f1
             WHERE Client.Name = $arg->{qclient})
";
    $self->dbh_do($query);

    if ($arg->{jclient_groups}) {
        $query = "
 INSERT INTO client_group_member (client_group_id, ClientId) 
    SELECT client_group_id, (SELECT ClientId
                FROM Client $f1
               WHERE Name = $arg->{qclient})
       FROM client_group $f2 WHERE client_group_name IN ($arg->{jclient_groups})
    
";
        $self->dbh_do($query);
    }

    $self->{dbh}->commit() or $self->error($self->{dbh}->errstr);

    $self->display_clients();
}

sub groups_edit
{
    my ($self) = @_;
    my $grp = $self->get_form(qw/qclient_group db_clients/);

    unless ($grp->{qclient_group}) {
        $self->can_do('r_group_mgnt');
        $self->display({ ID => $cur_id++,
                         client_group => "''",
                         %$grp,
                     }, "groups_edit.tpl");
        return;
    }

    unless ($self->cant_do('r_group_mgnt')) {
        $self->can_do('r_view_group');
    }

    my $query = "
SELECT Name AS name 
  FROM Client JOIN client_group_member using (ClientId)
              JOIN client_group using (client_group_id)
WHERE client_group_name = $grp->{qclient_group}
";

    my $row = $self->dbh_selectall_hashref($query, "name");
    $self->debug($row);
    $self->display({ ID => $cur_id++,
                     client_group => $grp->{qclient_group},
                     %$grp,
                     client_group_member => [ values %$row]}, 
                   "groups_edit.tpl");
}

sub groups_save
{
    my ($self) = @_;
    $self->can_do('r_group_mgnt');

    my $arg = $self->get_form(qw/qclient_group jclients qnewgroup qcomment/);
    if (!$arg->{qcomment}) {
        $arg->{qcomment} = "''";
    }

    if (!$arg->{qclient_group} and $arg->{qnewgroup}) {
        my $query = "
INSERT INTO client_group (client_group_name, comment) 
VALUES ($arg->{qnewgroup}, $arg->{qcomment})
";
        $self->dbh_do($query);
        $arg->{qclient_group} = $arg->{qnewgroup};
    }

    unless ($arg->{qclient_group}) {
        return $self->error("Can't get groups");
    }

    $self->{dbh}->begin_work();

    my $query = "
DELETE FROM client_group_member 
      WHERE client_group_id IN 
           (SELECT client_group_id 
              FROM client_group 
             WHERE client_group_name = $arg->{qclient_group})
";
    $self->dbh_do($query);

    if ($arg->{jclients}) {
        $query = "
    INSERT INTO client_group_member (ClientId, client_group_id) 
       SELECT  ClientId, 
                (SELECT client_group_id 
                   FROM client_group 
                  WHERE client_group_name = $arg->{qclient_group})
          FROM Client WHERE Name IN ($arg->{jclients})
       
";
        $self->dbh_do($query);
    }
    if ($arg->{qclient_group} ne $arg->{qnewgroup}) {
        $query = "
UPDATE client_group 
   SET client_group_name = $arg->{qnewgroup}, comment = $arg->{qcomment}
 WHERE client_group_name = $arg->{qclient_group}
";

        $self->dbh_do($query);
    }

    $self->{dbh}->commit() or $self->error($self->{dbh}->errstr);

    $self->display_groups();
}

sub groups_del
{
    my ($self) = @_;
    $self->can_do('r_group_mgnt');

    my $arg = $self->get_form(qw/qclient_group/);

    unless ($arg->{qclient_group}) {
        return $self->error("Can't get groups");
    }

    $self->{dbh}->begin_work();

    $self->dbh_do("
DELETE FROM client_group_member 
      WHERE client_group_id IN 
           (SELECT client_group_id 
              FROM client_group 
             WHERE client_group_name = $arg->{qclient_group})");

    $self->dbh_do("
DELETE FROM bweb_client_group_acl
      WHERE client_group_id IN
           (SELECT client_group_id 
              FROM client_group 
             WHERE client_group_name = $arg->{qclient_group})");

    $self->dbh_do("
DELETE FROM client_group
      WHERE client_group_name = $arg->{qclient_group}");

    $self->{dbh}->commit();
    $self->display_groups();
}

sub display_groups
{
    my ($self) = @_;
    my $arg;

    if ($self->cant_do('r_group_mgnt')) {
        $arg = $self->get_form(qw/db_client_groups filter/) ;
    } else {
        $arg = $self->get_form(qw/db_client_groups/) ;
    }

    if ($self->{dbh}->errstr) {
        return $self->error("Can't use groups with bweb, read INSTALL to enable them");
    }

    $self->debug($arg);

    $self->display({ ID => $cur_id++,
                     %$arg},
                   "display_groups.tpl");
}

###########################################################

# TODO: Get the role list from the catalog?
sub get_missing_roles
{
    my ($self) = @_;
    my @res;
    foreach my $r (qw/r_user_mgnt r_group_mgnt r_configure r_autochanger_mgnt
                   r_media_mgnt r_location_mgnt r_storage_mgnt r_delete_job
                   r_prune r_purge r_view_job r_view_log r_view_media
                   r_view_stat r_view_group r_view_running_job r_run_job
                   r_stop_job r_cancel_job r_restart_job r_restore_job
                   r_client_status r_client_setbwlimit r_bconsole/)
    {
        if (!$self->{security}->{$r}) {
            push @res, { role => $r };
        }
    }
    return \@res;
}

sub get_roles
{
    my ($self) = @_;
    my $known_in_bweb=0;

    if (not($self->{info}->{enable_security}) && $self->is_system_auth() == 0) {
        return 1;
    }

    unless ($self->{loginname}) {
        $self->set_username();
    }

    if (!$self->{loginname}) {
        $self->error("Can't get your login name, the enable_security option is set in bweb.conf and web server is not configured to handle authentification.");
        $self->edebug("Can't get your login name. To resolve this problem, you can remove enable_security option from bweb.conf or configure httpd.conf to use authentification");
        $self->display_end();
        exit 0;
    }
    # admin is a special user that can do everything
    if ($self->{loginname} eq 'admin') {
        return 1;
    }
    # already fill
    if (defined $self->{security}) {
        return 1;
    }
    $self->{security} = {};
    my $u = $self->dbh_quote($self->{loginname});
           
    my $query = "
 SELECT use_acl, rolename, tpl
  FROM bweb_user 
       JOIN bweb_role_member USING (userid)
       JOIN bweb_role USING (roleid)
 WHERE username = $u
";
    my $rows = $self->dbh_selectall_arrayref($query);
    # do cache with this role
    if ($rows && scalar(@$rows)) {
        $known_in_bweb = 1;
    }

    # Not defined in bweb, and no self user restore
    if (!$known_in_bweb) {
        if ($self->is_self_user_restore()) {
            # When self user restore service is enabled then default add
            # roles required to run restore job
            $self->{security}->{r_restore_job} = 1;
            $self->{security}->{r_view_job} = 1;
            $self->{security}->{r_view_running_job} = 1;
            $self->{security}->{use_acl} = 1;
        } else {
            $self->error("$self->{loginname} is not defined in BWeb");
            $self->display_end();
            exit 0;
        }
    } else {                    # Defined in bweb, so use it
        $self->{security} = {};
        foreach my $r (@$rows) {
            $self->{security}->{$r->[1]}=1;
        }
        $self->{security}->{use_acl} = $rows->[0]->[0];
        if ($rows->[0]->[2] =~ /^(\w+)$/) {
            $self->set_lang($1);
        }
    }
    return 1;
}

sub can_view_client
{
    my ($self, $client) = @_;

    my $filter = $self->get_client_filter();
    if (!$filter) {
        return 1;
    }
    my $cont = $self->dbh_selectrow_hashref("
 SELECT 1
   FROM Client $filter
  WHERE Name = '$client'
");
    return defined $cont;
}

sub cant_do
{
    my ($self, $action) = @_;
    # is security enabled in configuration ?
    if (not($self->{info}->{enable_security}) && $self->is_system_auth() == 0) {
        return 0
    }
    # must be logged
    if (!$self->{loginname}) {
        $self->{error} = "Can't do $action, your are not logged. " .
            "Check security with your administrator";
        return 1;
    }
    # admin is a special user that can do everything
    if ($self->{loginname} eq 'admin') {
        return 0;
    }
    if (!$self->get_roles()) {
        return 0;
    }
    if (!$self->{security}->{$action}) {
        $self->{error} =
            "$self->{loginname} sorry, but this action ($action) " .
            "is not permited. " .
            "Check security with your administrator";
        return 1;
    }
    return 0;
}

# make like an assert (program die)
sub can_do
{
    my ($self, $action) = @_;
    if ($self->cant_do($action)) {
        $self->error($self->{error});
        $self->display_end();
        exit 0;
    }
    return 1;
}

sub use_filter
{
    my ($self) = @_;

    my $session = $self->get_session();
    if ($self->is_self_user_restore() == 1 && $session->param('user') ne 'admin') {
        return 1;
    }

    if ((!$self->{info}->{enable_security} or !$self->{info}->{enable_security_acl}))
    {
        return 0 ;
    }
    
    if ($self->get_roles()) {
        return $self->{security}->{use_acl};
    } else {
        return 1;
    }
}

# JOIN Client USING (ClientId) " . $b->get_client_filter() . "
sub get_client_filter
{
    my ($self, $login) = @_;
    my $u;
    if ($login) {
        $u = $self->dbh_quote($login);
    } elsif ($self->use_filter()) {
        $u = $self->dbh_quote($self->{loginname});
    } else {
        return '';
    }

    if ($self->is_self_user_restore() == 1) {
        my $session = $self->get_session();
        my $user_res = $session->param('user_res');
        if (scalar(@{$user_res->{clients}}) > 0) {
            my @client_list = map { $self->dbh_quote($_); } @{$user_res->{clients}};
            return ' JOIN (SELECT ClientId FROM Client AS ctemp WHERE ctemp.Name IN (' .
                join(',', @client_list) . ')) AS filter USING (ClientId) ';
        } else {
            return ' JOIN (SELECT 0 AS ClientId) AS filter USING (ClientId) ';
        }
    }

    return "
 JOIN (SELECT ClientId FROM client_group_member
   JOIN client_group USING (client_group_id) 
   JOIN bweb_client_group_acl USING (client_group_id) 
   JOIN bweb_user USING (userid)
   WHERE bweb_user.username = $u 
 ) AS filter USING (ClientId)";
}

#JOIN client_group USING (client_group_id)" . $b->get_client_group_filter()
sub get_client_group_filter
{
    my ($self, $login) = @_;
    my $u;
    if ($login) {
        $u = $self->dbh_quote($login);
    } elsif ($self->use_filter()) {
        $u = $self->dbh_quote($self->{loginname});
    } else {
        return '';
    }
    return "
 JOIN (SELECT client_group_id 
         FROM bweb_client_group_acl
         JOIN bweb_user USING (userid)
   WHERE bweb_user.username = $u 
 ) AS filter USING (client_group_id)";
}

# role and username have to be quoted before
# role and username can be a quoted list
sub revoke
{
    my ($self, $role, $username) = @_;
    $self->can_do("r_user_mgnt");
    
    my $nb = $self->dbh_do("
 DELETE FROM bweb_role_member 
       WHERE roleid = (SELECT roleid FROM bweb_role
                        WHERE rolename IN ($role))
         AND userid = (SELECT userid FROM bweb_user
                        WHERE username IN ($username))");
    return $nb;
}

# role and username have to be quoted before
# role and username can be a quoted list
sub grant
{
    my ($self, $role, $username) = @_;
    $self->can_do("r_user_mgnt");

    my $nb = $self->dbh_do("
   INSERT INTO bweb_role_member (roleid, userid)
     SELECT roleid, userid FROM bweb_role, bweb_user 
      WHERE rolename IN ($role)
        AND username IN ($username)
     ");
    return $nb;
}

# role and username have to be quoted before
# role and username can be a quoted list
sub grant_like
{
    my ($self, $copy, $user) = @_;
    $self->can_do("r_user_mgnt");

    my $nb = $self->dbh_do("
  INSERT INTO bweb_role_member (roleid, userid) 
   SELECT roleid, a.userid 
     FROM bweb_user AS a, bweb_role_member 
     JOIN bweb_user USING (userid)
    WHERE bweb_user.username = $copy
      AND a.username = $user");
    return $nb;
}

# username can be a join quoted list of usernames
sub revoke_all
{
    my ($self, $username) = @_;
    $self->can_do("r_user_mgnt");

    $self->dbh_do("
   DELETE FROM bweb_role_member
         WHERE userid IN (
           SELECT userid 
             FROM bweb_user 
            WHERE username in ($username))");
    $self->dbh_do("
DELETE FROM bweb_client_group_acl 
 WHERE userid IN (
  SELECT userid 
    FROM bweb_user 
   WHERE username IN ($username))");
    
}

sub users_del
{
    my ($self) = @_;
    $self->can_do("r_user_mgnt");

    my $arg = $self->get_form(qw/jusernames/);

    unless ($arg->{jusernames}) {
        return $self->error("Can't get user");
    }

    $self->{dbh}->begin_work();
    {
        $self->revoke_all($arg->{jusernames});
        $self->dbh_do("
DELETE FROM bweb_user WHERE username IN ($arg->{jusernames})");
    }
    $self->{dbh}->commit();
    
    $self->display_users();
}

sub users_add
{
    my ($self) = @_;
    $self->can_do("r_user_mgnt");

    # we don't quote username directly to check that it is conform
    my $arg = $self->get_form(qw/username qpasswd qcomment jrolenames qcreate 
                                 lang qcopy_username jclient_groups/) ;

    if (not $arg->{qcreate}) {
        $arg = $self->get_form(qw/db_roles db_usernames db_client_groups/);
        $self->display($arg, "display_user.tpl");
        return 1;
    }

    my $u = $self->dbh_quote($arg->{username});
    
    $arg->{use_acl}=(CGI::param('use_acl')?'true':'false');

    if (!$arg->{qpasswd}) {
        $arg->{qpasswd} = "''";
    }
    if (!$arg->{qcomment}) {
        $arg->{qcomment} = "''";
    }

    # will fail if user already exists
    # UPDATE with mysql dbi does not return if update is ok
    ($self->dbh_do("
  UPDATE bweb_user 
     SET passwd=$arg->{qpasswd}, comment=$arg->{qcomment}, 
         use_acl=$arg->{use_acl}, tpl='$arg->{lang}'
   WHERE username = $u") 
#     and (! $self->dbh_is_mysql() )
     ) and
    $self->dbh_do("
  INSERT INTO bweb_user (username, passwd, use_acl, comment, tpl) 
        VALUES ($u, $arg->{qpasswd}, $arg->{use_acl}, 
                $arg->{qcomment}, '$arg->{lang}')");

    $self->{dbh}->begin_work();
    {
        $self->revoke_all($u);

        if ($arg->{qcopy_username}) {
            $self->grant_like($arg->{qcopy_username}, $u);
        } else {
            $self->grant($arg->{jrolenames}, $u);
        }

        if ($arg->{jclient_groups}) {
            $self->dbh_do("
INSERT INTO bweb_client_group_acl (client_group_id, userid)
 SELECT client_group_id, userid 
   FROM client_group, bweb_user
  WHERE client_group_name IN ($arg->{jclient_groups})
    AND username = $u
");
        }
    }
    $self->{dbh}->commit();

    $self->display_users();
}

# TODO: we miss a matrix with all user/roles
sub display_users
{
    my ($self) = @_;
    $self->can_do("r_user_mgnt");

    my $arg = $self->get_form(qw/db_usernames/) ;

    if ($self->{dbh}->errstr) {
        return $self->error("Can't use users with bweb, read INSTALL to enable them");
    }

    $self->display({ ID => $cur_id++,
                     %$arg},
                   "display_users.tpl");
}

sub display_user
{
    my ($self) = @_;
    $self->can_do("r_user_mgnt");

    my $arg = $self->get_form('username');

    my $userp = $self->get_user($arg->{username});
    if (!$userp) {
        return $self->error("Can't find $arg->{username} in catalog");
    }
    my $filter = $self->get_client_group_filter($arg->{username});
    my $scg = $self->dbh_selectall_hashref("
 SELECT client_group_name AS name 
   FROM client_group $filter
", 'name');

#  rolename  | userid
#------------+--------
# cancel_job |      0
# restore    |      0
# run_job    |      1

    my $user = $self->dbh_quote($arg->{username});
    my $role = $self->dbh_selectall_hashref("
SELECT rolename, max(here) AS userid FROM (
        SELECT rolename, 1 AS here
          FROM bweb_user 
          JOIN bweb_role_member USING (userid)
          JOIN bweb_role USING (roleid)
          WHERE username = $user
       UNION ALL
       SELECT rolename, 0 
         FROM bweb_role
) AS temp
GROUP by rolename", 'rolename');

    $arg = $self->get_form(qw/db_usernames db_client_groups/);    

    $self->display({
        db_usernames => $arg->{db_usernames},
        username => $userp->{username},
        comment => $userp->{comment},
        passwd => $userp->{passwd},
        lang => $userp->{tpl},
        use_acl => $userp->{use_acl},
        db_client_groups => $arg->{db_client_groups},
        client_group => [ values %$scg ],
        db_roles => [ values %$role], 
    }, "display_user.tpl");
}

# Get user record from the Catalog
# Returns: hashref with user data or undef when user doesn't exist
sub get_user {
    my ($self, $username) = @_;
    my $user;
    $username = $self->dbh_quote($username);
    my $result = $self->dbh_selectrow_hashref("
   SELECT userid, username, use_acl, enabled, comment, passwd, tpl
     FROM bweb_user
    WHERE username = $username");
    if ($result) {
        $user = $result;
    }
    return $user;
}

###########################################################

sub get_media_max_size
{
    my ($self, $type) = @_;
    my $query = 
"SELECT avg(VolBytes) AS size
  FROM Media 
 WHERE Media.VolStatus = 'Full' 
   AND Media.MediaType = '$type'
";
    
    my $res = $self->selectrow_hashref($query);

    if ($res) {
        return $res->{size};
    } else {
        return 0;
    }
}

sub update_media
{
    my ($self) = @_ ;

    my $media = $self->get_form('qmedia');

    unless ($media->{qmedia}) {
        return $self->error("Can't get media");
    }

    my $query = "
SELECT Media.Slot         AS slot,
       PoolMedia.Name     AS poolname,
       Media.VolStatus    AS volstatus,
       Media.InChanger    AS inchanger,
       Location.Location  AS location,
       Media.VolumeName   AS volumename,
       Media.MaxVolBytes  AS maxvolbytes,
       Media.MaxVolJobs   AS maxvoljobs,
       Media.MaxVolFiles  AS maxvolfiles,
       Media.VolUseDuration AS voluseduration,
       Media.VolRetention AS volretention,
       Media.CacheRetention AS cacheretention,
       Media.Comment      AS comment,
       PoolRecycle.Name   AS poolrecycle,
       Media.Enabled      AS enabled,
       Media.VolType      AS voltype

FROM Media INNER JOIN Pool AS PoolMedia ON (Media.PoolId = PoolMedia.PoolId)
           LEFT  JOIN Pool AS PoolRecycle ON (Media.RecyclePoolId = PoolRecycle.PoolId)
           LEFT  JOIN Location ON (Media.LocationId = Location.LocationId)

WHERE Media.VolumeName = $media->{qmedia}
";

    my $row = $self->dbh_selectrow_hashref($query);
    $row->{volretention} = human_sec($row->{volretention});

    # Cloud device type only
    if ($row->{voltype} == 14) {
        $row->{cacheretention} = human_sec($row->{cacheretention});
    }
    $row->{voluseduration} = human_sec($row->{voluseduration});
    $row->{enabled} = human_enabled($row->{enabled});

    my $elt = $self->get_form(qw/db_pools db_locations/);

    $self->display({
        %$elt,
        %$row,
    }, "update_media.tpl");
}

sub save_location
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');

    my $arg = $self->get_form('jmedias', 'qnewlocation') ;

    unless ($arg->{jmedias}) {
        return $self->error("Can't get selected media");
    }
    
    unless ($arg->{qnewlocation}) {
        return $self->error("Can't get new location");
    }

    my $query = "
 UPDATE Media 
     SET LocationId = (SELECT LocationId 
                       FROM Location 
                       WHERE Location = $arg->{qnewlocation}) 
     WHERE Media.VolumeName IN ($arg->{jmedias})
";

    my $nb = $self->dbh_do($query);

    print "$nb media updated, you may have to update your autochanger.";

    $self->display_media();
}

sub location_change
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');

    my $media = $self->get_selected_media_location();
    unless ($media) {
        return $self->error("Can't get media selection");
    }
    my $newloc = CGI::param('newlocation');

    my $user = CGI::param('user') || 'unknown';
    my $comm = CGI::param('comment') || '';
    $comm = $self->dbh_quote("$user: $comm");

    my $arg = $self->get_form('enabled');
    my $en = from_human_enabled($arg->{enabled});
    my $b = $self->get_bconsole();

    my $query;
    foreach my $vol (keys %$media) {
        $query = "
INSERT INTO LocationLog (Date,Comment,MediaId,LocationId,NewEnabled,NewVolStatus)
 SELECT $self->{sql}->{NOW}, $comm, Media.MediaId, Location.LocationId, 
        $en, VolStatus 
   FROM Media, Location
  WHERE Media.VolumeName = '$vol'
    AND Location.Location = '$media->{$vol}->{location}'
";
        $self->dbh_do($query);
        $self->debug($query);
        $b->send_cmd("update volume=\"$vol\" enabled=$en");
    }
    $b->close();

    my $q = new CGI;
    $q->param('action', 'update_location');
    my $url = $q->url(-full => 1, -query=>1);

    $self->display({ email  => $self->{info}->{email_media},
                     url => $url,
                     newlocation => $newloc,
                     # [ { volumename => 'vol1' }, { volumename => 'vol2'},..]
                     media => [ values %$media ],
                   },
                   "change_location.tpl");

}

sub _display_group_stats
{
    my ($self, %arg) = @_ ;

    my $carg = $self->get_form(qw/qclient_group/);

    unless ($carg->{qclient_group}) {
        return $self->error("Can't get group");
    }
    my $jobt = $self->get_stat_table();
    my ($limit, $label) = $self->get_limit(%arg);

    my $query = "
SELECT 
    count(Job.JobId)     AS nb_jobs,
    sum(Job.JobBytes)    AS nb_bytes,
    sum(Job.JobErrors)   AS nb_err,
    sum(Job.JobFiles)    AS nb_files,
    client_group.client_group_name  AS clientname
FROM $jobt AS Job
         JOIN Client USING (ClientId) 
         JOIN client_group_member ON (Client.ClientId = client_group_member.ClientId) 
         JOIN client_group USING (client_group_id)
WHERE 
    client_group.client_group_name = $carg->{qclient_group}
    $limit 
GROUP BY client_group.client_group_name
";

    my $row = $self->dbh_selectrow_hashref($query);

    $row->{ID} = $cur_id++;
    $row->{label} = $label;
    $row->{grapharg} = "client_group";

    $self->display($row, "display_client_stats.tpl");
}

# [ name, num, value, joberrors, nb_job ] =>
# {  items => 
#      [ { name => 'ALL',
#          events => [ { num => 1, label => '2007-01', 
#                        value => 'T', title => 10 },
#                      { num => 2, label => '2007-02', 
#                        value => 'R', title => 11 },
#                     ]
#         },
#         { name => 'Other',
#            ...
#       ]
# };
sub make_overview_tab
{
    my ($self, $q) = @_;
    my $ret = $self->dbh_selectall_arrayref($q);
    my @items;
    my $events=[];
    my $cur_name='';
    for my $elt (@$ret) {
        if ($cur_name and $cur_name ne $elt->[0]) { # order by name, num
            push @items, { name => $cur_name, events => $events};
            $events = [];
        }
        $cur_name = $elt->[0];
        push @$events, 
          { num => $elt->[1], status => $elt->[2], 
            joberrors => $elt->[3], title => "$elt->[4] jobs", date => $elt->[5]};
    }
    push @items, { name => $cur_name, events => $events};
    return \@items;
}

sub get_time_overview
{
    my ($self, $arg) = @_; # want since et age from get_form();
    my $type = $arg->{type} || 'day';
    if ($type =~ /^(day|week|hour|month)$/) {
        $type = uc($1);
    } else {
        $arg->{type}='day';
        $type = 'DAY';
    }
    my $jobt = $self->get_stat_table();
    my $stime1 = $self->{sql}->{"STARTTIME_P" . $type}; # get 1, 2, 3, 4
    $stime1 =~ s/Job.StartTime/date/;
    my $stime2 = $self->{sql}->{"STARTTIME_" . $type}; # get 2007-01-03, 2007-01-23

    my ($limit, $label) = $self->get_limit('since' => $arg->{since},
                                           'age' => $arg->{age});
    return ($stime1, $stime2, $limit, $label, $jobt);
}

#              lu ma me je ve sa di
#  groupe1     v  v  x  w  v  v  v    overview
#   |-- s1     v  v  v  v  v  v  v    overview_zoom
#   |-- s2     v  v  x  v  v  v  v
#   `-- s3     v  v  v  w  v  v  v
sub display_overview_zoom
{
    my ($self) = @_;
    $self->can_do('r_view_stat');

    my $arg = $self->get_form(qw/jclient_groups age since type level/);
    $arg->{type} = $arg->{type} || 'day';

    if (!$arg->{jclient_groups}) {
        return $self->error("Can't get client_group selection");
    }
    my ($filter2, undef) = $self->get_param(qw/level jobtype/);
    my ($stime1, $stime2, $limit, $label, $jobt) = $self->get_time_overview($arg);

    my $filter = $self->get_client_filter();
    my $q = "
SELECT name, $stime1 AS num,
       JobStatus AS value, joberrors, nb_job, date
FROM (
  SELECT $stime2        AS date,
         Client.Name    AS name,
         MAX(severity)  AS severity,
         COUNT(1)       AS nb_job,
         SUM(JobErrors) AS joberrors
    FROM $jobt AS Job
    JOIN client_group_member USING (ClientId)
    JOIN client_group        USING (client_group_id)
    JOIN Client              USING (ClientId)  $filter
    JOIN Status              USING (JobStatus)
   WHERE client_group_name IN ($arg->{jclient_groups})
     AND Job.JobStatus IN ('T', 'W', 'f', 'A', 'e', 'E')
         $limit $filter2
   GROUP BY Client.Name, date
) AS sub JOIN Status USING (severity)
 ORDER BY name, date
";
    my $items = $self->make_overview_tab($q);
    $self->display({label => $label,
                    type => $arg->{type},
                    action => "job;since=$arg->{since};level=$arg->{level};type=$arg->{type};age=$arg->{age};client=", 
                    items => $items}, "overview.tpl");
}

sub display_overview
{
    my ($self) = @_ ;
    $self->can_do('r_view_stat');

    my $arg = $self->get_form(qw/jclient_groups age since type level/);
    $arg->{type} = $arg->{type} || 'day';
    my ($filter2, undef) = $self->get_param(qw/client_groups level jobtype/);
    my $filter3 = $self->get_client_group_filter();
    my ($stime1, $stime2, $filter1, $label, $jobt) = $self->get_time_overview($arg);

    my $q = "
SELECT name, $stime1 AS num, 
       Status.JobStatus AS value, joberrors, nb_job, date
FROM (
  SELECT $stime2        AS date, 
         client_group_name AS name,
         MAX(severity)  AS severity,
         COUNT(1)       AS nb_job,
         SUM(JobErrors) AS joberrors
    FROM $jobt AS Job
    JOIN client_group_member USING (ClientId)
    JOIN client_group        USING (client_group_id) $filter3
    JOIN Status              USING (JobStatus)
   WHERE Job.JobStatus IN ('T', 'W', 'f', 'A', 'e', 'E')
       $filter1 $filter2
   GROUP BY client_group_name, date
) AS sub JOIN Status USING (severity)
 ORDER BY name, date
";
    my $items = $self->make_overview_tab($q);
    $self->display({label=>$label,
                    type => $arg->{type},
                    action => "overview_zoom;since=$arg->{since};level=$arg->{level};type=$arg->{type};age=$arg->{age};client_group=", 
                    items => $items}, "overview.tpl");

}

# poolname can be undef
sub display_pool
{
    my ($self, $poolname) = @_ ;
    $self->can_do('r_view_media');

    my $whereA = '';
    my $whereW = '';

    my $arg = $self->get_form('jmediatypes', 'qmediatypes');
    if ($arg->{jmediatypes}) { 
        $whereW = "WHERE MediaType IN ($arg->{jmediatypes}) ";
        $whereA = "AND   MediaType IN ($arg->{jmediatypes}) ";
    }
    
# TODO : afficher les tailles et les dates

    my $query = "
SELECT subq.volmax        AS volmax,
       subq.volnum        AS volnum,
       subq.voltotal      AS voltotal,
       Pool.Name          AS name,
       Pool.Recycle       AS recycle,
       Pool.VolRetention  AS volretention,
       Pool.VolUseDuration AS voluseduration,
       Pool.MaxVolJobs    AS maxvoljobs,
       Pool.MaxVolFiles   AS maxvolfiles,
       Pool.MaxVolBytes   AS maxvolbytes,
       subq.PoolId        AS PoolId,
       subq.MediaType     AS mediatype,
       $self->{sql}->{CAT_POOL_TYPE}  AS uniq
FROM
  (
    SELECT COALESCE(media_avg_size.volavg,0) * count(Media.MediaId) AS volmax,
           count(Media.MediaId)  AS volnum,
           sum(Media.VolBytes)   AS voltotal,
           Media.PoolId          AS PoolId,
           Media.MediaType       AS MediaType
    FROM Media
    LEFT JOIN (SELECT avg(Media.VolBytes) AS volavg,
                      Media.MediaType     AS MediaType
               FROM Media 
              WHERE Media.VolStatus = 'Full' 
              GROUP BY Media.MediaType
               ) AS media_avg_size ON (Media.MediaType = media_avg_size.MediaType)
    GROUP BY Media.MediaType, Media.PoolId, media_avg_size.volavg
  ) AS subq
LEFT JOIN Pool ON (Pool.PoolId = subq.PoolId)
$whereW
";

    my $all = $self->dbh_selectall_hashref($query, 'uniq') ;

    $query = "
SELECT Pool.Name AS name,
       sum(VolBytes) AS size
FROM   Media JOIN Pool ON (Media.PoolId = Pool.PoolId)
WHERE  Media.VolStatus IN ('Recycled', 'Purged')
       $whereA
GROUP BY Pool.Name;
";
    my $empty = $self->dbh_selectall_hashref($query, 'name');

    foreach my $p (values %$all) {
        if ($p->{volmax} > 0) { # mysql returns 0.0000
            # we remove Recycled/Purged media from pool usage
            if (defined $empty->{$p->{name}}) {
                $p->{voltotal} -= $empty->{$p->{name}}->{size};
            }
            $p->{poolusage} = sprintf('%.2f', $p->{voltotal} * 100/ $p->{volmax}) ;
        } else {
            $p->{poolusage} = 0;
        }

        $query = "
  SELECT VolStatus AS volstatus, count(MediaId) AS nb
    FROM Media 
   WHERE PoolId=$p->{poolid}
     AND Media.MediaType = '$p->{mediatype}'
         $whereA
GROUP BY VolStatus
";
        my $content = $self->dbh_selectall_hashref($query, 'volstatus');
        foreach my $t (values %$content) {
            $p->{"nb_" . $t->{volstatus}} = $t->{nb} ;
        }
    }

    $self->debug($all);
    $self->display({ ID => $cur_id++,
                     MediaType => $arg->{qmediatypes}, # [ { name => type1 } , { name => type2 } ]
                     Pools => [ values %$all ]},
                   "display_pool.tpl");
}

# With this function, we get an estimation of next jobfiles/jobbytes count
sub get_estimate_query
{
    my ($self, $mode, $job, $level) = @_;
    # get security filter
    my $filter = $self->get_client_filter();

    my $query;
   
    if ($self->dbh_is_pg()) { # mysql doesn't have statistics functions
        # postgresql have functions that permit to handle lineal regression
        # in y=ax + b
        # REGR_SLOPE(Y,X) = get x
        # REGR_INTERCEPT(Y,X) = get b
        # and we need y when x=now()
        # CORR gives the correlation
        # (TODO: display progress bar only if CORR > 0.8)
        my $now = $btime;
        $query = "
SELECT temp.jobname AS jobname, 
       COALESCE(CORR(jobbytes,jobtdate),0) AS corr_jobbytes,
       ($now*REGR_SLOPE(jobbytes,jobtdate) 
         + REGR_INTERCEPT(jobbytes,jobtdate)) AS jobbytes,
       COUNT(1) AS nb_jobbytes ";

    } else {                    # For MySQL and SQLite
        $query = "
SELECT jobname AS jobname, 
       0.1 AS corr_jobbytes, AVG(jobbytes) AS jobbytes,
       COUNT(1) AS nb_jobbytes ";
    }

    # if it's a differential, we need to compare since the last full
    # 
    #   F D D D F D D D      F I I I I D I I I
    # | #     # #     #    | #         #
    # | #   # # #   # #    | #         #
    # | # # # # # # # #    | # # # # # # # # #
    # +-----------------   +-------------------
    my $filter2='';
    if ($level eq 'D') {
        $filter2 = "
AND Job.StartTime > (
 SELECT StartTime 
   FROM Job 
  WHERE Job.Name = '$job' 
    AND Job.Level = 'F' 
    AND Job.JobStatus IN ('T', 'W') 
ORDER BY Job.StartTime DESC LIMIT 1
) ";
    }
    
    $query .= 
"
FROM (
   SELECT Job.Name AS jobname, 
          JobBytes AS jobbytes,
          JobTDate AS jobtdate
   FROM Job INNER JOIN Client USING (ClientId) $filter
   WHERE Job.Name = '$job'
     AND Job.Level = '$level'
     AND Job.JobStatus IN ('T', 'W')
     $filter2
   ORDER BY StartTime DESC
   LIMIT 4
) AS temp GROUP BY temp.jobname
";
 
    if ($mode eq 'jobfiles') {
        $query =~ s/jobbytes/jobfiles/g;
        $query =~ s/JobBytes/JobFiles/g;
    }
    return $query;
}

# Client bandwidth limit setting.
# Returns hash reference with command and output from command.
sub set_client_bwlimit
{
    my ($self) = @_;
    return if $self->cant_do('r_client_setbwlimit');

    my $arg = $self->get_form('client', 'bwlimit');
    return if (!$self->can_view_client($arg->{client}));

    my $bconsole = $self->get_bconsole();
    my $result = $bconsole->set_client_bwlimit($arg->{client}, $arg->{bwlimit});
    return $result;
}

# Set client bandwidth limit for specific client.
sub client_setbw
{
    my ($self) = @_;

    my $arg = $self->get_form('client');

    my $result = $self->set_client_bwlimit();
    $self->display({
        content => $result->{output},
        title => "Set Bandwidth for Client $arg->{client}",
        name => $result->{cmd},
        id => $cur_id++,
    }, "command.tpl");  
}

# Set debug for component (director|client|storage).
# Use it internal only.
sub set_debug
{
    my ($self, $component, $name, $debuglevel, $tags, $options) = @_;

    my $trace = ($debuglevel > 0) ? 1 : 0;

    # Usually tags will come from forms as quoted list, like '\'a\',\'b\',\'c\''.
    # From this reason here are removed all "'" characters.
    $tags =~ s/\'//g;
    $options =~ s/[\'\,]//g;

    my $bconsole = $self->get_bconsole();
    my $result = $bconsole->set_debug($component, $name, $debuglevel, $trace, $tags, $options);
    return $result;
}

# Set debug for specific client.
sub client_setdebug
{
    my ($self) = @_;

    return if $self->cant_do('r_client_status');

    my $arg_client = $self->get_form('client');
    return if (!$self->can_view_client($arg_client->{client}));

    my $arg = $self->get_form('component', 'debuglevel', 'tags');
    my $result = $self->set_debug('client', $arg_client->{client}, $arg->{debuglevel}, $arg->{tags});

    $self->display({
        content => $result->{output},
        title => "Set debug for Client $arg_client->{client}",
        name => $result->{cmd},
        id => $cur_id++,
    }, "command.tpl");  
}

# Get count of all statuses grouped by jobstatus
sub get_job_status_history
{
    my ($self, $jobname) = @_;

    my ($limit, $label) = $self->get_limit(age => 2592000); # 30 days

    my $query = "
SELECT jobstatus, COUNT(*) as count FROM Job WHERE
  name='$jobname'
  $limit GROUP BY jobstatus
";
    my $row = $self->dbh_selectall_arrayref($query);
    return $row;
}

# Get estimation files and bytes for backup
sub get_estimate_files_bytes
{
    my ($self, $jobname, $level) = @_;

    # for jobfiles, we use only last Full backup. status client= returns
    # all files that have been checked
    my $query1 = $self->get_estimate_query('jobfiles', $jobname, 'F');
    my $query2 = $self->get_estimate_query('jobbytes', $jobname, $level);

    # LEFT JOIN because we always have a previous Full
    my $query = "
SELECT  corr_jobbytes, jobbytes, corr_jobfiles, jobfiles
  FROM ($query1) AS A LEFT JOIN ($query2) AS B USING (jobname)
";
    my $row = $self->dbh_selectrow_hashref($query);

    if (!$row) {
        $row->{jobbytes} = $row->{jobfiles} = 0;
    }

    return $row;
}

sub display_running_job
{
    my ($self, $jobid) = @_;
    return if $self->cant_do('r_view_running_job');

    my $arg = $self->get_form('jobid', 'bwlimit');

    if (!defined($jobid)) {
        $jobid = $arg->{jobid};
    }

    return $self->error("Can't get jobid") unless ($jobid);

    # get security filter
    my $filter = $self->get_client_filter();

    my $query = "
SELECT Client.Name AS name, Job.Name AS jobname, 
       Job.Level AS level, Type AS type, JobStatus AS jobstatus
FROM Job INNER JOIN Client USING (ClientId) $filter
WHERE Job.JobId = $jobid
";

    my $row = $self->dbh_selectrow_hashref($query);

    if ($row) {
        $arg->{client} = $row->{name};
    } else {
        return $self->error("Can't get client");
    }

    my $status = $row->{jobstatus};

    if ($status =~ /[TfAaEWD]/) {
        $self->display_job_zoom($jobid);
        $self->get_job_log($jobid);
        return;
    }

    if ($row->{type} eq 'B') {
        $row = $self->get_estimate_files_bytes($row->{jobname}, $row->{level});
    }
    if ($status =~ /[RBSmMsjlL]/) {
        my $cli = new Bweb::Client(bweb => $self, name => $arg->{client});
        if ($arg->{bwlimit}) {
            my $b = $self->get_bconsole();
            $b->send_cmd("setbandwidth jobid=" . $jobid . " limit=" . $arg->{bwlimit});
        }
        $cli->display_running_job($self, $jobid, $row);
    }

    if ($status =~ /[TfAaEWD]/) {
        $self->display_job_zoom($jobid);
        $self->get_job_log($jobid);
        return;
    }

    if ($status eq 'C') {
        $self->display_running_job_zoom($jobid, $status);
        $self->get_job_log($jobid);
        return;
    }

    if ($jobid) {
        $self->get_job_log($jobid);
    }
}

sub display_running_jobs
{
    my ($self, $display_action) = @_;
    return if $self->cant_do('r_view_running_job');

    my $arg = $self->get_form('client');
    my $client = "";
    if ($arg->{client}) {
        $client = " AND Client.Name = \'$arg->{client}\' ";
    }

    # get security filter
    my $filter = $self->get_client_filter();

    my $query = "
SELECT Job.JobId AS jobid, 
       Job.Name  AS jobname,
       Job.Level     AS level,
       Job.StartTime AS starttime,
       Job.JobFiles  AS jobfiles,
       Job.JobBytes  AS jobbytes,
       Job.JobStatus AS jobstatus,
       $btime - Job.JobTDate AS duration,
       Client.Name AS clientname
FROM Job INNER JOIN Client USING (ClientId) $filter
WHERE 
  JobStatus IN ('C','R','B','e','F','S','m','M','s','j','c','d','t','p')
  $client
";      
    my $all = $self->dbh_selectall_hashref($query, 'jobid') ;
    
    $self->display({ ID => $cur_id++,
                     display_action => $display_action,
                     Jobs => [ values %$all ]},
                   "running_job.tpl") ;
}

sub display_group_stats
{
    my ($self) = @_;
    my $arg = $self->get_form('age', 'since');
    return if $self->cant_do('r_view_stat');

    my $filter = $self->get_client_group_filter();

    my $jobt = $self->get_stat_table();

    my ($limit, $label) = $self->get_limit(%$arg);
    my ($where, %param) = $self->get_param('client_groups', 'level');

    my $query = "
SELECT T1.client_group_name AS name, nb_byte, nb_file, nb_job, nb_err, nb_cancelled, nb_failed, nb_resto
  FROM (

    SELECT sum(JobBytes) AS nb_byte,
           sum(JobFiles) AS nb_file,
           count(1) AS nb_job, client_group_name 
      FROM $jobt AS Job JOIN client_group_member USING (ClientId) 
      JOIN client_group USING (client_group_id) $filter
     WHERE JobStatus IN ('T', 'W') AND Type IN ('M', 'B', 'g') 
           $where $limit
    GROUP BY client_group_name ORDER BY client_group_name

  ) AS T1 LEFT JOIN (

    SELECT count(1) AS nb_err, client_group_name 
      FROM $jobt AS Job JOIN client_group_member USING (ClientId) 
      JOIN client_group USING (client_group_id)
     WHERE JobStatus IN ('E','e','f','A') AND Type = 'B'
           $where $limit
    GROUP BY client_group_name ORDER BY client_group_name

  ) AS T3 USING (client_group_name) LEFT JOIN (

    SELECT count(1) AS nb_resto, client_group_name 
      FROM $jobt AS Job JOIN client_group_member USING (ClientId) 
      JOIN client_group USING (client_group_id)
     WHERE JobStatus IN ('T','W') AND Type = 'R'
           $where $limit
    GROUP BY client_group_name ORDER BY client_group_name

  ) AS T2 USING (client_group_name) LEFT JOIN (

    SELECT count(1) AS nb_cancelled, client_group_name 
      FROM $jobt AS Job JOIN client_group_member USING (ClientId) 
      JOIN client_group USING (client_group_id)
     WHERE JobStatus IN ('A') AND Type = 'B'
           $where $limit
    GROUP BY client_group_name ORDER BY client_group_name

  ) AS T4 USING (client_group_name) LEFT JOIN (

    SELECT count(1) AS nb_failed, client_group_name 
      FROM $jobt AS Job JOIN client_group_member USING (ClientId) 
      JOIN client_group USING (client_group_id)
     WHERE JobStatus IN ('E', 'e', 'f') AND Type = 'B'
           $where $limit
    GROUP BY client_group_name ORDER BY client_group_name

  ) AS T5 USING (client_group_name)

";
    $self->debug($query);
    my $all = $self->dbh_selectall_hashref($query, 'name') ;
    $self->debug($all);

    my $ret = { ID => $cur_id++,
                age => $arg->{age},
                since => $arg->{since},
                label => $label,
                Stats => [ values %$all ]};

    if ($param{client_groups}) {
        $ret->{client_groups} = [ map { { name => $_ } } @{$param{client_groups}} ];
    }

    $self->display($ret, "display_stats.tpl") ;
}

# return the autochanger list to update
sub eject_media
{
    my ($self) = @_;
    $self->can_do('r_media_mgnt');

    my %ret; 
    my $arg = $self->get_form('jmedias');

    unless ($arg->{jmedias}) {
        return $self->error("Can't get media selection");
    }

    my $query = "
SELECT Media.VolumeName  AS volumename,
       Storage.Name      AS storage,
       Location.Location AS location,
       Media.Slot        AS slot
FROM Media INNER JOIN Storage  ON (Media.StorageId  = Storage.StorageId)
           LEFT  JOIN Location ON (Media.LocationId = Location.LocationId)
WHERE Media.VolumeName IN ($arg->{jmedias})
  AND Media.InChanger = 1
";

    my $all = $self->dbh_selectall_hashref($query, 'volumename');

    foreach my $vol (values %$all) {
        my $a = $self->ach_get($vol->{location});
        next unless ($a) ;
        $ret{$vol->{location}} = 1;

        unless ($a->{have_status}) {
            $a->status();
            $a->{have_status} = 1;
        }
        # TODO: set enabled
        print "eject $vol->{volumename} from $vol->{storage} : ";
        if ($a->send_to_io($vol->{slot})) {
            print "<img src='/bweb/T.png' alt='ok'><br/>";
        } else {
            print "<img src='/bweb/E.png' alt='err'><br/>";
        }
    }
    return keys %ret;
}

sub move_email
{
    my ($self) = @_;

    my ($to, $subject, $content) = (CGI::param('email'),
                                    CGI::param('subject'),
                                    CGI::param('content'));
    $to =~ s/[^\w\d\.\@<>,]//;
    $subject =~ s/[^\w\d\.\[\]]/ /;    

    # will search for a custom mail command in /opt/bacula/bin
    my $cmd = $self->get_location("mail");
    if (!$cmd) {
        return $self->error("Unable to find a mail command");
    }
    open(MAIL, "|$cmd/mail -s '$subject' '$to'") ;
    print MAIL $content;
    close(MAIL);

    print "Mail sent";
}

sub old_restore
{
    my ($self) = @_;
    
    my $arg = $self->get_form('jobid', 'client');

    print CGI::header('text/brestore');
    print "jobid=$arg->{jobid}\n" if ($arg->{jobid});
    print "client=$arg->{client}\n" if ($arg->{client});
    print "\n\nYou have to assign this mime type with /usr/bin/brestore.pl\n";
    print "\n";
}

# Normal files restore page load
sub restore
{
    my ($self) = @_;
    my $res = {
        ver => $self->get_ver()
    };
    $self->display($res, "restore.tpl");
}

sub restore_exchange
{
    my ($self) = @_;
    my $res = {
        ver => $self->get_ver()
    };
    $self->display($res, "restore_exchange.tpl");
}
# VMware type restore single file page load
sub restore_vmware
{
    my ($self) = @_;
    my $res = {
        ver => $self->get_ver()
    };
    $self->display($res, "restore_vmware.tpl");
}

# Get restore single file working directory.
# It is place where teporary are stored files to restore
# up till time when they are downloaded and removed.
# Returns:
#  string working directory for storing files to restore
sub get_restore_single_file_workdir {
    my $self = shift;

    my $sd_opts = $self->get_component_options('Storage', 'Storage');
    my $workdir = "$sd_opts->[0]->{WorkingDirectory}/$restore_single_file_subdir";

    unless (-d "$workdir") {
        mkdir("$workdir", 0700);
    }

    return $workdir;
}

# Prepare single file restore files in single file restore functionality
# Params:
#  string $source_path path to provide as restore single file
#  string $access_method restore file access method ('zip', 'tar.gz', 'raw')
# Returns:
#  hashref destination file name and destination file path
#          for example: {file_name => 'abc.tar.gz', dest_path => '/opt/bacula/working'}
sub prepare_restore_single_file {
    my ($self, $root_path, $source_path, $access_method) = @_;
    my ($dest_file_name, $workdir, $result, $preparation_log);

    my @paths;
    if (ref($source_path) eq 'STRING') {
        push @paths, Encode::decode_utf8("${root_path}${source_path}");
    } elsif(ref($source_path) eq 'ARRAY') {
        for (my $i = 0; $i < scalar(@{$source_path}); $i++) {
            push @paths, Encode::decode_utf8("${root_path}$source_path->[$i]");
        }
    } else {
        $self->edebug("Unsupported paths format: " . ref($source_path));
    }

    if (scalar(@paths) > 0) {
        my $sd_workdir = $self->get_component_workdir('Storage');

        for (my $i = 0; $i < scalar(@paths); $i++) {
            if (!$self->validate_single_file_to_restore($paths[$i], $sd_workdir)) {
                $self->edebug("Invalid file name given '$paths[$i]'.");
                delete($paths[$i]);
            }
        }

        my @paths_cpy = @paths;
        my @rel_paths = map { s!^\Q${root_path}/\E!!; $_ } @paths_cpy;
        my $source_paths = '"' . join('" "', @rel_paths) . '"';

        if ($access_method =~ /^(zip|tar\.gz|raw)$/) {
            $workdir = $self->get_restore_single_file_workdir();
        }

        # Add file preparation date and time to file name
        $dest_file_name = $restore_filename_prefix . strftime("%Y-%m-%d_%H-%M-%S", localtime());
        if ($access_method eq 'zip') {
            chdir($root_path);
            $preparation_log = `zip -r "$workdir/$dest_file_name.zip" $source_paths 2>&1`;
            if ($? != 0) {
                $dest_file_name = "";
                $self->edebug("Zip compressing file '$workdir/$dest_file_name.zip' failed: $!");
            } else {
                $dest_file_name .= ".zip";
            }
        } elsif ($access_method eq 'tar.gz') {
            chdir($root_path);
            $preparation_log = `tar zcvf "$workdir/$dest_file_name.tar.gz" $source_paths 2>&1`;
            if ($? != 0) {
                $dest_file_name = "";
                $self->edebug("Tar.gz compressing file '$workdir/$dest_file_name.tar.gz' failed: $!");
            } else {
                $dest_file_name .= ".tar.gz";
            }
        } elsif ($access_method eq 'raw' && scalar(@paths) == 1 && -e $paths[0]) {
            my $file_name = basename($paths[0]);
            my $dest_path = "$workdir/$file_name";
            $self->copy_single_file($paths[0], $dest_path);
            $dest_file_name = "$file_name";
        }
    }

    $result = {dest_path => $workdir, file_name => $dest_file_name, preparation_log => $preparation_log};
    return $result;
}

sub validate_single_file_to_restore {
    my ($self, $path, $root_dir) = @_;
    my $valid = 0;

    my $file_name = basename($path);
    my $full_source_path = File::Spec->rel2abs($path);

    # do nothing with '.' and '..' names and when file path does not start from sd workdir
    if ($file_name =~ /^\.{1,2}$/ || !defined($root_dir) || $full_source_path !~ /^$root_dir/) {
        $valid = 0;
    # allow only directories, files and symbolic links
    } elsif (! -d $path || -e $path || -l $path) {
        $valid = 1;
    }
    return $valid;
}

# TODO : move this to Bweb::Autochanger ?
# TODO : make this internal to not eject tape ?
use Bconsole;

sub display_files
{
    my ($self) = @_ ;
    $self->can_do('r_view_job');

    my $arg = $self->get_form(qw/limit offset jobid/);

    $arg->{jobid} = $self->filter_jobid($arg->{jobid});

    if (!$arg->{jobid}) {
        return $self->error("Can't get jobid");
    }

    $self->display({
        title => "Content of JobId $arg->{jobid}  ",
        name => "list files jobid=$arg->{jobid}",
        notail => 1,
        id => $cur_id++,
    }, "command.tpl");  

    my $b = $self->get_bconsole();
    $b->display_one_cmd("list files jobid=$arg->{jobid} limit=$arg->{limit}"); # TODO: add offset

    $self->display({
        nohead => 1,
    }, "command.tpl");  
}

sub ach_get
{
    my ($self, $name) = @_;
    
    unless ($name) {
        return $self->error("Can't get your autochanger name ach");
    }

    unless ($self->{info}->{ach_list}) {
        return $self->error("Could not find any autochanger");
    }
    
    my $a = $self->{info}->{ach_list}->{$name};

    unless ($a) {
        $self->error("Can't get your autochanger $name from your ach_list");
        return undef;
    }

    $a->{bweb}  = $self;
    $a->{debug} = $self->{debug};

    return $a;
}

sub ach_register
{
    my ($self, $ach) = @_;
    $self->can_do('r_configure');
    my $conf = $self->{info}->get_current_conf();
    $conf->{ach_list}->{$ach->{name}} = $ach;

    $self->{info}->save();
    
    return 1;
}

sub ach_edit
{
    my ($self) = @_;
    $self->can_do('r_configure');

    my $arg = $self->get_form('ach');
    if (!$arg->{ach} 
        or !$self->{info}->{ach_list} 
        or !$self->{info}->{ach_list}->{$arg->{ach}}) 
    {
        return $self->error("Can't get autochanger name");
    }

    my $ach = $self->{info}->{ach_list}->{$arg->{ach}};

    my $i=0;
    $ach->{drives} = 
        [ map { { name => $_, index => $i++ } } @{$ach->{drive_name}} ] ;

    my $b = $self->get_bconsole();

    my @storages = $b->list_storage() ;

    $ach->{devices} = [ map { { name => $_ } } @storages ];
    
    $self->display($ach, "ach_add.tpl");
    delete $ach->{drives};
    delete $ach->{devices};
    return 1;
}

sub ach_del
{
    my ($self) = @_;
    $self->can_do('r_configure');

    my $arg = $self->get_form('ach');

    if (!$arg->{ach} 
        or !$self->{info}->{ach_list} 
        or !$self->{info}->{ach_list}->{$arg->{ach}}) 
    {
        return $self->error("Can't get autochanger name");
    }
   
    delete $self->{info}->{ach_list}->{$arg->{ach}} ;
   
    $self->{info}->save();
    $self->{info}->view();
}

sub ach_add
{
    my ($self) = @_;
    $self->can_do('r_configure');

    my $arg = $self->get_form('ach', 'mtxcmd', 'device', 'precmd');

    my $b = $self->get_bconsole();
    my @storages = $b->list_storage() ;

    unless ($arg->{ach}) {
        $arg->{devices} = [ map { { name => $_ } } @storages ];
        return $self->display($arg, "ach_add.tpl");
    }

    my @drives ;
    foreach my $drive (CGI::param('drives'))
    {
        unless (grep(/^$drive$/,@storages)) {
            return $self->error("Can't find $drive in storage list");
        }

        my $index = CGI::param("index_$drive");
        unless (defined $index and $index =~ /^(\d+)$/) {
            return $self->error("Can't get $drive index");
        }

        $drives[$index] = $drive;
    }

    unless (@drives) {
        return $self->error("Can't get drives from Autochanger");
    }

    my $a = new Bweb::Autochanger(name   => $arg->{ach},
                                  precmd => $arg->{precmd},
                                  drive_name => \@drives,
                                  device => $arg->{device},
                                  mtxcmd => $arg->{mtxcmd});

    $self->ach_register($a) ;
    
    $self->{info}->view();
}

sub delete
{
    my ($self) = @_;
    $self->can_do('r_delete_job');

    my $arg = $self->get_form('jobid');
    $arg->{jobid} = $self->filter_jobid($arg->{jobid});

    if ($arg->{jobid}) {
        my $b = $self->get_bconsole();
        my $ret = $b->send_cmd("delete jobid=\"$arg->{jobid}\"");

        $self->display({
            content => $ret,
            title => "Delete a job ",
            name => "delete jobid=$arg->{jobid}",
            id => $cur_id++,
        }, "command.tpl");      
    }
}

sub do_update_media
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');

    my $arg = $self->get_form(qw/media volstatus inchanger pool
                                 slot volretention voluseduration 
                                 maxvoljobs maxvolfiles maxvolbytes
                                 qcomment poolrecycle enabled cacheretention
                              /);

    unless ($arg->{media}) {
        return $self->error("Can't find media selection");
    }

    my $update = "update volume=$arg->{media} ";

    if ($arg->{volstatus}) {
        $update .= " volstatus=$arg->{volstatus} ";
    }
    
    if ($arg->{inchanger}) {
        $update .= " inchanger=yes " ;
        if ($arg->{slot}) {
            $update .= " slot=$arg->{slot} ";
        }
    } else {
        $update .= " slot=0 inchanger=no ";
    }

    if ($arg->{enabled}) {
        $update .= " enabled=$arg->{enabled} ";
    }

    if ($arg->{pool}) {
        $update .= " pool=$arg->{pool} " ;
    }

    if (defined $arg->{volretention}) {
        $update .= " volretention=\"$arg->{volretention}\" " ;
    }

    if (defined $arg->{cacheretention}) {
        $update .= " cacheretention=\"$arg->{cacheretention}\" " ;
    }

    if (defined $arg->{voluseduration}) {
        $update .= " voluse=\"$arg->{voluseduration}\" " ;
    }

    if (defined $arg->{maxvoljobs}) {
        $update .= " maxvoljobs=$arg->{maxvoljobs} " ;
    }
    
    if (defined $arg->{maxvolfiles}) {
        $update .= " maxvolfiles=$arg->{maxvolfiles} " ;
    }    

    if (defined $arg->{maxvolbytes}) {
        $update .= " maxvolbytes=$arg->{maxvolbytes} " ;
    }    

    if (defined $arg->{poolrecycle}) {
        $update .= " recyclepool=\"$arg->{poolrecycle}\" " ;
    } else {
        $update .= " recyclepool= " ;
    }
    
    my $b = $self->get_bconsole();

    $self->display({
        content => $b->send_cmd($update),
        title => "Update a volume ",
        name => $update,
        id => $cur_id++,
        hide_output => 1,
    }, "command.tpl");  


    my @q;
    my $media = $self->dbh_quote($arg->{media});

    my $loc = CGI::param('location') || '';
    if ($loc) {
        $loc = $self->dbh_quote($loc); # is checked by db
        push @q, "LocationId=(SELECT LocationId FROM Location WHERE Location=$loc)";
    }
    if (!$arg->{qcomment}) {
        $arg->{qcomment} = "''";
    }
    push @q, "Comment=$arg->{qcomment}";
    

    my $query = "
UPDATE Media 
   SET " . join (',', @q) . "
 WHERE Media.VolumeName = $media
";
    $self->dbh_do($query);

    $self->update_media();
}

sub update_slots
{
    my ($self) = @_;
    my $slots;
    
    $self->can_do('r_autochanger_mgnt');

    my $arg = $self->get_form('ach', 'slots', 'drive');

    my $ach = $self->ach_get($arg->{ach});
    unless ($ach) {
        return $self->error("Bad autochanger name");
    }

    $self->display({
        title => "Scanning autochanger content ",
        name => "update slots",
        notail => 1,
        id => $cur_id++,
    }, "command.tpl");  

    my $b = $self->get_bconsole();

    $b->update_slots($ach->{name}, $arg->{drive}, $arg->{slots});

    $self->display({
        nohead => 1,
    }, "command.tpl");  
}

sub get_job_log
{
    my ($self, $jobid) = @_;
    if ($self->cant_do('r_view_log')) {
        return 0;
    }

    my $arg = $self->get_form('jobid', 'limit', 'offset');

    if (!defined($jobid)) {
        $jobid = $arg->{jobid};
    }

    unless ($jobid) {
        return $self->error("Can't get jobid");
    }

    if ($arg->{limit} == 100) {
        $arg->{limit} = 1000;
    }
    # get security filter
    my $filter = $self->get_client_filter();

    my $query = "
SELECT Job.Name as name, Job.JobStatus as jobstatus, Client.Name as clientname
 FROM  Job INNER JOIN Client USING (ClientId) $filter
 WHERE JobId = $jobid
";

    my $row = $self->dbh_selectrow_hashref($query);

    unless ($row) {
        return $self->error("Can't find $jobid in catalog");
    }

    # display only Error and Warning messages
    $filter = '';
    if (CGI::param('error')) {
        $filter = " AND LogText $self->{sql}->{MATCH} 'Error|Warning|ERR=' ";
    }

    my $logtext;
    if (CGI::param('time') || $self->{info}->{display_log_time}) {
        $logtext = $self->dbh_strcat('Time', " ' ' ", 'LogText');
    } else {
        $logtext = 'LogText';
    }

    $query = "
SELECT count(1) AS nbline, 
       GROUP_CONCAT($logtext $self->{sql}->{CONCAT_SEP}) AS logtxt, id
  FROM  (
    SELECT 1 AS id, Time, LogText
    FROM Log 
   WHERE ( Log.JobId = $jobid 
      OR (Log.JobId = 0 
          AND Time >= (SELECT StartTime FROM Job WHERE JobId=$jobid) 
          AND Time <= (SELECT COALESCE(EndTime,$self->{sql}->{NOW}) 
                         FROM Job WHERE JobId=$jobid)
       ) ) $filter
 ORDER BY LogId
 LIMIT $arg->{limit}
 OFFSET $arg->{offset}
 ) AS temp
 GROUP BY id

";

    my $log = $self->dbh_selectrow_hashref($query);
    unless ($log) {
        my $lines = {};
        if ($row->{jobstatus} eq 'C') {
           $lines->{lines} = " ";
        }
        # if the "lines" value is reference to empty hash
        # an error message will be printed.
        $self->display($lines, 'display_log.tpl');
        return;
    }
    $log->{logtxt} =~ s/(\0|\\,)//g;
    $self->display({ lines=> $log->{logtxt},
                     nbline => $log->{nbline},
                     jobid => $jobid,
                     name  => $row->{name},
                     client => $row->{clientname},
                     offset => $arg->{offset},
                     limit  => $arg->{limit},
                 }, 'display_log.tpl');
}

sub display_running_job_zoom {
    my ($self, $jobid, $jobstatus) = @_;

    my $b = $self->get_bconsole();
    my $status = $b->get_job_status($jobid);

    $self->display({jobid => $jobid,
                    jobname => $status->{name},
                    jobtype => $status->{type},
                    status => $jobstatus,
                    status_desc => $status->{status_desc},
                    client => $status->{clientname},
                    storage => $status->{storage},
                    fileset => $status->{fileset},
                    level => $status->{level},
                    starttime => $status->{starttime},
                    jobfiles => $status->{jobfiles},
                    jobbytes => $status->{jobbytes},
                    joberrors => $status->{errors}
                }, 'running_job_zoom.tpl');
}

sub cancel_future_job
{
    my ($self) = @_;
    $self->can_do('r_cancel_job');

    my $arg = $self->get_form(qw/job pool level client when/);

    if ( !$arg->{job} or !$arg->{pool} or !$arg->{level} 
         or !$arg->{client} or !$arg->{when})
    {
        return $self->error("Can't get enough information to mark this job as canceled");
    }

    $arg->{level} =~ s/^(.).+/$1/; # we keep the first letter
    my $jobtable = $self->{info}->{stat_job_table} || 'JobHisto';

    if ($jobtable =~ /^Job$/i) {
        return $self->error("Can add records only in history table");
    }
    my $jname = "$arg->{job}.$arg->{when}";
    $jname =~ s/\s/_/g;

    my $found = $self->dbh_selectrow_hashref("
SELECT 1
  FROM $jobtable
 WHERE JobId = 0
   AND Job = '$jname'
   AND Name = '$arg->{job}'
");
    if ($found) {
        return $self->error("$jname is already in history table");
    }

        $self->dbh_do("
INSERT INTO $jobtable 
  (JobId, Name, Job, Type, Level, JobStatus, SchedTime, StartTime, EndTime, 
   RealEndTime, ClientId, PoolId) 
 VALUES 
  (0, '$arg->{job}', '$jname', 'B', '$arg->{level}', 'A',
   '$arg->{when}', '$arg->{when}', '$arg->{when}', '$arg->{when}',
   (SELECT ClientId FROM Client WHERE Name = '$arg->{client}'),
   (SELECT PoolId FROM Pool WHERE Name = '$arg->{pool}')
  )
");
    $self->display({ Filter => "Dummy record for $jname",
                     ID => 1,
                     Jobs => 
                         [{ jobid => 0,
                            client => $arg->{client},
                            jobname => $arg->{job},
                            pool => $arg->{pool},
                            level => $arg->{level},
                            starttime => $arg->{when},
                            duration => 0,
                            jobfiles => 0,
                            jobbytes => 0,
                            joberrors => 0,
                            jobstatus => 'A',
                     }]
                   },
                   "display_job.tpl");
}

sub add_media
{
    my ($self) = @_ ;
    $self->can_do('r_media_mgnt');
    my $arg = $self->get_form('storage', 'pool', 'nb', 'media', 'offset');
    my $b = $self->get_bconsole();

    if (!$arg->{storage} || !$arg->{pool} || not defined $arg->{nb} || !$arg->{media} || !$arg->{offset}) {
        CGI::param(offset => 0);
        $arg = $self->get_form('db_pools');
        $arg->{storage} = [ map { { name => $_ } }$b->list_storage()];
        $self->display($arg, 'add_media.tpl');
        return 1;
    }
    my $cmd = "add pool=\"$arg->{pool}\" storage=\"$arg->{storage}\"\n";
    if ($arg->{nb} > 0) {
        $arg->{offset} = $arg->{offset}?$arg->{offset}:1; 
        $cmd .= "$arg->{nb}\n";
        $cmd .= "$arg->{media}\n";
        $cmd .= "$arg->{offset}\n";
        $cmd .= "0\n";          # slot
        $cmd .= "no\n";         # inchanger (might not be used...)

    } else {
        $cmd .= "0\n";
        $cmd .= "$arg->{media}\n";
        $cmd .= "0\n";          # slot      (might not be used...)
        $cmd .= "no\n";         # inchanger
    }
    
    $self->edebug($cmd);
    $self->edebug($b->send_one_cmd($cmd));

    sleep(1);

    CGI::param('media', '');
    CGI::param('re_media', $arg->{media});
    $self->display_media();
}

sub label_barcodes
{
    my ($self) = @_ ;
    $self->can_do('r_autochanger_mgnt');

    my $arg = $self->get_form('ach', 'slots', 'drive', 'pool');

    unless ($arg->{ach}) {
        return $self->error("Can't find autochanger name");
    }

    my $a = $self->ach_get($arg->{ach});
    unless ($a) {
        return $self->error("Can't find autochanger name in configuration");
    } 

    my $storage = $a->get_drive_name($arg->{drive});
    unless ($storage) {
        return $self->error("Can't get your drive name");
    }

    my $slots_sql = '';
    my $t = 300 ;
    if ($arg->{slots}) {
        $slots_sql = " AND Slot IN ($arg->{slots}) ";
        $t += 60*scalar( CGI::param('slot') ) ;
    }
    my $pool = $arg->{pool} || 'Scratch';
    my $b = $self->get_bconsole();
    print "<h1>This command can take long time, be patient...</h1>";
    print "<pre>" ;
    $b->label_barcodes(storage => $storage,
                       drive => $arg->{drive},
                       pool  => $pool,
                       slots => $arg->{slots}) ;
    $b->close();
    print "</pre>";

    $self->dbh_do("
  UPDATE Media 
       SET LocationId =   (SELECT LocationId 
                             FROM Location 
                            WHERE Location = '$arg->{ach}')

     WHERE (LocationId = 0 OR LocationId IS NULL)
       $slots_sql
");

}

sub purge
{
    my ($self) = @_;
    $self->can_do('r_purge');

    my @volume = CGI::param('media');

    unless (@volume) {
        return $self->error("Can't get media selection");
    }

    my $b = $self->get_bconsole(timeout => 60);

    foreach my $v (@volume) {
        $self->display({
            content => $b->purge_volume($v),
            title => "Purge media",
            name => "purge volume=$v",
            id => $cur_id++,
        }, "command.tpl");
    }   
    $b->close();
}

sub prune
{
    my ($self) = @_;
    $self->can_do('r_prune');

    my @volume = CGI::param('media');
    unless (@volume) {
        return $self->error("Can't get media selection");
    }

    my $b = $self->get_bconsole(timeout => 60);
    my $output;
    foreach my $v (@volume) {
        $output = $b->prune_volume($v);
        if (!$output) {
            $output = $self->{translate}->get('Nothing to prune.');
        }
        $self->display({
            content => $output,
            title => "Prune volume",
            name => "prune volume=$v",
            id => $cur_id++,
        }, "command.tpl");
    }
    $b->close();
}

sub stop_or_cancel_job
{
    my ($self, $what) = @_;
    $self->can_do("r_${what}_job");

    my $cmdarg = '';

    if (CGI::param('all')) {
        $cmdarg = "all yes";

    } else {
        my $arg = $self->get_form('jobids');
    
        if ($arg->{jobids}) {
            $arg->{jobids} = $self->filter_jobid($arg->{jobids});
        }

        unless ($arg->{jobids}) {
            return $self->error("Can't get jobid");
        }

        $cmdarg = "jobid=$arg->{jobids}";
    }

    my $b = $self->get_bconsole();
    $self->display({
        content => $b->send_cmd("$what $cmdarg"),
        title => "$what job",
        name => "$what $cmdarg",
        id => $cur_id++,
    }, "command.tpl");  
}

sub restart_job
{
    my ($self) = @_;
    $self->can_do('r_restart_job');

    my $arg = $self->get_form('jobid');
    $arg->{jobid} = $self->filter_jobid($arg->{jobid});

    unless ($arg->{jobid}) {
        return $self->error("Can't get jobid");
    }

    my $jobid = $arg->{jobid};
    my $b = $self->get_bconsole();
    $self->display({            # TODO: handle other job type
        content => $b->send_cmd("restart incomplete jobid=$jobid"),
        title => "Restart job",
        name => "restart incomplete jobid=$arg->{jobid}",
        id => $cur_id++,
    }, "command.tpl");  

    print "<br>You can follow job (jobid=$jobid) execution <a href='?action=dsp_cur_job;jobid=$jobid'> here </a><script type='text/javascript' language='JavaScript'>setTimeout(function() { window.location='?action=dsp_cur_job;jobid=$jobid' + get_current_director_arg(\";\")},2000);</script>";
}

sub fileset_view
{
    # Warning, we display current fileset
    my ($self) = @_;

    $self->can_do('r_view_job');
    my $arg = $self->get_form('fileset');

    if ($arg->{fileset}) {
        my $b = $self->get_bconsole();
        my $ret = $b->get_fileset($arg->{fileset});
        $self->display({ fileset => $arg->{fileset},
                         %$ret,
                     }, "fileset_view.tpl");
    } else {
        $self->error("Can't get fileset name");
    }
}

sub director_show_sched
{
    my ($self) = @_ ;
    $self->can_do('r_view_job');
    my $arg = $self->get_form('days');

    my $b = $self->get_bconsole();
    my $ret = $b->director_get_sched( $arg->{days} );

    $self->display({
        id => $cur_id++,
        list => $ret,
    }, "scheduled_job.tpl");
}

sub enable_disable_job
{
    my ($self, $what) = @_ ;
    $self->can_do('r_run_job');

    my $arg = $self->get_form('jobs');
    if (!$arg->{jobs}) {
        return $self->error("Can't find job name");
    }

    my $b = $self->get_bconsole();

    my $cmd;
    if ($what) {
        $cmd = "enable";
    } else {
        $cmd = "disable";
    }

    my $out = '';
    my $cmds = '';
    my $fullcmd;
    for my $job (split(/,/, $arg->{jobs})) {
        $fullcmd = "$cmd job=\"$job\"";
        $out .= $b->send_cmd($fullcmd);
        $cmds .= "${fullcmd}; ";
    }

    $self->display({
        content => $out,
        title => $cmd,
        name => $cmds,
        id => $cur_id++,
    }, "command.tpl");  
}

sub clear_director
{
    my ($self) = @_;
    $self->dbh_disconnect();
    delete $self->{info}->{current_conf};
}

sub set_lang
{
    my ($self, $lang) = @_;
    $self->{current_lang} = $lang;
}

sub set_director
{
    my ($self, $dir) = @_;
    $self->clear_director();
    if ($dir) {
        # if we have a director in argument, we overwrite some fields
        if (!$self->{info}->apply_subconf($dir)) {
            return $self->error("Can't use $dir as director, check your configuration");
        }
    }

    if ($self->{info}->{dbi} && $self->{info}->{dbi} =~ /DBI:(\w+):/i) {
        $self->{sql} = $sql_func{$1};
    }

    return 1;
}

sub get_bconsole
{
    my ($self, @opts) = @_;
    my ($uid, $gid);
    if ($self->is_system_auth() == 1) {
        my $session = $self->get_session();
        $uid = $session->param('uid');
        $gid = $session->param('gid');
    }
    return new Bconsole(pref => $self->{info}, loginname => $self->{loginname}, uid => $uid, gid => $gid, @opts);
}


# Get clients for remote address
# It has meaning when system user authentication is used
# Returns:
#  arrayref list with clients (usually one client)
sub get_clients_per_remote_addr
{
    my $self = shift;
    my $b = $self->get_bconsole();
    my $remote_addr = $self->get_session()->query()->remote_addr();
    my @clients = $b->get_clients_per_remote_addr($remote_addr);
    return @clients;
}

sub storage_command
{
    my ($self, $arg) = @_;

    $self->can_do('r_storage_mgnt');

    my $ret = {cmd => undef, output => undef, sd_exitcode => undef};
    if ($arg->{storage} and $arg->{storage_cmd}) {
        my $b = $self->get_bconsole();
        my $cmd;
        my $dev;
        if(defined($arg->{device_name})) {
            $dev = "device=\"$arg->{device_name}\" drive=0";
        } else {
            $dev = "drive=$arg->{drive}";
        }
        if ($arg->{storage_cmd} eq 'setdebug') {
            $cmd = "setdebug storage=\"$arg->{storage}\" trace=1 level=$arg->{debuglevel}";

        } else {
            $cmd = "$arg->{storage_cmd} storage=\"$arg->{storage}\" $dev slot=$arg->{slot}";
        }
        my $output = $b->send_cmd($cmd);
        $ret->{cmd} = $cmd;
        $ret->{output} = $output;
        foreach my $l (split(/\n/, $output)) {
            if ($l =~ /^(3\d{3})\s+\S.*/) {
                $ret->{sd_exitcode} = $1;
                last;
            }
        }
    }
    return $ret;
}

sub cmd_storage
{
    my ($self) = @_;

    my $arg = $self->get_form(qw/storage storage_cmd drive device_name slot debuglevel/);

    if ($arg->{storage} and $arg->{storage_cmd}) {
        my $ret = $self->storage_command($arg);
        $self->display({
            content => $ret->{output},
            title => "Storage ",
            name => $ret->{cmd},
            id => $cur_id++,
        }, "command.tpl");
    } else {
        my $b = $self->get_bconsole();
        my $storages= [ map { { name => $_ } } $b->list_storage()];
        $self->display({ storage => $storages}, "cmd_storage.tpl");
    }

}

sub run_parse_show_output
{
    my ($self, $ouput) = @_;

    my %arg;
    my $key;
    my $val;
    my $curr_section;
    $self->debug($ouput);
    foreach my $l (split(/\r?\n/, $ouput)) {
        $self->debug($l);
        if ($l =~ /(\w+): name=([\w\d\.\s-]+?)(\s+\w+=.+)?$/i) {
            $key = lc($1);
            $val = $2;
            if (!exists($arg{$key})) {
                $arg{$key} = $val;
            } elsif ($key =~ /^(autochanger|storage)$/i) {
                # Set read and write storage for jobs that use NextPool
                # directive (Copy and Migration job types).
                if ($curr_section =~ /^pool$/i) {
                    $arg{read_storage} = $val;
                } elsif ($curr_section =~ /^nextpool$/i) {
                    $arg{write_storage} = $val;
                }
            }
            $curr_section = $key;
            $l = $3 
                if ($3) ;
        }

        $l =~ s/\s+$//g;
        if (my @l = $l =~ /(\w+)=([\w\d\,\.\- %\*]+)(?=\s|$)/g) {
            for (my $i = 0; $i < scalar(@l); $i+=2) {
                $key = lc($l[$i]);
                $val = $l[$i+1];
                if (!exists($arg{$key})) {
                    $arg{$key} = $val;
                }
            }
        } elsif ($l =~ /(where|command|mailcmd|opcmd|writebootstrap)=([\S\s]+)$/ui) {
            # utf8 directives (usually paths)
            $arg{$1} = $2;
        }
    }

    my %lowcase ;
    foreach my $k (keys %arg) {
        $lowcase{lc($k)} = $arg{$k} ;
    }
    $self->debug(\%lowcase);
    return \%lowcase;
}

# Get job default directives taken from 'show job=xxx' bconsole command output
sub get_job_directives {
    my ($self, $job) = @_;
    $self->can_do('r_run_job');

    if (!$job) {
        return $self->error("Can't get job name");
    }

    my $b = $self->get_bconsole();

    # we take informations from director, and we overwrite with user wish
    my $info = $b->send_cmd("show job=\"$job\"");
    my $attr = $self->run_parse_show_output($info);
    return $attr;
}

# Get pool default directives taken from 'show pool=xxx' bconsole command output
sub get_pool_directives {
    my ($self, $pool) = @_;
    $self->can_do('r_run_job');

    if (!$pool) {
        return $self->error("Can't get pool name");
    }

    my $b = $self->get_bconsole();

    my $info = $b->send_cmd("show pool=\"$pool\"");
    my $attr = $self->run_parse_show_output($info);
    return $attr;
}

sub run_job_mod
{
    my ($self) = @_;
    $self->can_do('r_run_job');

    my $b = $self->get_bconsole();
    my $arg = $self->get_form(qw/pool level client fileset storage media job comment/);

    if (!$arg->{job}) {
        return $self->error("Can't get job name");
    }

    my $attr = $self->get_job_directives($arg->{job});
    
    if (!$arg->{pool} and $arg->{media}) {
        my $r = $self->dbh_selectrow_hashref("
SELECT Pool.Name AS name
  FROM Media JOIN Pool USING (PoolId)
 WHERE Media.VolumeName = '$arg->{media}'
   AND Pool.Name != 'Scratch'
");
        if ($r) {
            $arg->{pool} = $r->{name};
        }
    }

    my %job_opt = (%$attr, %$arg);
    
    my $jobs   = [ map {{ name => $_ }} $b->list_job() ];

    my $pools  = [ map { { name => $_ } } $b->list_pool() ];
    my $clients = [ map { { name => $_ } }$b->list_client()];
    my $filesets= [ map { { name => $_ } }$b->list_fileset() ];
    my $storages= [ map { { name => $_ } }$b->list_storage()];

    $self->display({
        jobs     => $jobs,
        pools    => $pools,
        clients  => $clients,
        filesets => $filesets,
        storages => $storages,
        %job_opt,
    }, "run_job_mod.tpl");
}

sub run_job
{
    my ($self) = @_;
    $self->can_do('r_run_job');

    my $b = $self->get_bconsole();
    
    my $jobs   = [ map {{ name => $_ }} $b->list_job() ];
    if ($b->{error}) {
        return $self->error("Bconsole returns an error, check your setup. ERR=$b->{error}");
    }
    $self->display({
        jobs     => $jobs,
    }, "run_job.tpl");
}

sub run_job_start
{
    my ($self) = @_;
    $self->can_do('r_run_job');
    my $arg = $self->get_form(qw/db_pools job/);
    my $clients = $self->clients_list();
    my $pools = $arg->{db_pools};
    my $b = $self->get_bconsole();
    my $storages = [ map { { name => $_ } } $b->list_storage() ];
    my $jobs_all = [ map { { name => $_ } } $b->list_job() ];
    my @restores = $b->list_restore();
    my $filesets= [ map { { name => $_ } }$b->list_fileset() ];
    my @jobs;

    # create new array without restore jobs (not needed)
    foreach my $job (@{$jobs_all}) {
        unless (grep($job->{name} eq $_, @restores)) {
            push @jobs, $job;
        }
    }

    my $res = {
       jobs => \@jobs,
       clients => JSON::encode_json($clients),
       pools => JSON::encode_json($pools),
       filesets => JSON::encode_json($filesets),
       storages => JSON::encode_json($storages),
       job => $arg->{job},
       ver => $self->get_ver()
    };
    $self->display($res, "run_job_start.tpl");
}

sub run_job_now
{
    my ($self) = @_;
    $self->can_do('r_run_job');

    my $b = $self->get_bconsole();
    
    # TODO: check input (don't use pool, level)

    my $arg = $self->get_form(qw/pool level client priority when 
                                 fileset job storage comment nextpool
                                 verifyjob jobid/);
    if (!$arg->{job}) {
        return $self->error("Can't get your job name");
    }

    my %params = (
        job => $arg->{job},
        client => $arg->{client},
        priority => $arg->{priority},
        level => $arg->{level},
        storage => $arg->{storage},
        pool => $arg->{pool},
        fileset => $arg->{fileset},
        when => $arg->{when},
        comment => $arg->{comment}
    );

    if ($arg->{nextpool}) { # for copy/migrate job types
        $params{nextpool} = $arg->{nextpool};
    }

    # for verify job type
    if ($arg->{verifyjob}) {
        $params{verifyjob} = $arg->{verifyjob};
    } elsif ($arg->{jobid}) {
        $params{jobid} = $arg->{jobid};
    }

    my $jobid = $b->run(%params);

    if ($b->{error}) {
        print $b->{error};
    }

    $self->display_running_job($jobid);
}

sub display_next_job
{
    my ($self) = @_;
    $self->can_do('r_view_job');

    my $arg = $self->get_form(qw/job begin end/);
    if (!$arg->{job}) {
        return $self->error("Can't get job name");
    }

    my $b = $self->get_bconsole();

    my $job = $b->send_cmd("show job=\"$arg->{job}\"");
    my $attr = $self->run_parse_show_output($job);
    
    if (!$attr->{schedule}) {
        return $self->error("Can't get $arg->{job} schedule");
    }
    my $jpool=$attr->{pool} || '';

    my $sched = new Bweb::Sched(bconsole => $b, name => $attr->{schedule},
                                begin => $arg->{begin}, end => $arg->{end});

    my $ss = $sched->get_scheds($attr->{schedule}); 
    my @ret;

    foreach my $s (@$ss) {
        my $level = $sched->get_level($s);
        my $pool  = $sched->get_pool($s) || $jpool;
        my $evt = $sched->get_event($s);
        push @ret, map { "$_ : $pool ($level)\n" } @$evt;
    }
    
    print "<b>$arg->{job}:</b><pre>", sort @ret, "</pre><br>";
}

# permit to verify for higher level backup
# we attempt a Increment, we made a Full, that ok
# TODO: Pool may have change
sub get_higher_level
{
    my ($self, $level) = @_;
    if ($level eq 'F') {
        return "'F'";
    } elsif ($level eq 'D') {
        return "'F', 'D'";
    } elsif ($level eq 'I') {
        return "'F', 'D', 'I'";
    }
    return "''";
}

# check jobs against their schedule
sub check_job
{
    my ($self, $sched, $schedname, $job, $job_pool, $client, $type) = @_;
    return undef if (!$self->can_view_client($client));

    $self->debug("checking $job, $job_pool, $client, $type, $schedname");

    my $sch = $sched->get_scheds($schedname);    
    return undef if (!$sch);

    my @ret;
    foreach my $s (@$sch) {
        my $pool;
        if ($type eq 'B') {     # we take the pool only for backup job
            $pool = $sched->get_pool($s) || $job_pool;
        }
        my $level = $sched->get_level($s);
        my ($l) = ($level =~ m/^(.)/); # we keep the first letter
        $l = $self->get_higher_level($l);
        my $evts = $sched->get_event($s);
        my $end = $sched->{end}; # this backup must have start before the next one
        foreach my $evt (reverse @$evts) {
            my $all = $self->dbh_selectrow_hashref("
 SELECT 1
   FROM Job
   JOIN Client USING (ClientId) LEFT JOIN Pool USING (PoolId)
  WHERE Job.StartTime >= '$evt' 
    AND Job.StartTime <  '$end'
    AND Job.Name = '$job'
    AND Job.Type = '$type'
    AND Job.JobStatus IN ('T', 'W')
    AND Job.Level IN ($l)
" . ($pool?" AND Pool.Name = '$pool' ":'') . "
    AND Client.Name = '$client'
 LIMIT 1
");             
            if ($all) {
                $self->debug("found job record for $job on $client");
            } else {
                push @{$self->{tmp}}, {date => $evt, level => $level,
                                       type => 'Backup', name => $job,
                                       pool => $pool, volume => $pool,
                                       client => $client};
            }
            $end = $evt;
        }
    }
}

sub display_missing_job
{
    my ($self) = @_;
    my $arg = $self->get_form(qw/begin end age/);

    if (!$arg->{begin}) { # TODO: change this
        $arg->{begin} = strftime('%F %T', localtime($btime - $arg->{age}));
    }
    if (!$arg->{end}) {
        $arg->{end} = strftime('%F %T', localtime($btime));
    }
    $self->{tmp} = [];          # check_job use this for result

    my $bconsole = $self->get_bconsole();

    my $sched = new Bweb::Sched(bconsole => $bconsole,
                                begin => $arg->{begin},
                                end => $arg->{end});
    $self->debug($sched);
    my $job = $bconsole->send_cmd("show job");
    my ($jname, $jsched, $jclient, $jpool, $jtype);
    foreach my $j (split(/\r?\n/, $job)) {
        if ($j =~ /Job: name=([\w\d\-]+?) JobType=(\d+)/i) {
            if ($jname and $jsched) {
                $self->check_job($sched, $jsched, $jname, 
                                 $jpool, $jclient, $jtype);
            }
            $jname = $1;
            $jtype = chr($2);
            $jclient = $jpool = $jsched = undef;
        } elsif ($j =~ /Client: name=(.+?) enabled=/i) {
            $jclient = $1;
        } elsif ($j =~ /Pool: name=([\w\d\-]+) PoolType=/i) {
            $jpool = $1;
        } elsif ($j =~ /Schedule: name=([\w\d\-]+)/i) {
            $jsched = $1;
        }
    }
    $self->display({
        id => $cur_id++,
        title => "Missing Jobs (from $arg->{begin} to $arg->{end})",
        list => $self->{tmp},
        wiki_url => $self->{info}->{wiki_url},
        missing_mode => 1,
    }, "scheduled_job.tpl");

    delete $self->{tmp};
}

sub filter_jobid
{
    my ($self, $jobid) = (@_);
    my $filter = $self->get_client_filter();

    if (!$filter) {
        return $jobid;
    }

    my $q = "SELECT JobId FROM Job JOIN Client USING (ClientId) $filter WHERE JobId IN ($jobid)";
    my $a = $self->dbh_selectall_arrayref($q);
    return join(",", map { $_->[0] } @$a );
}

sub config_check
{
    my ($self) = @_;
    my @content;

    if ($self->dbh_is_mysql()) {

    } elsif ($self->dbh_is_pg()) {

# Tests to do on  pg_settings
#
# syntax : <level> <variable> <op>
#   <level >   alerte level, ie. WARNING or CRITICAL
#   <variable> variable of pg_settings
#   <op>       operation to do in perl lang
#
        my $pg_settings_check = "
W  stats_start_collector ne 'on'
W  enable_hashagg        ne 'on'
W  enable_hashjoin       ne 'on'
W  enable_indexscan      ne 'on'
W  enable_mergejoin      ne 'on'
W  enable_nestloop       ne 'on'
W  enable_seqscan        ne 'on'
W  enable_sort           ne 'on'
W  enable_tidscan        ne 'on'
W  max_fsm_pages          < 10000
W  log_executor_stats    ne 'off'
W  log_parser_stats      ne 'off'
W  log_planner_stats     ne 'off'
W  log_statement_stats   ne 'off'
W  log_checkpoints       ne 'on'
W  random_page_cost      != 4
W  autovacuum            ne 'on'
W  wal_buffer             < 4194304
W  wal_buffer             > 8388608
W  checkpoint_segments    < 16
W  checkpoint_segments    > 512
W  checkpoint_timeout     < 1200
W  effective_cache_size   < 131072
W  work_mem               < 32768
W  work_mem               > 1048576
W  maintenance_work_mem   < 102400
W  shared_buffers         < 131072
W  shared_buffers         > 1310720
f  fsync                 ne 'on'
";

        my $all = $self->dbh_selectall_hashref("SELECT name, setting, unit, short_desc FROM pg_settings", 
                                               "name");
        # parse conf 
        foreach my $l (split('\n', $pg_settings_check)) {

            next if ($l eq "");

            if (! ($l =~ /(\w+)\s+(\w+)\s+(.*)$/) ) {
                next;
            }

            my ($level, $key, $op) = ($1, $2, $3);

            # not in pg_settings
            next if (not defined $all->{$key});

            my $res = -1;
            eval("\$res = ( \$all->{\$key}->{setting} $op )");
            
            if ($res == 1) {
		push @content, {level => $level,
                                key   => $key,
                                desc  => $all->{$key}->{short_desc},
                                unit  => $all->{$key}->{unit},
                                setting => $all->{$key}->{setting},
                                op    => $op};
            }
        }
    }

    my $row = $self->dbh_selectall_arrayref("SELECT 1 FROM bweb_role LIMIT 1");
    if (!$row || !$row->[0] || $row->[0]->[0] != 1) {
        push @content, {level => 'f',
                        key => 'Bweb SQL functions',
                        unit => '',
                        setting => 'Not found',
                        desc => 'Check if bweb-postgresql.sql or bweb-mysql.sql are loaded',
                        op => ''}
    }

    $self->display({
        content => \@content,
        id => $cur_id++,
    }, "catalog_check.tpl");

    # Get components version
    my @comps = keys(%Services);
    @comps = sort(@comps);
    my @comp_ver  = ();
    foreach my $comp (@comps) {
        my $comp_params = $self->get_component_version($comp);
        $comp_params->{component} = $comp;
        push @comp_ver, $comp_params;
    }
    $self->display({
        comp_ver => \@comp_ver
    }, "component_version.tpl");

    ### Checking available features ###
    my $features_support = $self->get_support(qw/ssh smb sir/);
    $self->display({
        ssh_keys => $features_support->{ssh}->{support},
        smb_tools => $features_support->{smb}->{support},
        vmware_restore => $features_support->{sir}->{vmware},
        exchange_restore => $features_support->{sir}->{exchange}
    }, "features_check.tpl");

    ### Configuration check (selftest suite) ###

    $self->display({
        title => "Testing bweb configuration",
        name => "bweb_testconfig",
        notail => 1,
        id => $cur_id++,
    }, "command.tpl");  

    system("./bweb_testconfig.pl");

    $self->display({
        nohead => 1,
    }, "command.tpl");  
}

# Getting SD status via '.api 2' command for:
# - headers
# - devices
# - running jobs
# - terminated jobs
sub get_storage_status
{
    my ($self) = @_;

    my $arg = $self->get_form('storage');

    my $bconsole = $self->get_bconsole();
    my $storage_devices = $bconsole->get_storage_devices();
    my $status = {header => {}, devices => {}, running => {}, terminated => {}};
    my $output;
    my $device;
    my $jobid;
    my $autochanger = undef;
    for my $type (keys %{$status}) {
        $output = $bconsole->send_cmd_api(".status storage=\"$arg->{storage}\" $type");
        foreach my $l (@$output) {
            if ($l =~ /^(\S+)=([\S\s]+)\n$/) {
                if ($type eq 'header') {
                    $status->{$type}{$1} = $2;
                } elsif ($type eq 'devices') {
                    if ($1 eq 'autochanger') {
                        $autochanger = $2;
                        next;
                    }
                    if ($1 eq 'name') {
                        $device = $2;
                        if(!defined($status->{$type}{$device})) {
                            $status->{$type}{$device} = {};
                        }
                    }
                    if (defined($autochanger) && $1 eq 'device') {
                        $status->{$type}{$device}{autochanger} = $autochanger;
                        $status->{$type}{$device}{storage} = $storage_devices->{$autochanger}
                    }
                    if ($1 eq 'archive_device') {
                        $autochanger = undef;
                    }
                    if (defined($device)) {
                        if (!defined($status->{$type}{$device}{storage})) {
                            $status->{$type}{$device}{storage} = $storage_devices->{$device};
                        }
                        $status->{$type}{$device}{$1} = $2;
                    }
                    if ($1 eq 'volume') {
                        my $media = $self->get_media_by_name($2);
			$status->{$type}{$device}{slot} = $media->{slot};
                    }
                } elsif ($type eq 'terminated' || $type eq 'running') {
                    if ($1 eq 'jobid') {
                        $jobid = $2;
                        $status->{$type}{$jobid} = {};
                    }
                    if (defined($jobid)) {
                        $status->{$type}{$jobid}{$1} = $2;
                    }
                }
            }
        }
    }
    $status->{storages} = $storage_devices;
    return $status;
}

sub status_storage
{
    my ($self) = @_;
    $self->can_do('r_storage_mgnt');

    my $arg = $self->get_form('storage');

    $self->display({
        storage => $arg->{storage},
    }, "status_storage.tpl");
}

# Getting Client/FileDaemon status via '.api 2' command for:
# - headers
# - running jobs
# - terminated jobs
sub get_client_status
{
    my ($self) = @_;

    my $arg = $self->get_form('client');

    my $bconsole = $self->get_bconsole();
    my $status = {header => {}, running => {}, terminated => {}};
    my $output;
    my $jobid;
    for my $type (keys %{$status}) {
        $output = $bconsole->send_cmd_api(".status client=\"$arg->{client}\" $type");
        foreach my $l (@$output) {
            if ($l =~ /^(\S+)=([\S\s]+)\n$/) {
                if ($type eq 'header') {
                    $status->{$type}{$1} = $2;
                } elsif ($type eq 'terminated' || $type eq 'running') {
                    if ($1 eq 'jobid') {
                        $jobid = $2;
                        $status->{$type}{$jobid} = {};
                    }
                    if (defined($jobid)) {
                        $status->{$type}{$jobid}{$1} = $2;
                    }
                }
            }
        }
    }
    $status->{header}{tls_possible} = Bweb::tls_support_check($status->{header}{version});

    return $status;
}

sub status_client
{
    my ($self) = @_;
    $self->can_do('r_client_status');

    my $arg = $self->get_form('client');
    my $client_prop = {};

    if (defined($arg->{client})) {
        $client_prop = $self->show_client($arg->{client});
        $client_prop->{client} = $arg->{client};
        my $client = $self->dbh_quote($arg->{client});
        my $query = "
 SELECT
    count(Job.JobId)     AS nb_jobs,
    sum(Job.JobBytes)    AS nb_bytes,
    sum(Job.JobErrors)   AS nb_err,
    sum(Job.JobFiles)    AS nb_files,
    Client.Name          AS clientname,
    Client.Uname         AS uname
 FROM Job JOIN Client USING (ClientId)
 WHERE
    Client.Name = $client
 GROUP BY Client.Name";
        my $row = [ $self->dbh_selectrow_hashref($query) ];
        $client_prop->{nb_jobs} = $row->[0]{nb_jobs};
        $client_prop->{nb_bytes} = human_size($row->[0]{nb_bytes});
        $client_prop->{nb_err} = $row->[0]{nb_err};
        $client_prop->{nb_files} = $row->[0]{nb_files};

        # Uname from catalog database, not from status client.
        # The uname is not taken by regexp due to dificult to guess
        # uname values. Safer is use index in this case.
        my $uname_idx = index($row->[0]{uname}, ')');
        $client_prop->{uname} = substr($row->[0]{uname}, ++$uname_idx);
    }

    my $clients = $self->clients_list();
    $client_prop->{clients} = [ values %$clients ];
    $client_prop->{ver} = $self->get_ver();

    $self->display($client_prop, "status_client.tpl");
}

# Parse 'show client=xxx' command output
# Catalog values are put with trailing 'cat_' string in every key as prefix.
sub show_client
{
    my ($self, $client) = @_;

    my $client_prop = {};
    my $bconsole = $self->get_bconsole();
    my $output = $bconsole->send_cmd("show client=\"$client\"");
    my $prefix = "";
    my $key = undef;
    for my $l (split(/\r?\n/, $output)) {
        chomp($l);
	for my $n (split(/=/, $l)) {
            chomp($n);
            if ($n =~ /Client\:\s+([^\s][\w]+)/) { # first line in show client
                $key = $1;
                next;
            } elsif ($n =~ /\S+\s+Catalog\:\s+([^\s][\w]+)/) { # first line in client catalog data
                $key = $1;
                $prefix = "cat_";
                next;
            } elsif ($n =~ /^\s+(\w+)$/) {
                $key = $1;
                next
            }

            if (defined($key) && $n =~ /([\S\s]+)\s+([^\s]+\w+)$/) {
                $client_prop->{"$prefix$key"} = $1;
                $key = $2;
                next;
            }

            if (defined($key)) {
               $client_prop->{"$prefix$key"} = $n;
               $key = undef;
            }
	}
    }
    return $client_prop;
}

sub get_cloud_status
{
    my ($self, $storage) = @_;

    my $arg = $self->get_form('storage');
    my $bconsole = $self->get_bconsole();
    my %status = (header => {}, uploads => [], downloads => []);
    my $state = '';
    my $context = '';
    my $device = '';
    my (%content, %transfers);
    my $output = $bconsole->send_cmd_api(".status storage=\"$arg->{storage}\" cloud");
    foreach my $l (@$output) {
        chomp($l);
        if ($l =~ /^(\S+)=([\S\s]+)$/) {
             if ($state eq 'uploads' || $state eq 'downloads') {
                 $content{$1} = $2;
                 $context = $state;
             } elsif ($state eq 'transfers') {
                 $transfers{$1} = $2;
             }
        } elsif ($l =~ /^(\S+):[\S\s]*$/) {
            $state = $1;
             if ($state !~ /^(uploads|downloads|transfers)$/) {
                 $device = $state;
            }
        } elsif ($l =~ /^$/ && $state eq 'transfers') {
            if (%transfers) {
                push @{$content{transfers}}, { %transfers };
                %transfers = ();
            }
        } elsif ($l =~ /^\]$/) {
            if (%transfers) {
                push @{$content{transfers}}, { %transfers };
                %transfers = ();
            }
	    if (!defined($arg->{device_name}) || $device ne $arg->{device_name}) {
                $content{device} = $device;
                push @{$status{$context}}, { %content };
            }
            %content = ();
            $state = '';
        }
    }
    return \%status;
}

# Get component (Storage, Director, File Daemon) version
# and release date.
# As param is taken component name 'Storage', 'Director' or 'FileDaemon'.
sub get_component_version
{
    my ($self, $component) = @_;

    my $ver = {demo => undef, version => undef, date => undef};
    if (defined($Services{$component})) {
        my $comp_bin = $Services{$component};
        my $bin_path = $self->get_location($comp_bin);

        if(defined($bin_path)) {
            if (-X $bin_path) {
                my @output = `LANG=C $bin_path/$comp_bin -? 2>&1`;

                for my $l (@output) {
                    if ($l =~ /^(Demo\s+)*Version:\s+([\d\.a-z]+)\s+\(([\S\s]+)\)\s*$/) {
                        $ver->{demo} = (defined($1)) ? 1 : 0;
                        $ver->{version} = $2;
                        $ver->{date} = $3;
                        last;
                    }
                }
            } else {
                $self->debug("Can't execute binary file '$Services{$component}' for component: '$component'.");
            }
        } else {
            $self->edebug("Can't get path of binary file '$Services{$component}' for component: '$component'.");
        }
    } else {
            $self->error("Unknown component name '$component'.");
    }

    return $ver;
}

# Get component options by component name and config resource name
# Params:
#  string $comp component name ('Director', 'Storage', 'FileDaemon', 'Console')
#  string $res resource type ('Pool', 'Device', 'Messages'...etc.)
# Resturns:
#  hash reference with options
sub get_component_options {
    my ($self, $comp, $res) = @_;
    my $result;
    if (defined($JSONTools{$comp})) {
        my $bin = $self->get_location($JSONTools{$comp});
        my $output = `$bin/$JSONTools{$comp} -r "$res" -D`;
	$result = JSON::decode_json($output);
    }
    return $result;
}

# Get component working directory
# Params:
#  string $comp component name ('Director', 'Storage', 'FileDaemon')
# Returns:
#  mixed string with working directory path or undef if there is not possible
#        to get any working directory value
sub get_component_workdir {
    my ($self, $comp) = @_;
    my $workdir;

    my $comp_opts = $self->get_component_options($comp, $comp);
    if (defined($comp_opts) && defined($comp_opts->[0]->{WorkingDirectory})) {
        $workdir = Encode::encode_utf8($comp_opts->[0]->{WorkingDirectory});
    } else {
        $self->edebug("Unale to get working directory for component '$comp'.");
    }
    return $workdir;
}

# Get Global Endpoint Deduplication usage statistics for specific storage.
sub get_sd_dedup_stats {
    my ($self, $storage) = @_;

    my $b = $self->get_bconsole();
    my $stats = $b->get_sd_dedup_stats($storage);
    return $stats;
}

sub dedup_stats {
    my ($self) = @_;
    $self->can_do("r_storage_mgnt");
    my $arg = $self->get_form('storage');

    my $b = $self->get_bconsole();
    my @storage_list = $b->list_storage();
    my $storages = [ map { { name => $_ } } @storage_list ];

    $self->display({
        'storages' => $storages,
        'default_storage' => $arg->{storage}
    }, 'dedup_stats.tpl');
}

# Copy single file from one place to another
# Params:
#  string $source_file source file path
#  string $dest_file destination file path
# Returns:
#  integer 0 - copying error, 1 - copied successfully
sub copy_single_file {
    my ($self, $source_file, $dest_file) = @_;
    my $result = copy($source_file, $dest_file);
    if ($result != 1) {
        $self->edebug("Error while copying file '$source_file' to '$dest_file' location. $!");
    }
    return $result;
}

sub unix_to_iso
{ 
    return strftime('%m-%d', localtime($_[0])) ;
}

# Convert ISO date to unix timestamp.
# Note, it works also for date without time for example '2016-06-07' (not ISO).
sub iso_to_unix
{
    my $iso_date = shift;
    my $unix_timestamp;

    if ($iso_date =~ /^(\d{4})-(\d{1,2})-(\d{1,2})( (\d{1,2}):(\d{2}):(\d{2}))?$/) {
        my $hour = 0;
        my $minute = 0;
        my $second = 0;
        if ($4) {
            $hour = $5;
            $minute = $6;
            $second = $7;
        }
        # do integer conversion and prepare month and year values
        # (years are in two digits form: 116 means 2016, months counted from 0 not from 1)
        my $year = sprintf("%d", $1) - 1900;
        my $month = sprintf("%d", $2) - 1;
        my $day = $3;
        $unix_timestamp = mktime($second, $minute, $hour, $day, $month, $year);
    }
    return $unix_timestamp;
}

#### CHECK FEATURES SUPPORT ####

# Check if in given component version is supported and possible
# to get TLS connection information
# Returns: 1 - supported, 0 - not supported
sub tls_support_check {
    my $component_version = shift;
    if (!$component_version) {
        return 0;
    }
    my $tls_support;
    my ($release, $major, $minor) = split(/\./, $component_version, 3);
    if (($release == $tls_support_version->{release} && $major >= $tls_support_version->{major}) ||
         $release > $tls_support_version->{release}) {
        $tls_support = 1;
    } else {
        $tls_support = 0;
    }
    return $tls_support;
}

# Check if Samba tools feature is supported.
# Returns: hashref with general support state (property: 'support') and
# support state for particular tools.
# {
#   support => 0/1,
#   net => 0/1,
#   smbclient => 0/1
# }
sub check_smb_support
{
    my $self = shift;
    my $support;
    my $net_support;
    my $smbclient_support;

    # Note, 'net' command returns exit code 255 for all calls which cause
    # showing help message. From this reason here is checked 'net' version
    # by -V parameter.
    my $output = `net -V 2>&1`;
    my $exitcode = $?;
    if ($exitcode == 0) {
        $net_support = 1;
    } else {
        $net_support = 0;
        # if binary was not executed (no binary file, permission problem ...) then
        # output is 'undef' value. Assing empty string for this case.
        $output ||= '';
        $self->debug("'net' command is not available. Output=$output, Exit=$exitcode");
    }

    $output = `smbclient -V 2>&1`;
    $exitcode = $?;
    if ($exitcode == 0) {
        $smbclient_support = 1;
    } else {
        $smbclient_support = 0;
        # if binary was not executed (no binary file, permission problem ...) then
        # output is 'undef' value. Assing empty string for this case.
        $output ||= '';
        $self->debug("'smbclient' command is not available. Output=$output, Exit=$exitcode");
    }

    $support = ($net_support && $smbclient_support) ? 1 : 0;

    my $ret = {
        support => $support,
        net => $net_support,
        smbclient => $smbclient_support
    };

    return $ret;
}

# Check if SSH function is supported.
# Returns: hashref with general support state (property: 'support') and
# support state for particular SSH functions.
# {
#   support => 0/1,
#   lib => 0/1,
#   passphrase => 0/1
# }
sub check_ssh_support
{
    my $self = shift;
    my $support;
    my $lib_support = 1;
    my $passphrase_support = 1;

    eval "use Net::SSH::Perl";
    if ($@) {
        $lib_support = 0;
        $self->debug("Net::SSH::Perl library is not available. Output=$@");
    }
    eval "use Net::SSH::Perl::Util::Term";
    if ($@) {
        $lib_support = 0;
        $self->debug("Net::SSH::Perl::Util::Term library is not available. Output=$@");
    } elsif (!defined(&Net::SSH::Perl::Util::Term::_read_passphrase)) {
        $passphrase_support = 0;
        $self->debug("No _read_passphrase() function in Net::SSH::Perl::Util::Term.");
    }
    eval "use Net::SSH::Perl::Key::RSA";
    if ($@) {
        $lib_support = 0;
        $self->debug("Net::SSH::Perl::Key::RSA library is not available. Output=$@");
    } elsif (!defined(&Net::SSH::Perl::Key::RSA::read_private)) {
        $passphrase_support = 0;
        $self->debug("No read_private() method in Net::SSH::Perl::Key::RSA.");
    }

    $support = ($lib_support && $passphrase_support) ? 1 : 0;

    my $ret = {
        support => $support,
        lib => $lib_support,
        passphrase => $passphrase_support
    };

    return $ret;
}

# Check if single item restore for VMware and Exchange is supported.
# Returns: hashref with general support state (property: 'support') and
# support state for particular single item restore types.
# {
#   support => 0/1,
#   vmware => 0/1,
#   exchange => 0/1
# }
sub check_single_item_restore_support
{
    my $self = shift;
    my $support;
    my $vmware_support = defined($self->get_location('mount-vmware'));
    my $exchange_support = defined($self->get_location('mount-exchange'));

    # This state says that both vmware and exchange restore are supported
    # or both are not supported.
    $support = ($vmware_support && $exchange_support) ? 1 : 0;

    my $ret = {
        support => $support,
        vmware => $vmware_support,
        exchange => $exchange_support
    };

    return $ret;
}

# Get support state for specific features.
# If no feature given, it checks and returns support for all defined features.
# Sample return:
# {
#   ssh => {
#     support => 0/1,
#     lib => 0/1,
#     passphrase => 0/1
#   },
#   smb => {
#     support => 0/1,
#     net => 0/1,
#     smbclient => 0/1
#   },
#   sir => {
#     support => 0/1,
#     vmware => 0/1,
#     exchange => 0/1
#   }
# }
# Param: array with features to check
# Returns: hashref with support states
sub get_support
{
    my ($self, @features) = @_;
    my $support = {};

    # TODO: Add support checking for TLS
    my %func = (
        ssh => sub { $self->check_ssh_support(); },
        smb => sub { $self->check_smb_support(); },
        sir => sub { $self->check_single_item_restore_support(); }
    );

    if (scalar(@features) > 0) {
        # check support for a few features
        for my $feature (@features) {
            if (defined($func{$feature})) {
                $support->{$feature} = $func{$feature}();
            }
        }
    } else {
        # check support for all features
        for my $feature (keys %func) {
            $support->{$feature} = $func{$feature}();
        }
    }

    return $support;
}

sub user_info
{
    my $self = shift;
    my $arg = $self->get_form('username');
    my $user;
    if ($arg->{username} eq $self->{loginname}) {
        $user = $self->{loginname};
    } elsif ($arg->{username} && $arg->{username} ne $self->{loginname} && $self->{loginname} eq 'admin') {
        $user = $arg->{username};
    }

    my $user_info = $self->get_user_info($user);
    if (!defined($user_info)) {
        $self->display({
            username => $arg->{username},
            error => 1
        },
        'user_info.tpl');
    } else {
        $self->display($user_info, 'user_info.tpl');
    }
}

# Execute command in shell.
# If $out_dir and $out_file given, then output is saved in file.
# If $out_dir given (without $out_file) then new temporary file is created to
# store command output (the file full path is visible in 'out' property in
# returned hashref.
#
# Params:
#  string $bin script/program full path to execute
#  string $out_dir directory to save $bin output
#  string $out_file file to save output
#  integer $stdout_only if set then to file is saved STDOUT output only
#  integer $background if set then $bin execution is done in background
#  string $file_template if set then this file template is used for new out file
# Returns:
#  hashref execution result (output, exitcode, error message, output file)
sub shell_exec {
    my ($self, $bin, $out_dir, $out_file, $stdout_only, $background, $file_template) = @_;
    my $out;

    if (defined($out_dir)) {
        if ($out_dir !~ /\/$/) {
            $out_dir .= '/';
        }
        if (!defined($out_file)) {
            unless($file_template) {
                $file_template = '';
	    }
            my ($fh, $tmp_file) = File::Temp::tempfile(DIR => $out_dir, TEMPLATE => "${file_template}XXXXX", SUFFIX => '.out');
            $out = $tmp_file; # It is file with full path
        } else {
            $out = $out_dir . $out_file;
        }
    }

    my $cmd;
    if ($out) {
        if (defined($stdout_only)) {
            $cmd = "$bin 1> '$out'";
        } else {
            $cmd = "$bin &> '$out'";
        }
    } else {
        $cmd = $bin;
    }

    if (defined($background)) {
        $cmd .= ' &';
    }
    my $output = `$cmd`;
    my $exitcode = $?;
    my $error = $!;
    my $ret = {
        output => $output,
        exitcode => $exitcode,
        error => $error,
        out => $out
    };
    return $ret;
}

# Display login page
sub display_system_login {
    my $self = shift;
    my $arg = $self->get_form('username');
    my $error = 0;
    if ($arg->{username} && $self->is_system_user_logged() == 0) {
        $error = 1;
    }
    $self->display({
        error => $error,
        ver => $self->get_ver()
    }, 'system_auth.tpl');
}

# Display page after log out
sub display_system_logout {
    my $self = shift;
    my $session = $self->get_session();
    $session->param("login", 0);
    my @to_clear = ("user", "uid", "gid", "registered", "user_res");
    $session->clear(\@to_clear);
    $self->display({
        ver => $self->get_ver()
    }, 'system_auth.tpl');
}

# Check if system user is logged in
# Returns:
#  1 - user is logged in, 0 - user isn't logged in
sub is_system_user_logged {
    my $self = shift;
    my $success = 0;
    my $login = $self->get_session()->param("login");

    if ($self->is_system_auth() == 0 || (defined($login) && $login == 1)) {
        $success = 1;
    }
    return $success;
}

# Check user and password.
# Here is try to log in to system.
# @TODO: Protect password (env variable ?)
# Retruns:
#  0 - user and password are valid, 0< - user and password are invalid
sub validate_system_user {
    my ($self, $user, $password) = @_;
    my $ret = 1;
    if ($self->is_system_auth() == 1) {
        my $msg;
        my $result = {error => '', exitcode => 0};
        if ($> == 0) {
            # root doesn't use password to switch to a user
            $msg = "ERROR: The system user auth doesn't support using web server as 'root' user.";
            $self->edebug($msg);
        } else {
            my $cmd_name = 'su';
            my $cmd_path = $self->get_location($cmd_name);
            my $cmd = "$cmd_path/$cmd_name -l $user";
            my $cmd_expect = new Expect;
            unless ($cmd_expect->spawn($cmd)) {
                $msg = "ERROR: Cannot spawn 'su' command: $!";
                $self->edebug($msg);
                $result->{exitcode} = 1;
                $result->{error} = $msg;
            } else {
                $cmd_expect->log_stdout(0);
                $cmd_expect->expect(5,
                [qr'Password:'i, sub {
                    my $h = shift;
                    print $h "$password\n\r";
                    $self->debug("Try to 'su' login using user/password.\n");
                    exp_continue;
                }],
                [EOF => sub {
                    my $h = shift;
                    $msg = "ERROR: Could not login by 'su', unexpected EOF occurred!\n";
                    $self->edebug($msg);
                    $result->{exitcode} = 1;
                    $result->{error} = $msg;
                }],
                [timeout => sub {
                    my $h = shift;
                    $msg = "ERROR: Could not login by 'su', timeout occurred!\n";
                    $self->edebug($msg);
                    $result->{exitcode} = 1;
                    $result->{error} = $msg;
                }],
                [qr'[$>:~] $'i => sub {
                    my $h = shift;
                    $msg = "Login OK.\n";
                    $self->debug($msg);
                    $result->{exitcode} = 0;
                    $result->{error} = $msg;
                }],
                '-re', qr'[#>:~] $',
                );
                my $exitcode = $cmd_expect->exitstatus();
                if ($result->{exitcode} == 0 && defined($exitcode) && $exitcode != 0) {
                    $result->{exitcode} = $exitcode;
                    $result->{error} .= $cmd_expect->before();
                }
                $cmd_expect->hard_close();
            }
        }
        $ret = $result->{exitcode};
        if ($result->{exitcode} != 0) {
            $self->edebug("ExitCode=$result->{exitcode}, ErrorMsg=$result->{error}");
        }
    }
    return $ret;
}

# Authenticate user in system.
# Login params are stored in CGI session.
# Returns:
#  1 - auth finished successfully, 0 - auth error
sub system_auth {
    my $self = shift;
    my $success = 0;
    my $arg = $self->get_form(qw/username password/);
    my $user = defined($arg->{username}) ? $arg->{username} : '';
    my $pwd = defined($arg->{password}) ? $arg->{password} : '';
    my $session = $self->get_session();

    unless ($self->validate_system_user($user, $pwd)) {
        my $user_info = $self->get_user_info($user);
        my $registered = defined($self->get_user($user)) ? 1 : 0;
        $session->param("login", 1);
        $session->param("user", $user);
        $session->param("group", $user_info->{group});
        $session->param("uid", $user_info->{uid});
        $session->param("gid", $user_info->{gid});
        $session->param("registered", $registered);
        $success = 1;
    } elsif ($self->is_system_auth() == 1) {
        $session->param("login", 0);
        my @to_clear = ("user", "group", "uid", "gid", "registered", "user_res");
        $session->clear(\@to_clear);
    }
    return $success;
}

# Get user information
# Returns hashref with info or undef if user doesn't exist or problem with user access
sub get_user_info
{
    my ($self, $user) = @_;
    my $ret;
    my @params;

    if ($user) {
        @params = getpwnam($user);
    }

    if (scalar(@params) > 0) {
        my ($username, $passwd, $uid, $gid, $quota, $comment, $gcos, $dir, $shell, $expire) = @params;
        my $group = getgrgid($gid);
        $ret = {
            username => $username,
            group => $group,
            uid => $uid,
            gid => $gid,
            quota => $quota,
            comment => $comment,
            gcos => $gcos,
            dir => $dir,
            shell => $shell,
            expire => $expire
        };
    }
    return $ret;
}

# Perform system user authenticate.
# Returns:
#  1 - auth OK, 0 - no auth
sub authenticate {
    my ($self, $action) = @_;

    # Authenticate system user
    my $auth = 0;
    if (defined($action) && $$action eq 'system_logout') {
        # Logout
        $self->display_system_logout();
        $auth = 0;
    } elsif ($self->is_system_auth() == 0 || $self->is_system_user_logged()) {
        # Already logged
        $auth = 1;
    } elsif ((!defined($action) || $$action ne 'begin') && $self->system_auth()) {
        # User just logged in
        $auth = 1;
        if ($$action eq 'system_auth') {
            $$action = 'begin';
        }
    } else {
        # Login form
        $self->display_system_login();
        $auth = 0;
    }
    return $auth;
}

sub component_network_test
{
    my ($self, $client, $storage, $bytes) = @_;
    my $b = $self->get_bconsole();
    my $ret = $b->component_network_test($client, $storage, $bytes);
    return $ret;
}

# Command from web console passthrough to bconsole
sub bconsole_command
{
   my ($self, $command) = @_;
   $self->can_do("r_bconsole");
   my $ret;
   if ($self->{info}->{enable_web_bconsole}) {
     my $b = $self->get_bconsole();
     $ret = $b->send_one_cmd($command);
   }
   return $ret;
}

1;

#####################################################################

package Bweb::Lang;

use base q/Bweb::Gui/;

=head1 PACKAGE

    Bweb::Lang - read and use translations files

=head2 DESCRIPTION

    This package is for using translations texts from BWeb server side.

=head2 USAGE

    $lang = new Bweb::Lang(lang_file => '/path/to/lang_file');
    $lang->load();

    $lang->get('Hello world!');
    $lang->get('Some text written in English...');

=head2 NOTES

    lang_file:
     It points to the translated language file in currently used language.
     This file is in Data::Dumper format and loaded is by load() method.

    In future it is possible to use Locale::gettext Perl module for read
    particular texts directly from *.po files. Currently is used pre-defined
    template file way.

=cut

# Translations file for server side texts.
# The file exists in every supported language and is generated together with
# all templates. Its format is Data::Dumper format.
our $lang_filename = "translations.tpl";

# Loading translations file.
sub load {
    my ($self) = @_;

    my $lang_texts = "";
    if (open(FH, "<$self->{lang_file}")) {
        local $/;
        $lang_texts = <FH>;
        close(FH);
    } else {
        $self->edebug("ERROR: Problem while opening lang file: ".
        "'$self->{lang_file}' $!\n");
    }

    # I did not find any easier way to avoid throwing warning when eval used
    # in strict mode. Here strict is temporary disabled on eval time.
    no strict;
    $self->{translation} = eval "$lang_texts";
    use strict;
}

# Translate a server side text.
# Main method used for translate texts from server side.
# For add new texts, there is need to add it to $lang_filename
sub get {
    my ($self, $msgid) = @_;
    my $msgstr;

    if (defined($self->{translation}->{$msgid})) {
        $msgstr = $self->{translation}->{$msgid};
    } else {
        $self->edebug("ERROR: Translations text with msgid: '$msgid'".
        " did not find in $self->{info}->{lang_file}. Used original text.");
        $msgstr = $msgid;
    }
    return $msgstr;
}
1;

#####################################################################

package Bweb::FilesBrowser;

use Encode;
use File::stat;
use Fcntl ':mode';
use base q/Bweb::Gui/;

=head1 PACKAGE

    Bweb::FilesBrowser - list files and directories from local filesystems

=head2 DESCRIPTION

    This package is for getting and listing real files and their attributes.

=head2 USAGE

    $lang = new Bweb::FilesBrowser(path => '/path/to/listing', offset => 10, limit => 50);
    $lang->list();

=head2 NOTES

    path:
     It points to the path which files and directories will be listed.

=cut

# Document root for mounted filesystem.
# Note, for root path access /, set empty string (use carefully).
our $document_root = '';

# Offset used as default when no offset given
my $default_files_offset = 0;

# Limit used as default when no limit given
my $default_files_limit = 500;

# Load directory content (file and directory names)
# Returns:
#  array reference with current directory content ('bin', 'boot', 'etc'...etc.)
sub load {
    my $self = shift;

    my @files;

    unless ($self->{document_root}) {
        $self->{document_root} = $document_root;
    }
    unless ($self->{offset}) {
        $self->{offset} = $default_files_offset;
    }

    unless ($self->{limit}) {
        $self->{limit} = $default_files_limit;
    }

    # cut of last slash (if exist)
    if ($self->{path} =~ /(.*)\/$/) {
        $self->{path} = $1;
    }

    # Add document root
    $self->{path} = "$self->{document_root}$self->{path}";

    if (opendir(DIR, "$self->{path}/")) {
        @files = map { Encode::decode_utf8($_); } readdir(DIR);
        closedir(DIR);

        # sort before splice
        @files = sort(@files);

        # apply offset and limit
        @files = splice(@files, $self->{offset}, $self->{limit});
    } else {
        $self->edebug("ERROR: Problem while opening directory: ".
        "'$self->{path}' $!\n");
    }
    return \@files;
}

# List detailed directory content
# Returns:
#  return array reference contains files and dirs data
sub list {
    my $self = shift;

    my @content_details;
    my $detail;
    my $content = $self->load();
    for (my $i = 0; $i < scalar(@{$content}); $i++) {
        $detail = $self->get_stat_info($content->[$i]);
        if (!defined($detail)) {
            next;
        }
        push @content_details, $detail;
    }
    return \@content_details;
}

=head1 FUNCTION

    get_stat_info - Get stat information for specific file

=head2 DESCRIPTION

 Params:
  filename - single file name (without path)
 Returns:
  hash reference with detailed file data like 'name', 'mtime', 'size'...etc.

 Example return:
 {
     "name": ".thumbnails",
     "size": 4096,
     "mtime": 1418657845,
     "ctime": 1418657845,
     "atime": 1439736308,
     "isdir": 1,
     "islink": 0,
     "uid": 1000,
     "gid": 1000,
     "dev": 2065,
     "rdev": 0,
     "blksize": 4096,
     "blocks": 8,
     "nlink": 3,
     "ino": 681653
 }

=cut

sub get_stat_info {
    my ($self, $filename) = @_;
    my $st = lstat($self->{path} . '/' . Encode::encode_utf8($filename));
    my $ret;
    if (defined($st)) {
        $ret = {
            name => $filename,
            size => $st->size,
            atime => $st->atime,
            mtime => $st->mtime,
            ctime => $st->ctime,
            isdir => S_ISDIR($st->mode),
            islink => S_ISLNK($st->mode),
            nlink => $st->nlink,
            blksize => $st->blksize,
            blocks => $st->blocks,
            dev => $st->dev,
            ino => $st->ino,
            uid => $st->uid,
            gid => $st->gid,
            rdev => $st->rdev
        };
    } else {
        $self->edebug("There is not possible to get stat() attribues from file '$self->{path}/$filename'.");
    }
    return $ret;
}

1;
#####################################################################

package Bweb::SessionKey;

use File::Basename;
use File::stat;
use Digest::MD5 qw/md5_hex/;
use base q/Bweb::Gui/;

=head1 PACKAGE

    Bweb::SessionKey - generate a token for a file

=head2 DESCRIPTION

    This package is responsible for prepare tokens for files shared via HTTP(S) protocol

=head2 USAGE

    $sessionkey = new Bweb::SessionKey();
    # generate and save token
    $sessionkey->generate_token('/path/to/file');
    # verify token with filepath
    $sessionkey->verify_token('/path/to/file', 'sometoken');

=head2 NOTES

    path:
     It points to the path which files and directories will be listed.

=cut

# Default directory to storing session key files.
# Note, it is used only when no SD and DIR working directory found.
# If exists SD or DIR working directory, then it is used to storing
# session key data.
our $session_key_dir = '/tmp';

# Subdirectory in which are stored session key files.
# If not exist, it is created.
our $session_key_subdir = 'bsessions';

# Extension for session file
our $session_file_extention = '.bsess';

# Session index file to store files share properties
our $index_file = 'sessions.index';

# Session index fields (in this order)
our @index_fields = qw/file_name file_path user_name last_download_timestamp downloads_count/;

# Get session directory for storing session files.
# Returns:
#  string path to session directory to store session data
sub get_session_dir {
    my $self = shift;

    # For using session is required one Storage Daemon or Director component
    # from BWeb local side. Here is probed working dir (first for SD, next for DIR)
    # and if no working directory found then is used default working directory
    # from $session_key_dir value.
    my $workdir = $self->{bweb}->get_component_workdir('Storage');
    unless ($workdir) {
       $workdir = $self->{bweb}->get_component_workdir('Director');
    }

    if (defined($workdir)) {
        $workdir = "$workdir/$session_key_subdir";
    } else {
        # No SD and DIR working dir found, using default directory
        $workdir = "$session_key_dir/$session_key_subdir";
    }

    unless(-d $workdir) {
        unless (mkdir($workdir, 0700)) {
            $self->edebug("Problem while creating working subdirectory '$workdir': $!");
        }
    }

    return $workdir;
}

# Generate random session key that will be used to prepare a token.
# Session keys are stored in session key directory.
sub generate_session_key {
    my $self = shift;
    my @session_key = map{('a'..'z','A'..'Z',0..9)[rand 62]} 0..32;
    my $key = join('', @session_key);
    return $key;
}

# Generate token by using session key.
# Token is generated by crypt(file_attrs, session_key).
# Please note that system crypt() function has no decrypt function.
# Tokens are not stored in any file.
# Params:
#  string $file_path path to file for which token will be generated
#  integer $expiration_time expiration session unix timestamp (with miliseconds part)
# Returns:
#  string new token value
sub generate_token {
    my ($self, $file_path, $expiration_time, $remove_after_download) = @_;
    my $salt = $self->get_stat_salt($file_path);
    my $session_key = $self->generate_session_key();
    my $token = crypt($salt, $session_key);
    $self->save_session_data($token, $session_key, $expiration_time, $remove_after_download);
    return $token;
}

# Save generated session key in session key directory.
# File name in which is stored session key is taken from token hash in next way:
# md5_hex(token) -> file name to storing session key
# Params:
#  string $token token value
#  string $session_key session key to storing
#  integer $expiration_time unix timestamp after which session will be expired
sub save_session_data {
    my ($self, $token, $session_key, $expiration_time, $remove_after_download) = @_;
    my $keyfile = md5_hex($token);
    my $session_dir = $self->get_session_dir();
    if (open(FP, ">$session_dir/${keyfile}${session_file_extention}")) {
        print FP "$session_key:$expiration_time:$remove_after_download";
        close(FP);
    } else {
            $self->edebug("Problem during saving session key in key file '$session_dir/${keyfile}${session_file_extention}'. $!");
    }
}

# Read session key from file.
# Return session key string
# Params:
#  string $token token value
#  integer $is_token_hash (optional) determine if $token param is given as raw token or as hash,
#                          0 - raw token string given, 1 - hash token given
# Returns:
#  hashref session key with its expiration time {session_key => 'somesesskey', expiration_time => 123456789}
sub read_session_data {
    my ($self, $token, $is_token_hash) = @_;
    my $result;
    my $session_key;
    my $expiration_time;
    my $remove_after_download;
    my $keyfile = defined($is_token_hash) ? $token : md5_hex($token);
    my $session_dir = $self->get_session_dir();
    if ( -e "$session_dir/${keyfile}${session_file_extention}") {
        if (open(FP, "<$session_dir/${keyfile}${session_file_extention}")) {
            my $session_data = <FP>;
            chomp($session_data);
            $session_key = $1 if $session_data =~ /^(\S+)\:\d+:[0-1]{1}$/;
            $expiration_time = $1 if $session_data =~ /^\S+\:(\d+):[0-1]{1}$/;
            $remove_after_download = $1 if $session_data =~ /^\S+\:\d+:([0-1]{1})$/;
            if ($expiration_time) {
                #conversion to integer
                $expiration_time *= 1;
            }
            close(FP);
        } else {
            $self->edebug("Problem with reading key file '$session_dir/${keyfile}${session_file_extention}' $!.");
        }
    } else {
            $self->edebug("Key file does not exist '$session_dir/${keyfile}${session_file_extention}'.");
    }
    if (defined($session_key) && defined($expiration_time)) {
        $result = {
            session_key => $session_key,
            expiration_time => $expiration_time,
            remove_after_download => $remove_after_download
        };
    }
    return $result;
}

# List opened/active sessions data.
#
# Returns array of hashes contains following fields:
# [
#  {
#   token => 'sometoken',
#   file => 'abc.txt',
#   file_dir => '/home/aaa',
#   expiration_time => 123456789
#  },
#  {
#   token => 'someanothertoken',
#   file => 'zzz.png',
#   file_dir => '/usr/share/images/',
#   expiration_time => 987654321
#  }
#  ...etc.
# ]
#
# Params:
#  string $files_dir location storing data files for which sessions are created
# Returns:
#  array reference in result is given array of hashes (each hash is one session item).
sub list_sessions {
    my ($self, $files_dir) = @_;

    my $session_dir = $self->get_session_dir();
    my @files_session_data;
    my @data_files;
    my @session_files;
    my ($salt, $session_key_data, $token_hash);

    if (opendir(DIR, $files_dir)) {
        @data_files = sort(grep(!/^(\.\.?)$/, readdir(DIR)));
        closedir(DIR);
    } else {
        $self->edebug("ERROR: Problem while opening directory: '$files_dir': $!");
    }

    if (opendir(SESSDIR, $session_dir)) {
        @session_files = sort(grep(!/^(\.\.?)$/, readdir(SESSDIR)));
        closedir(SESSDIR);
    } else {
        $self->edebug("ERROR: Problem while opening session directory: '$session_dir': $!");
    }

    DATADIR_LOOP: for (my $i = 0; $i < scalar(@data_files); $i++) {

        $salt = $self->get_stat_salt("$files_dir/$data_files[$i]");

        SESSDIR_LOOP: for (my $j = 0; $j < scalar(@session_files); $j++) {
            if ($session_files[$j] =~ /^([a-fA-F0-9]+)\Q$session_file_extention\E$/) {
                $token_hash = $1;
            } else {
                next SESSDIR_LOOP;
            }

            $session_key_data = $self->read_session_data($token_hash, 1);
            if (defined($session_key_data)) {
                my $token = crypt($salt, $session_key_data->{session_key});
                my $user_token_hash = md5_hex($token);
                if ($user_token_hash eq $token_hash) {
                    my $sess_item = {
                        token => $token,
                        file => $data_files[$i],
                        file_dir => "${files_dir}/",
                        expiration_time => $session_key_data->{expiration_time}
                    };
                    push @files_session_data, $sess_item;
                    last SESSDIR_LOOP;
                }
            }
        }
    }
    return \@files_session_data;
}

# Remove session file.
# Params:
#  string $token token value for which session key will be removed
sub remove_session_file {
    my ($self, $token) = @_;
    my $session_key;
    my $keyfile = md5_hex($token);
    my $session_dir = $self->get_session_dir();
    if ( -e "$session_dir/${keyfile}${session_file_extention}") {
        unlink("$session_dir/${keyfile}${session_file_extention}");
    } else {
            $self->edebug("Key file does not exist '$session_dir/${keyfile}${session_file_extention}'.");
    }
}

# Get specific for each file salt.
# In this case to determine salt there are used concatenated lstat() values.
# Params:
#  string $file_path file path to get stat salt
# Returns:
#  string stat salt value
sub get_stat_salt {
    my ($self, $file_path) = @_;

    my $st = lstat("$file_path");

    if (!defined($st)) {
        $self->edebug("There is not possible to get stat() attribues from file '$file_path'.");
	return;
    }

    my $salt = $st->size . $st->atime . $st->mtime . $st->ctime . $st->mode . $st->nlink .
                $st->blksize . $st->blocks . $st->dev . $st->ino . $st->uid . $st->gid . $st->rdev;

    return $salt;
}

# Verify matching token and file path.
# By this function is possible to determine if token is valid.
# If token is compatible with file path, then return 1, otherwise 0.
# Params:
#  string $file_path path to file for which token will be verified
#  string $token token value to verify
# Returns:
#  integer 0 - token invalid, 1 - token valid
sub verify_file_token {
    my ($self, $file_path, $token) = @_;
    my $valid = 0;
    if (! -e "$file_path") {
        $self->edebug("File does not exist '$file_path'.");
        return 0;
    }

    my $salt = $self->get_stat_salt($file_path);

    my $session_key_data = $self->read_session_data($token);
    if (defined($session_key_data)) {
        my $orig_token = crypt($salt, $session_key_data->{session_key});
        if ($orig_token eq $token) {
            $valid = 1;
        } else {
            $self->edebug("Ivalid token given '$token'.");
        }
    }
    return $valid;
}

# Remove single session data and remove related file (file to restore)
# with session.
# Params:
#  string $file_path path to file to remove
#  string $token token value from session to remove
sub remove_session_data_and_file {
    my ($self, $file_path, $token) = @_;

    $self->remove_session_file($token);

    unless (unlink($file_path)) {
        $self->edebug("Remove file '$file_path' failed: $!");
    }
}

# Remove all sessions (clean up) and all related data with these sessions.
# Params:
#  string $files_dir directory path that stores data files protected by session
#  integer $only_expired 0 - remove all session data and related files, 1 - remove
#                        only expired sessions.
sub remove_all_sessions_and_files {
    my ($self, $files_dir, $only_expired) = @_;

    my $sessions = $self->list_sessions($files_dir);

    # compute in timestamp format date prevoius day at 23:59
    my $time_now = (sprintf("%.0f", (time() / 86400)) * 86400) - 1;
    my ($file_path, $remove);

    for my $sess (@{$sessions}) {
        $remove = 0;
        $file_path = "$sess->{file_dir}$sess->{file}";
        if ($only_expired) {
            if ($sess->{expiration_time} < $time_now) {
                $remove = 1;
            }
        } else {
            $remove = 1;
        }

        if ($remove == 1) {
            $self->remove_session_data_and_file($file_path, $sess->{token});
        }
    }
}

# Add single item to index (download history)
# Params:
#  array @item_values list of values according to @index_fields array
#                     They are: ('file_name', 'file_path', 'user_name',
#                     'last_download_timestamp', 'download_counts'...)
sub add_index_item {
    my ($self, @item_values) = @_;

    my $item = join(',', @item_values);

    my $session_dir = $self->get_session_dir();
    my $index_path = "${session_dir}/${index_file}";
    if (open(FH, ">>$index_path")) {
        print FH "$item\n";
        close(FH);
    } else {
        $self->edebug("Error while opening '$index_path' file: $!.");
    }
}

# Read and get all index items.
# Returns:
#  array of arrays items list with each item in separate array inside one
#                  main array
sub read_index_items {
    my $self = shift;
    my @index_items;
    my $item;

    my $session_dir = $self->get_session_dir();
    my $index_path = "${session_dir}/${index_file}";
    if (open(FH, "<$index_path")) {
        while (my $line = <FH>) {
            chomp($line);

            my @item =  split(',', $line, scalar(@index_fields));
            push @index_items, \@item;
        }
        close(FH);
    } else {
        $self->edebug("Error while opening '$index_path' file: $!.");
    }
    return \@index_items;
}

# Get single index item by item name and item value
# Params:
#  string $item_name name in index (for example: 'file_name' or 'user_name'...)
#  string $item_value value for name from first param (for example: 'abc.tar.gz,
#                     or 'admin'...)
# Returns:
#  array of arrays list all items matched to given parameters ($item_name
#                  and $item_value)
sub get_index_item {
    my ($self, $item_name, $item_value) = @_;
    my @result;

    my $items = $self->read_index_items();
    my $item_pos = -1;

    for my $item (@{$items}) {
        ($item_pos) = grep { $index_fields[$_] eq $item_name} 0..$#index_fields;
        if ($item_pos > -1 && defined($item_value) && $item->[$item_pos] ne $item_value) {
            next;
        }
        push @result, $item;
        $item_pos = -1;
    }
    return \@result;
}

# Check if item index with given item name and value exists.
# Params:
#  string $item_name name in index (for example: 'file_name' or 'user_name'...)
#  string $item_value value for name from first param (for example: 'abc.tar.gz,
#                     or 'admin'...)
# Returns:
#  integer 1 - index item exists, 0 - index item does not exists
sub index_item_exists {
    my ($self, $item_name, $item_value) = @_;

    my $items = $self->get_index_item($item_name, $item_value);
    my $exists = (scalar(@{$items}) == 1) ? 1 : 0;
    return $exists;
}

# Truncate index file to zero value
sub truncate_index_file {
    my $self = shift;

    my $session_dir = $self->get_session_dir();
    my $index_path = "${session_dir}/${index_file}";

    if (open(FH, ">$index_path")) {
        print FH '';
        close(FH);
    } else {
        $self->edebug("Error while opening '$index_path' file: $!.");
    }

}

# Update/set single item value
# Note, index file is rewrited from scratch.
# TODO: Add index file locking
#  string $item_name name in index (for example: 'file_name' or 'user_name'...)
#  string $item_value value for name from first param (for example: 'abc.tar.gz,
#                     or 'admin'...)
#  string $update_item item name to update in matched row from first two params
#  string $update_value value to update in matched row from first two params
sub update_index_item {
    my ($self, $item_name, $item_value, $update_item, $update_value) = @_;

    my $item = $self->get_index_item($item_name, $item_value);
    my $items = $self->read_index_items();
    my ($update_item_pos) = grep { $index_fields[$_] eq $update_item} 0..$#index_fields;

    if (scalar(@{$item}) > 0) {
        for (my $i = 0; $i < scalar(@{$item}); $i++) {
            $item->[$i]->[$update_item_pos] = $update_value;
        }

        # Truncate index file to zero size (empty file)
        $self->truncate_index_file();

        # Recreate index file with updated values
        for (my $i = 0; $i < scalar(@{$items}); $i++) {
            for (my $j = 0; $j < scalar(@{$item}); $j++) {
                if ($item->[$j]->[0] eq $items->[$i]->[0]) {
                    $self->add_index_item(@{$item->[$j]});
                } else {
                    $self->add_index_item(@{$items->[$i]});
                }
            }
        }
    }
}

# Remove single item value
# Note, index file is rewrited from scratch.
#  string $item_name name in index (for example: 'file_name' or 'user_name'...)
#  string $item_value value for name from first param (for example: 'abc.tar.gz,
#                     or 'admin'...)
sub remove_index_item {
    my ($self, $item_name, $item_value) = @_;

    my $item = $self->get_index_item($item_name, $item_value);
    my $items = $self->read_index_items();

    if (scalar(@{$item}) > 0) {
        # Truncate index file to zero size (empty file)
        $self->truncate_index_file();

        # Recreate index file with updated values
        for (my $i = 0; $i < scalar(@{$items}); $i++) {
            for (my $j = 0; $j < scalar(@{$item}); $j++) {
                if ($item->[$j]->[0] eq $items->[$i]->[0]) {
                    next;
                }
                $self->add_index_item(@{$items->[$i]});
            }
        }
    }
}

1;

#####################################################################

package Bweb::ExchangeAccess;

use base q/Bweb::Gui/;

use JSON;

=head1 PACKAGE

    ExchangeAccess - Base package for calling mount-exchange

=head2 DESCRIPTION

 The package make make available interface to direct work on
 mount-exchange.

=head2 USAGE

    $exchange_access = new Bweb::ExchangeAccess(bweb => $bweb_obj);
    $exchange_access->mount_action(1);
    my @params = ('Client=127.0.0.1-fd');
    $exchange_acces->mount_action(2, @params);
    ...etc.

=cut

# Workdir path taken when is not possible to determine
# Storage Daemon working directory
my $default_exchange_workdir = '/tmp';

# Subdir is using for separating restore exchange sessions data.
# It enables to use exchange restore by many admins at the same time.
my $exchange_subdir_prefix = 'mount-exchange-';

# Run mount-exchange script in background.
# If script is running in background then getting its output is possible
# by persistent content feature (@see PersistentContent).
my $exchange_run_in_background = 1;

# Get the current data step
#  string $session_id CGI session identifier
#  array @cmd_params current command parameters
sub mount_action {
    my ($self, $session_id, @cmd_params) = @_;

    # Always call request about Catalog name
    my $result = $self->mount_exchange($session_id);

    if ($result->{exitcode} == 0) {
        $result = $self->parse_output($result->{output});
        # Temporary take first Catalog name
        # TODO: Support for more catalogs than one
        push(@cmd_params, "Catalog=$result->[0]");

        $result = $self->mount_exchange($session_id, \@cmd_params, $exchange_run_in_background);
        if ($exchange_run_in_background && defined($result->{out})) {
            # New persistent content result
            $result = $self->{content}->new_identity('file', $result->{out}, 'exchange_restore');
        } else {
            $result = $self->parse_output($result->{output});
        }
    }
    return $result;
}

# Prepare and parse mount Exchange resources output.
#  string $output mount-exchange command output
sub parse_output {
    my ($self, $output) = @_;
    my $result = {};

    # decode_json() in eval because it exits (by croak) from script if it is not
    # able to decode string. In this place there is no sure that output will be
    # decoded properly (errors are checked after decoding).
    my $obj = eval { JSON::decode_json($output) };
    if (!$@ && exists($obj->{Error})) {
        # No error in content. Content valid.
        $result = exists($obj->{Data}) ? $obj->{Data} : $obj;
    } else {
        # Error detected in output content.
        my $errmsg;
        if ($obj && exists($obj->{ErrMsg})) {
            my $type = exists($obj->{Type}) ? $obj->{Type} : 'No type';
            # Error != 0 or no Error property
            my $errorcode = exists($obj->{Error}) ? $obj->{Error} : 'No errorcode';
            $errmsg = "Error while getting data for Type: %s, ErrorCode: %s, ErrorMsg: %s.";
            $errmsg = sprintf($errmsg, $type, $errorcode, $obj->{ErrMsg});
        } else {
            # Not possible to parse output (non-JSON string)
            $errmsg = "Unknown output received while getting data. Output=$output.";
        }

        if (!$obj && $exchange_run_in_background) {
            # In background mode the output can be invalid because it can be incomplete
            # up till time when process in background becomes finished. To avoid
            # sending many false error messages in log, the error messages are visible
            # in logs only when debug is enabled.
            $self->debug($errmsg);
        } else {
            $self->edebug($errmsg);
        }
    }
    return $result;
}

# Execute mount Exchange backup commands via mount-exchange interface.
# Params:
#  string $session_id CGI session identifier
#  arrayref $params parameters given to command line
#  integer $background if set, run script in background, otherwise run script normally (in foreground)
# Returns:
#  hashref script execution result (output, exitcode, error message...)
sub mount_exchange {
    my ($self, $session_id, $params, $background) = @_;
    my $ret;

    my $sir_support = $self->{bweb}->get_support(qw/sir/);
    if ($sir_support->{sir}->{exchange}) {
        my $mount_script = $self->{bweb}->get_location('mount-exchange');

        my $mount_params = '';
        if (defined($params) && ref($params) eq "ARRAY") {
            $mount_params = '"' . join('" "', @{$params}) . '"';
        }
        my $workdir = $self->get_exchange_workdir($session_id);
        my $bindir = $self->{bweb}->get_location('bacula-sd');
        my $confdir = $self->{bweb}->get_location('bacula-sd.conf');

        my $cmd = "$mount_script/mount-exchange -w \"$workdir\" -b \"$bindir\" -C \"$workdir/mount-cache\" -c \"$confdir\" -j $mount_params";
        if ($background) {
            $ret = $self->{bweb}->shell_exec($cmd, $workdir, undef, 1, 1, $session_id);
            if ($ret->{exitcode} != 0) {
                $self->edebug("Error while background executing $cmd. Error=$ret->{error}, Output=$ret->{output}.");
            }
        } else {
            $ret = $self->{bweb}->shell_exec($cmd);
            if ($ret->{exitcode} != 0) {
                $self->edebug("Error while direct executing $cmd. Error=$ret->{error}, Output=$ret->{output}.");
            }
        }
    }
    return $ret;
}

# Get Exchange restore working directory.
# The working directory is prepared per session. It means that there can be opened
# many restore sessions by different users.
# Params:
#  string $session_id CGI session identifier
# Returns:
#  string Exchange restore working directory
sub get_exchange_workdir {
    my ($self, $session_id) = @_;
    my $workdir = $self->{bweb}->get_component_workdir('Storage');
    if (!$workdir || ! -d $workdir) {
        # No SD working dir found, using default working directory
        $workdir = $default_exchange_workdir;
    }
    $workdir .= "/${exchange_subdir_prefix}${session_id}";
    return $workdir;
}

1;

#####################################################################

package Bweb::VMwareAccess;

use base q/Bweb::Gui/;

use JSON;

=head1 PACKAGE

    VMwareAccess - Base package for calling mount-vmware

=head2 DESCRIPTION

 The package make make available interface to direct work on
 mount-vmware.

=head2 USAGE

    $vmware_access = new Bweb::VMwareAccess(bweb => $bweb_obj);
    $vmware_access->mount_action(1);
    my @params = ('Client=127.0.0.1-fd');
    $vmware_acces->mount_action(2, @params);
    ...etc.

=cut

# Workdir path taken when is not possible to determine
# Storage Daemon working directory
my $default_vmware_workdir = '/tmp';

# Subdir is using for separating restore vmware sessions data.
# It enables to use vmware restore by many admins at the same time.
my $vmware_subdir_prefix = 'mount-vmware-';

# Run mount-vmware script in background.
# If script is running in background then getting its output is possible
# by persistent content feature (@see PersistentContent).
my $vmware_run_in_background = 1;

# Get the current data step
#  string $session_id CGI session identifier
#  array @cmd_params current command parameters
sub mount_action {
    my ($self, $session_id, @cmd_params) = @_;

    # Always call request about Catalog name
    my $result = $self->mount_vmware($session_id);

    if ($result->{exitcode} == 0) {
        $result = $self->parse_output($result->{output});

        # Temporary take first Catalog name
        # TODO: Support for more catalogs than one
        push(@cmd_params, "Catalog=$result->[0]");

        $result = $self->mount_vmware($session_id, \@cmd_params, $vmware_run_in_background);
        if ($vmware_run_in_background && defined($result->{out})) {
            # New persistent content result
            $result = $self->{content}->new_identity('file', $result->{out}, 'vmware_restore');
        } else {
            $result = $self->parse_output($result->{output});
        }
    }
    return $result;
}

# Prepare and parse mount VMware resources output.
#  string $output mount-vmware command output
sub parse_output {
    my ($self, $output) = @_;
    my $result = {};

    # decode_json() in eval because it exits (by croak) from script if it is not
    # able to decode string. In this place there is no sure that output will be
    # decoded properly (errors are checked after decoding).
    my $obj = eval { JSON::decode_json($output) };
    if (!$@ && exists($obj->{Error})) {
        # No error in content. Content valid.
        $result = exists($obj->{Data}) ? $obj->{Data} : $obj;
    } else {
        # Error detected in output content.
        my $errmsg;
        if ($obj && exists($obj->{ErrMsg})) {
            my $type = exists($obj->{Type}) ? $obj->{Type} : 'No type';
            # Error != 0 or no Error property
            my $errorcode = exists($obj->{Error}) ? $obj->{Error} : 'No errorcode';
            $errmsg = "Error while getting data for Type: %s, ErrorCode: %s, ErrorMsg: %s.";
            $errmsg = sprintf($errmsg, $type, $errorcode, $obj->{ErrMsg});
        } else {
            # Not possible to parse output (non-JSON string)
            $errmsg = "Unknown output received while getting data. Output=$output.";
        }

        if (!$obj && $vmware_run_in_background) {
            # In background mode the output can be invalid because it can be incomplete
            # up till time when process in background becomes finished. To avoid
            # sending many false error messages in log, the error messages are visible
            # in logs only when debug is enabled.
            $self->debug($errmsg);
        } else {
            $self->edebug($errmsg);
        }
    }
    return $result;
}

# Execute mount VMware backup commands via mount-vmware interface.
# Params:
#  string $session_id CGI session identifier
#  arrayref $params parameters given to command line
#  integer $background if set, run script in background, otherwise run script normally (in foreground)
# Returns:
#  hashref script execution result (output, exitcode, error message...)
sub mount_vmware {
    my ($self, $session_id, $params, $background) = @_;
    my $ret;

    my $sir_support = $self->{bweb}->get_support(qw/sir/);
    if ($sir_support->{sir}->{vmware}) {
        my $mount_script = $self->get_location('mount-vmware');
        my $mount_params = '';
        if (defined($params) && ref($params) eq "ARRAY") {
            $mount_params = '"' . join('" "', @{$params}) . '"';
        }

        my $workdir = $self->get_vmware_workdir($session_id);
        my $bindir = $self->get_location('bacula-sd');
        my $confdir = $self->get_location('bacula-sd.conf');
        my $cmd = "$mount_script/mount-vmware -w \"$workdir\" -b \"$bindir\" -C \"$workdir/mount-cache\" -c \"$confdir\" -j $mount_params";
        if ($background) {
            $ret = $self->{bweb}->shell_exec($cmd, $workdir, undef, 1, 1, $session_id);
            if ($ret->{exitcode} != 0) {
                $self->edebug("Error while background executing $cmd. Error=$ret->{error}, Output=$ret->{output}.");
            }
        } else {
            $ret = $self->{bweb}->shell_exec($cmd);
            if ($ret->{exitcode} != 0) {
                $self->edebug("Error while direct executing $cmd. Error=$ret->{error}, Output=$ret->{output}.");
            }
        }
    }
    return $ret;
}

# Get VMware restore working directory.
# The working directory is prepared per session. It means that there can be opened
# many restore sessions by different users.
# Params:
#  string $session_id CGI session identifier
# Returns:
#  string VMware restore working directory
sub get_vmware_workdir {
    my ($self, $session_id) = @_;
    my $workdir = $self->{bweb}->get_component_workdir('Storage');
    if (!$workdir || ! -d $workdir) {
        # No SD working dir found, using default working directory
        $workdir = $default_vmware_workdir;
    }
    $workdir .= "/${vmware_subdir_prefix}${session_id}";
    return $workdir;
}

=head1 PACKAGE

    PersistentContent - module to manage persistent content

=head2 DESCRIPTION

 The package enables to remember on server side persistent information.
 Each of persistent content remembered by this module has assigned data identify
 information that describes basic data properties:
    - source type (file, database, remote content ...etc.)
    - source name (file name, db params, URL ...etc.)
    - action - string that says what action call just after getting data
      (it can be validator, formatter, parser... and others)

 The identify information is available after providing valid content identifier
 called "content_id".

 To get data in specified source type, the module should have defined function,
 which will be capable to read the data. Currently supported is only "file" source type.
 To support other types required is to add appropriate function to read data from
 the new source type.

=head2 USAGE

    # Get new identity for new content
    $identity = $persistent_content->new_identity('file', $content, 'some_action');

    # Register content as persistent
    $content_id = $content->set_content($identity, \&identity_setter);

    # Get persistent content
    $persistent_content->get_content($content_id, \&identity_getter);

    # Unregister persistent content (end work with data)
    $persistent_content->clear_content($content_id, \&identity_getter, \&identity_clear);

    # identity_getter, identity_setter and identity_clear are function handlers that enable to
    # store identity somewhere (for example, to remember identity in CGI scripts can be used
    # CGI session functions). identity_getter and identity_clear take one parameter (key),
    # identity_setter takes two parameters (key, value).

=cut

package Bweb::PersistentContent;

use base q/Bweb::Gui/;

# Supported source types and methods to read data and clear data.
my %source_types = (
    file => { read => \&read_file, clear => \&remove_file }
);

# Get new identity.
# Params:
#  string $source_type where data is stored
#  string $source_name param or params to get data
#  string $action action after getting data
# Returns:
#  hashref new identity
sub new_identity {
    my ($self, $source_type, $source_name, $action) = @_;
    my @rand_str = map{('a'..'z','A'..'Z',0..9)[rand 62]} 0..32;
    my $content_id = join('', @rand_str);
    my $ret = {
        content_id => $content_id,
        source_type => $source_type,
        source_name => $source_name,
        action => $action
    };
    return $ret;
}

# Get already registered identity.
# Params:
#  string $content_id content identifier
#  function $getter reference to handler that gets identity
# Returns:
#  hashref identity
sub get_identity {
    my ($self, $content_id, $getter) = @_;
    my $identity = {content_id => $content_id};
    my $key = $self->get_key($content_id);
    my $value = $getter->($key);
    if ($value) {
        my ($source_type, $source_name, $action) = split(':', $value, 3);
        $identity->{source_type} = $source_type;
        $identity->{source_name} = $source_name;
        $identity->{action} = $action;
    }
    return $identity;
}

# Set/store identity.
# It is used to register new persistent content.
# Params:
#  hashref $identity identity
#  function $setter reference to handler that sets identity
# Returns:
#  hashref content identifier
sub set_identity {
    my ($self, $identity, $setter) = @_;
    my $key = $self->get_key($identity->{content_id});
    my $value = $self->get_value($identity);
    $setter->($key, $value);
    my $ret = {content_id => $identity->{content_id}};
    return $ret;
}

# Clear/remove identity.
# It is used to unregister persistent content.
# Params:
#  string $content_id content identifier
#  function $clear reference to handler that clears identity
# Returns:
#  (depending on clear function) state about clearing success or not
sub clear_identity {
    my ($self, $content_id, $clear) = @_;
    my $key = $self->get_key($content_id);
    my $state = $clear->($key);
    return $state;
}

# Register new persistent content.
# If the 'content' parameter is identity type, then the identity
# is registered. If it isn't valid identity, the 'content' is returned as is.
# Params:
#  mixed $content valid identity
#  function $setter reference to handler that sets identity
# Returns:
#  content identifier (content_id) or original content
sub set_content {
    my ($self, $content, $setter) = @_;
    my $ret;
    if ($self->is_identity_type($content)) {
        $ret = $self->set_identity($content, $setter);
    } else {
        # Non-persistent content, return it as is
        $ret = $content;
    }
    return $ret;
}

# Get already registered content.
# Params:
#  string $content_id content identifier
#  function $getter reference to handler that gets identity
# Returns:
#  hashref persistent content with action property
sub get_content {
    my ($self, $content_id, $getter) = @_;
    my $identity = $self->get_identity($content_id, $getter);
    my $data = $self->data_action('read', $identity->{source_type}, $identity->{source_name});
    my $content = {content => $data, action => $identity->{action}};
    return $content;
}

# Clear (unregister) persistent content.
# Params:
#  string $content_id content identifier
#  function $getter reference to handler that gets identity
#  function $clear reference to handler that clears identity
#  integer $without_data_clearing don't clear data
# Returns:
#  integer 1 - successfully cleared, 0 - error during clearing
sub clear_content {
    my ($self, $content_id, $getter, $clear, $without_data_clearing) = @_;
    my $identity = $self->get_identity($content_id, $getter);
    my $ret = 1;
    unless ($without_data_clearing) {
        $ret = $self->data_action('clear', $identity->{source_type}, $identity->{source_name});
    }
    my $state = $self->clear_identity($content_id, $clear);
    my $cleared = ($ret && $state) ? 1 : 0;
    return $cleared;
}

# Do action on data (read, clear ...etc.)
# This module has to support action that is called here.
# Note: Data action is something different than 'action' in identity.
# Params:
#  string $action data action
#  string $source_type where data is stored
#  string $source_name param or params to get data
# Returns:
#  (depending on called action) data - on read, 0/1 - on clear
sub data_action {
    my ($self, $action, $source_type, $source_name) = @_;
    my $result;
    if (exists($source_types{$source_type}->{$action})) {
        my $func = $source_types{$source_type}->{$action};
        $result = $self->$func($source_name);
    }
    return $result;
}

# Validate identity.
# Params:
#  hashref $identity identity
# Returns:
#  integer 1 - identity valid, 0 - identity invalid
sub is_identity_type {
    my ($self, $identity) =  @_;
    my $valid = 0;
    if (ref($identity) eq "HASH") {
        if (exists($identity->{content_id}) && exists($identity->{source_type}) && exists($identity->{source_name}) && exists($identity->{action})) {
            $valid = 1;
        }
    }
    return $valid;
}

# Get key to store identity.
# The key is used in getter, setter and clear functions.
# Params:
#  string $content_id content identifier
# Returns:
#  string key
sub get_key {
    my ($self, $content_id) = @_;
    my $key = "content_$content_id";
    return $key;
}

# Get value to store identity.
# The value is used in setter function.
# Params:
#  hashref $identity identity
# Returns:
#  string value
sub get_value {
    my ($self, $identity) = @_;
    my $value = "$identity->{source_type}:$identity->{source_name}:$identity->{action}";
    return $value;
}

### Methods (plugins) to support specified data types ###

# File type methods

# Read file data.
# Params:
#  string $file file full path
# Returns:
#  string file content
sub read_file {
    my ($self, $file) = @_;
    my $max_file_size = 1048576; # 1MiB
    my $chunk_size = 1024;
    my $content = '';
    if (-f $file) {
        if (open(my $fh, '<', $file)) {
            my $length = 0;
            my $ret;
            my $chunk = '';
            do {
                $length = length($chunk);
                $ret = sysread($fh, $chunk, $chunk_size, $length);
            } while ($ret && $length < $max_file_size);
            close($fh);
            $content = $chunk;
        } else {
            $self->edebug("Could not open '$file'. Error=$!.");
        }
    } else {
        $self->edebug("File '$file' does not exist or the file is not accessible.");
    }
    return $content;
}

# Remove file data.
# Params:
#  string $file file full path
# Returns:
#  integer remove state, 1 - success, 0 - error
sub remove_file {
    my ($self, $file) = @_;
    my $state = unlink($file);
    if (!$state) {
        $self->edebug("Unable to remove file '$file'. Error=$!.");
    }
    # Note: unlink() returns number of files it successfully deleted
    return $state;
}

1;
